#!/usr/bin/env python

#****************************************************************************
# cmdline.py, provides a class to read and execute command line arguments
#
# TreeLine, an information storage program
# Copyright (C) 2004, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#*****************************************************************************

"""
Usage:

   treeline [infile]

-or- (non-interactive):

   treeline [import-option] [export-option] [misc-options] infile [infile2 ...]

Import-options:

   --import-tabbed     import a tab indented text file
   --import-table      import a tab-delimitted text table, one node per line
   --import-lines      import plain text, one node per line
   --import-para       import plain text, one node per paragraph
   --import-treepad    import a Treepad text-node file
   --import-xbel       import an XML bookmark file in XBEL format
   --import-mozilla    import an html bookmark file in Mozilla format

Export-options:

   --export-html       export a single HTML file
   --export-dir        export HTML in directories
   --export-xslt       export an XSLT file
   --export-tabbed     export a tab indented text file
   --export-table      export a text table of the first children only
   --export-xbel       export an XML bookmark file in XBEL format
   --export-mozilla    export an html bookmark file in Mozilla format

Misc-options:

   -o, --outfile=FILE  the output filename, not used with multiple infiles
   -n, --no-root       exclude the root node form the output if applicable
   --add-header        add a header and footer to HTML exports
   --indent=NUM        change the indent amount for HTML exports (default = 20)
   -q, --quiet         supress normal status information, only give errors
   -h, --help          display this information and exit

No more than one import-option and one export-option may be specified.  If
either are not present, the native TreeLine file format is assumed.

The output filename option can only be specified if there is only one input
file.  If it is not specified, the input's base file name is used with the
appropriate output file extension.  If the extensions are the same, an
underscore is added before the extension.  Note that this avoids overwriting
the input file, but other files may be overwritten without notification if
they share the output file's name.
"""

import sys, os.path
from treedoc import TreeDoc, ReadFileError
import docparam


class CmdLine:
    """Parses and executes command line arguments for file translations"""
    options = 'o:nqh'
    longOptions = ['import-tabbed', 'import-table', 'import-lines', \
                   'import-para', 'import-treepad', 'import-xbel', \
                   'import-mozilla', 'export-html', 'export-dir', \
                   'export-xslt', 'export-tabbed', 'export-table', \
                   'export-xbel', 'export-mozilla', 'outfile=', 'no-root', \
                   'indent=', 'quiet', 'help']
    def __init__(self, opts, args):
        self.doc = None
        self.outfile = ''
        self.includeRoot = 1
        self.addHeader = 0
        self.indent = 20
        self.quiet = 0
        importType = ''
        exportType = ''
        for opt, arg in opts:
            if opt in ('-h', '--help'):
                print __doc__
                return
            elif opt.startswith('--import-'):
                if importType:
                    print __doc__
                    sys.exit(2)
                importType = opt[9:]
            elif opt.startswith('--export-'):
                if exportType:
                    print __doc__
                    sys.exit(2)
                exportType = opt[9:]
            elif opt in ('-o', '--outfile'):
                self.outfile = arg
                if len(args) > 1:
                    print __doc__
                    sys.exit(2)
            elif opt in ('-n', '--no-root'):
                self.includeRoot = 0
            elif opt == '--add-header':
                self.addHeader = 1
            elif opt == '--indent':
                try:
                    self.indent = int(arg)
                except ValueError:
                    print __doc__
                    sys.exit(2)
            elif opt in ('-q', '--quiet'):
                self.quiet = 1
        if not args:
            print __doc__
            sys.exit(2)
        if not importType:
            importType = 'trl'
        if not exportType:
            exportType = 'trl'
        errorFlag = 0
        for fileName in args:
            self.doc = TreeDoc()
            try:
                self.importFile(fileName, importType)
            except IOError:
                print 'Error - could not read file %s' % fileName
                errorFlag = 1
                continue
            except ReadFileError, e:
                print 'Error in %s - %s' % (fileName, e)
                errorFlag = 1
                continue
            newFileName = self.exportFile(fileName, exportType)
            if newFileName:
                if not self.quiet:
                    print 'File "%s" (%s type) was exported to "%s" (%s type)' \
                          % (fileName, importType, newFileName, exportType)
            else:
                errorFlag = 1
        if errorFlag:
            sys.exit(1)

    def importFile(self, fileName, importType):
        """Import file using importType"""
        if importType == 'trl':
            importType = 'File'
        eval('self.doc.read%s' % importType.title())(fileName)

    def exportFile(self, oldFileName, exportType):
        """Export file using exportType, return new filename on success"""
        if exportType == 'trl':
            fileName = self.newFileName(oldFileName, '.trl')
            result = self.doc.writeFile(fileName)
        elif exportType == 'html':
            fileName = self.newFileName(oldFileName, '.html')
            result = self.doc.exportHtml(fileName, self.doc.root, \
                                         self.includeRoot, 0, self.indent, \
                                         self.addHeader)
        elif exportType == 'dir':
            dirName = os.path.split(oldFileName)[0]
            if not dirName:
                dirName = '.'
            fileName = dirName
            result = self.doc.exportDir(dirName, self.doc.root, self.addHeader)
        elif exportType == 'xslt':
            fileName = self.newFileName(oldFileName, '.xsl')
            result = self.doc.exportXslt(fileName, self.doc.root, self.indent)
        elif exportType == 'tabbed':
            fileName = self.newFileName(oldFileName, '.txt')
            result = self.doc.exportTabbedTitles(fileName, self.doc.root, \
                                                 self.includeRoot)
        elif exportType == 'table':
            fileName = self.newFileName(oldFileName, '.tbl')
            result = self.doc.exportTable(fileName, self.doc.root)
        elif exportType == 'xbel':
            fileName = self.newFileName(oldFileName, '.xml')
            result = self.doc.exportXbel(fileName, self.doc.root)
        elif exportType == 'mozilla':
            fileName = self.newFileName(oldFileName, '.html')
            result = self.doc.exportHtmlBookmarks(fileName, self.doc.root)
        if result:
            return fileName
        return ''

    def newFileName(self, oldFileName, newExt):
        """Return a new filename with a new extension, add an underscore to
           base name if the names are the same"""
        if self.outfile:
            return self.outfile
        baseName, oldExt = os.path.splitext(oldFileName)
        if oldExt == newExt:
            baseName += '_'
        return baseName + newExt
