#!/usr/bin/env python

#****************************************************************************
# fieldformat.py, provides non-GUI base classes for field formating
#
# TreeLine, an information storage program
# Copyright (C) 2003, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#****************************************************************************

import re
from xml.sax.saxutils import escape
from gennumber import GenNumber, GenNumberError
from gendate import GenDate, GenDateError
from gentime import GenTime, GenTimeError
from genboolean import GenBoolean, GenBooleanError
import treedoc
import docparam

_errorStr = '#####'

class TextFormat:
    """Holds format info for a normal text field"""
    typeName = 'Text'
    sortSequence = 20

    #field format edit options:
    defaultFormat = ''
    formatMenuList = ()
    htmlOption = 1

    hasEditChoices = 0

    def __init__(self, name, attrs={}):
        """Any prefix, suffix, html info in attrs dict"""
        self.name = name
        self.prefix = attrs.get(u'prefix', '')
        self.suffix = attrs.get(u'suffix', '')
        # defaults to no html (line breaks preserved)
        self.html = attrs.get(u'html', '').startswith('y') and 1 or 0
        self.format = attrs.get(u'format', self.defaultFormat)
        self.parentLevel = 0
        self.useFileInfo = 0
        self.showInDialog = 1
        self.initFormat()

    def initFormat(self):
        """Called by base init and after class change"""
        pass

    def sepName(self):
        """Return name enclosed with {* *} separators"""
        if not self.useFileInfo:
            return '{*%s*}' % self.name
        return '{*!%s*}' % self.name

    def writeXml(self):
        """Return text for xml attributes"""
        text = ' type="%s"' % self.__class__.typeName
        if self.format:
            text += ' format="%s"' % escape(self.format, treedoc.escDict)
        if self.prefix:
            text += ' prefix="%s"' % escape(self.prefix, treedoc.escDict)
        if self.suffix:
            text += ' suffix="%s"' % escape(self.suffix, treedoc.escDict)
        if self.html:
            text += ' html="y"'
        return text

    def outputText(self, inText, titleMode):
        """Return formatted text, properly escaped if not in titleMode"""
        if titleMode:
            inText = inText.split('\n', 1)[0]    # truncate to 1st line
        elif not self.html:
            inText = escape(inText).replace('\n', '<br />')
        if docparam.formHtml or titleMode:
            text = '%s%s%s' % (self.prefix, inText, self.suffix)
        else:
            text = '%s%s%s' % (escape(self.prefix), inText, \
                               escape(self.suffix))
        return text

    def editText(self, inText, editFormats):
        """Return text in edit format, found in editFormats dict"""
        return inText

    def extractEditValue(self, inText, editFormats):
        """Return stored text from edited text using editFormats dict"""
        return inText

    def sortValue(self, inText):
        """Return value to be compared to like types for sorting"""
        return inText.upper()

    def xslText(self):
        """Return what we need to write into an  XSL file for this type"""
        return '<xsl:value-of select="./%s"/>' % self.name

    def xslTestText(self):
        """Return XSL file test for data existance"""
        return 'normalize-space(./%s)' % self.name


class LongTextFormat(TextFormat):
    """Holds format info for a long text field"""
    typeName = 'LongText'
    sortSequence = 20
    def __init__(self, name, attrs={}):
        """Any format, prefix, suffix, html info in attrs dict"""
        TextFormat.__init__(self, name, attrs)


class NumberFormat(TextFormat):
    """Holds format info for a number field"""
    typeName = 'Number'
    sortSequence = 10

    #field format edit options:
    defaultFormat = '#.##'
    formatMenuList = (('Optional Digit\t#', '#'), ('Required Digit\t0', '0'), \
                      ('Digit or Space\t<space>', ' '), None, \
                      ('Optional Sign\t-', '-'), ('Required Sign\t+', '+'), \
                      None, ('Exponent (capital)\tE', 'E'), \
                      ('Exponent (small)\te', 'e'))

    def __init__(self, name, attrs={}):
        """Any format, prefix, suffix, html info in attrs dict"""
        TextFormat.__init__(self, name, attrs)

    def outputText(self, inText, titleMode):
        """Return formatted text, properly escaped if not in titleMode"""
        try:
            text = GenNumber(inText).numStr(self.format)
        except GenNumberError:
            text = _errorStr
        return TextFormat.outputText(self, text, titleMode)

    def editText(self, inText, editFormats):
        """Return text in edit format, found in editFormats dict"""
        try:
            return GenNumber(inText).numStr(editFormats[self.typeName])
        except GenNumberError:
            return ''     #  no value indicates blank field or an error

    def extractEditValue(self, inText, editFormats):
        """Return stored text from edited text using editFormats dict"""
        try:
            return repr(GenNumber().setFromStr(inText, \
                                        editFormats[self.__class__.typeName]))
        except GenNumberError:
            return ''      #  no value indicates blank field or an error

    def sortValue(self, inText):
        """Return value to be compared to like types for sorting"""
        try:
            return GenNumber(inText).num
        except GenNumberError:
            return ''


class ChoiceFormat(TextFormat):
    """Holds format info for a field with one of several text options"""
    typeName = 'Choice'
    sortSequence = 20
    editSep = '/'

    #field format edit options:
    defaultFormat = '1/2/3/4'
    formatMenuList = (('Separator\t/', '/'), None, \
                      ('"/" Character\t//', '//'), None, \
                      ('Example\t1/2/3/4', '1/2/3/4'))

    hasEditChoices = 1

    def __init__(self, name, attrs={}):
        """Any format, prefix, suffix, html info in attrs dict"""
        TextFormat.__init__(self, name, attrs)

    def initFormat(self):
        """Called by base init and after class change"""
        self.formatList = self.splitText(self.format)

    def outputText(self, inText, titleMode):
        """Return formatted text, properly escaped if not in titleMode"""
        if inText not in self.formatList:
            inText = _errorStr
        return TextFormat.outputText(self, inText, titleMode)

    def editText(self, inText, editFormats):
        """Return text in edit format"""
        if inText in self.formatList:
            return inText
        return ''       #  no value indicates blank field or an error

    def extractEditValue(self, inText, editFormats):
        """Return stored text from edited text"""
        if inText in self.formatList:
            return inText
        return ''       #  no value indicates blank field or an error

    def getEditChoices(self, editFormats, currentText=''):
        """Return list of choices for combo box, 
           each a tuple of edit text and annotated text"""
        return [(text, text) for text in self.formatList]

    def splitText(self, textStr):
        """Split textStr using editSep, double sep's become char"""
        return [text.strip().replace('\0', self.editSep) for text in \
                textStr.replace(self.editSep * 2, '\0').\
                split(self.editSep)]


class CombinationFormat(ChoiceFormat):
    """Holds format info for a field of combinations of text options"""
    typeName = 'Combination'

    outputSepList = (',', ';', ':', '|', '/', '\\', '~')

    def __init__(self, name, attrs={}):
        """Any format, prefix, suffix, html info in attrs dict"""
        ChoiceFormat.__init__(self, name, attrs)

    def initFormat(self):
        """Called by base init and after class change"""
        ChoiceFormat.initFormat(self)
        fullFormat = ''.join(self.formatList)
        try:
            self.sep = [sep for sep in CombinationFormat.outputSepList \
                        if sep not in fullFormat][0] + ' '
        except IndexError:
            self.sep = CombinationFormat.outputSepList[0] + ' '

    def sortedChoices(self, inText):
        """Return tuple of choices from inText sorted like format and
           1 if all splits are valid and included"""
        choices = self.splitText(inText)
        sortedChoices = [text for text in self.formatList if text in choices]
        if len(choices) == len(sortedChoices):
            return (sortedChoices, 1)
        else:
            return (sortedChoices, 0)

    def outputText(self, inText, titleMode):
        """Return formatted text, properly escaped if not in titleMode"""
        choices, valid = self.sortedChoices(inText)
        if valid:
            result = self.sep.join(choices)
        else:
            result = _errorStr
        return TextFormat.outputText(self, result, titleMode)

    def editText(self, inText, editFormats):
        """Return text in edit format"""
        for choice in self.splitText(inText):
            if choice not in self.formatList:
                return ''      #  no value indicates blank field or an error
        return inText

    def extractEditValue(self, inText, editFormats):
        """Return stored text from edited text"""
        choices, valid = self.sortedChoices(inText)
        if valid:
            return self.editSep.join(choices)
        else:
            return ''       #  no value indicates blank field or an error

    def getEditChoices(self, editFormats, currentText=''):
        """Return list of choices for combo box,
           each a tuple of edit text and annotated text"""
        currentChoices, valid = self.sortedChoices(currentText)
        nonChoices = [text for text in self.formatList \
                      if text not in currentChoices]
        results = []
        for choice in nonChoices:      # menu entries to add a choice
            allChoices = currentChoices + [choice]
            allChoices = [text for text in self.formatList \
                          if text in allChoices]
            results.append((self.editSep.join(allChoices), choice + '  (add)'))
        results.append((None, None))   # separator
        for choice in currentChoices:   # menu entries to remove a choice
            allChoices = currentChoices[:]
            allChoices.remove(choice)
            allChoices = [text for text in self.formatList \
                          if text in allChoices]
            results.append((self.editSep.join(allChoices), \
                            choice + '  (remove)'))
        return results


class DateFormat(TextFormat):
    """Holds format info for a date field"""
    typeName = 'Date'
    sortSequence = 5

    #field format edit options:
    defaultFormat = 'mmmm d, yyyy'
    formatMenuList = (('Day (1 or 2 digits)\td', 'd'), \
                      ('Day (2 digits)\tdd', 'dd'), None, \
                      ('Month (1 or 2 digits)\tm', 'm'), \
                      ('Month (2 digits)\tmm', 'mm'), \
                      ('Month Abbreviation\tmmm', 'mmm'), \
                      ('Month Name\tmmmm', 'mmmm'), None, \
                      ('Year (2 digits)\tyy', 'yy'), \
                      ('Year (4 digits)\tyyyy', 'yyyy'), None, \
                      ('Weekday (1 digit)\tw', 'w'), \
                      ('Weekday Abbreviation\twww', 'www'), \
                      ('Weekday Name\twwww', 'wwww'))

    hasEditChoices = 1

    def __init__(self, name, attrs={}):
        """Any format, prefix, suffix, html info in attrs dict"""
        TextFormat.__init__(self, name, attrs)

    def outputText(self, inText, titleMode):
        """Return formatted text, properly escaped if not in titleMode"""
        try:
            text = GenDate(inText).dateStr(self.format)
        except GenDateError:
            text = _errorStr
        return TextFormat.outputText(self, text, titleMode)

    def editText(self, inText, editFormats):
        """Return text in edit format, found in editFormats dict"""
        try:
            return GenDate(inText).dateStr(editFormats[self.typeName])
        except GenDateError:
            return ''     #  no value indicates blank field or an error

    def extractEditValue(self, inText, editFormats):
        """Return stored text from edited text using editFormats dict"""
        try:
            return repr(GenDate().setFromStr(inText, \
                                             editFormats[self.typeName]))
        except GenDateError:
            return ''      #  no value indicates blank field or an error

    def getEditChoices(self, editFormats, currentText=''):
        """Return list of choices for combo box, 
           each a tuple of edit text and annotated text"""
        today = GenDate().dateStr(editFormats[self.typeName])
        yesterday = (GenDate() - 1).dateStr(editFormats[self.typeName])
        tomorrow = (GenDate() + 1).dateStr(editFormats[self.typeName])
        return [(today, '%s (today)' % today), \
                (yesterday, '%s (yesterday)' % yesterday), \
                (tomorrow, '%s (tomorrow)' % tomorrow)]


class TimeFormat(TextFormat):
    """Holds format info for a time field"""
    typeName = 'Time'
    sortSequence = 6

    #field format edit options:
    defaultFormat = 'h:MM:SS aa'
    formatMenuList = (('Hour (0-23, 1 or 2 digits)\tH', 'H'), \
                      ('Hour (00-23, 2 digits)\tHH', 'HH'), \
                      ('Hour (1-12, 1 or 2 digits)\th', 'h'), \
                      ('Hour (01-12, 2 digits)\thh', 'hh'), None, \
                      ('Minute (1 or 2 digits)\tM', 'M'), \
                      ('Minute (2 digits)\tMM', 'MM'), None, \
                      ('Second (1 or 2 digits)\tS', 'S'), \
                      ('Second (2 digits)\tSS', 'SS'), \
                      ('Fractional Seconds\ts', 's'), None, \
                      ('AM/PM\tAA', 'AA'), ('am/pm\taa','aa'))

    hasEditChoices = 1

    def __init__(self, name, attrs={}):
        """Any format, prefix, suffix, html info in attrs dict"""
        TextFormat.__init__(self, name, attrs)

    def outputText(self, inText, titleMode):
        """Return formatted text, properly escaped if not in titleMode"""
        try:
            text = GenTime(inText).timeStr(self.format)
        except GenTimeError:
            text = _errorStr
        return TextFormat.outputText(self, text, titleMode)

    def editText(self, inText, editFormats):
        """Return text in edit format, found in editFormats dict"""
        try:
            return GenTime(inText).timeStr(editFormats[self.typeName])
        except GenTimeError:
            return ''     #  no value indicates blank field or an error

    def extractEditValue(self, inText, editFormats):
        """Return stored text from edited text using editFormats dict"""
        try:
            return repr(GenTime(inText))
        except GenTimeError:
            return ''      #  no value indicates blank field or an error

    def getEditChoices(self, editFormats, currentText=''):
        """Return list of choices for combo box, 
           each a tuple of edit text and annotated text"""
        now = GenTime().timeStr(editFormats[self.typeName])
        choices = [(now, '%s (now)' % now)]
        for hr in (6, 9, 12, 15, 18, 21, 0):
            time = GenTime((hr, 0)).timeStr(editFormats[self.typeName])
            choices.append((time, time))
        return choices


class BooleanFormat(ChoiceFormat):
    """Holds format info for a bool field"""
    typeName = 'Boolean'
    sortSequence = 1

    #field format edit options:
    defaultFormat = 'yes/no'
    formatMenuList = (('true/false', 'true/false'), ('T/F', 'T/F'), None, \
                      ('yes/no', 'yes/no'), ('Y/N', 'Y/N'), None, \
                      ('1/0', '1/0'))

    hasEditChoices = 1

    def __init__(self, name, attrs={}):
        """Any format, prefix, suffix, html info in attrs dict"""
        ChoiceFormat.__init__(self, name, attrs)

    def outputText(self, inText, titleMode):
        """Return formatted text, properly escaped if not in titleMode"""
        if inText not in self.formatList:
            try:
                inText = GenBoolean(inText).boolStr(self.format)
            except GenBooleanError:
                inText = _errorStr
        return TextFormat.outputText(self, inText, titleMode)

    def editText(self, inText, editFormats):
        """Return text in edit format"""
        if inText in self.formatList:
            return inText
        try:
            return GenBoolean(inText).boolStr(self.format)
        except GenBooleanError:
            return ''     #  no value indicates blank field or an error

    def extractEditValue(self, inText, editFormats):
        """Return stored text from edited text"""
        try:
            return repr(GenBoolean(inText))
        except GenBooleanError:
            if inText in self.formatList:
                return inText
            return ''      #  no value indicates blank field or an error


class URLFormat(TextFormat):
    """Holds format info for a field with a URL path"""
    typeName = 'URL'
    sortSequence = 8

    htmlOption = 0
    URLMethod = 'http://'

    def __init__(self, name, attrs={}):
        """Any format, prefix, suffix, html info in attrs dict"""
        TextFormat.__init__(self, name, attrs)

    def initFormat(self):
        """Called by base init and after class change"""
        self.html = 1

    def outputText(self, inText, titleMode):
        """Return formatted text, properly escaped and with
           a link reference if not in titleMode"""
        if titleMode:
            return TextFormat.outputText(self, inText, titleMode)
        paths = inText.split('\n')
        results = []
        for url in paths:
            path = url
            if not ':' in path:
                path = '%s%s' % (self.URLMethod, path)
            path = '<a href="%s">%s</a>' % (path, url)
            results.append(TextFormat.outputText(self, path, titleMode))
        return '<br />'.join(results)

    def xslText(self):
        """Return what we need to write into an  XSL file for this type"""
        return '<xsl:for-each select = "./%s"><xsl:choose>' \
               '<xsl:when test="contains(., \':\')"><a href="{.}">' \
               '<xsl:value-of select="."/></a></xsl:when><xsl:otherwise>' \
               '<a href="%s{.}"><xsl:value-of select="."/></a>' \
               '</xsl:otherwise></xsl:choose></xsl:for-each>' % \
               (self.name, self.URLMethod)


class PathFormat(URLFormat):
    """Holds format info for a field with a local path"""
    typeName = 'Path'

    URLMethod = 'file:///'

    def __init__(self, name, attrs={}):
        """Any format, prefix, suffix, html info in attrs dict"""
        URLFormat.__init__(self, name, attrs)


class EmailFormat(URLFormat):
    """Holds format info for a field with a local path"""
    typeName = 'Email'

    URLMethod = 'mailto:'

    def __init__(self, name, attrs={}):
        """Any format, prefix, suffix, html info in attrs dict"""
        URLFormat.__init__(self, name, attrs)


class PictureFormat(TextFormat):
    """Holds format info for a field with a link to a picture"""
    typeName = 'Picture'
    sortSequence = 8

    htmlOption = 0

    def __init__(self, name, attrs={}):
        """Any format, prefix, suffix, html info in attrs dict"""
        TextFormat.__init__(self, name, attrs)

    def initFormat(self):
        """Called by base init and after class change"""
        self.html = 1

    def outputText(self, inText, titleMode):
        """Return formatted text, properly escaped and with
           a link to the picture if not in titleMode"""
        if titleMode:
            return TextFormat.outputText(self, inText, titleMode)
        paths = inText.split('\n')
        results = ['<img src="%s">' % url for url in paths]
        return '<br />'.join(results)


class ParentFormat(TextFormat):
    """Placeholder format for references to specific parents"""
    typeName = 'Parent'

    def __init__(self, name, parentLevel=1):
        TextFormat.__init__(self, name, {})
        self.parentLevel = parentLevel

    def sepName(self):
        """Return name enclosed with {* *} separators"""
        return '{*%s%s*}' % (self.parentLevel * '*', self.name)

    def xslText(self):
        """Return what we need to write into an  XSL file for this type"""
        return '<xsl:value-of select="%s%s"/>' % (self.parentLevel * '../', \
                                                  self.name)

    def xslTestText(self):
        """Return XSL file test for data existance"""
        return 'normalize-space(%s%s)' % (self.parentLevel * '../', self.name)


class AncestorFormat(TextFormat):
    """Placeholder format for references to any parent with data"""
    typeName = 'Ancestor'

    def __init__(self, name):
        TextFormat.__init__(self, name, {})
        self.parentLevel = -1

    def sepName(self):
        """Return name enclosed with {*? *} separators"""
        return '{*?%s*}' % (self.name)

    def xslText(self):
        """Return what we need to write into an  XSL file for this type"""
        return '<xsl:value-of select="ancestor::*/%s"/>' % self.name

    def xslTestText(self):
        """Return XSL file test for data existance"""
        return 'normalize-space(ancestor::*/%s)' % self.name
