#!/usr/bin/env python

#****************************************************************************
# formatlist.py, provides non-GUI base classes for stroing node format info
#
# TreeLine, an information storage program
# Copyright (C) 2004, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#****************************************************************************

from UserList import UserList
import treeformat, docparam

class FormatList(UserList):
    """A list to store node formats in the doc"""
    def __init__(self, initList=[], setDefaults=0, defaultFieldAttrs={}):
        UserList.__init__(self, initList)
        if setDefaults:
            self.data.extend([treeformat.NodeFormat('ROOT', 'Name', \
                                                    defaultFieldAttrs), \
                              treeformat.NodeFormat('DEFAULT', 'Name', \
                                                    defaultFieldAttrs)])
            self.data[0].childType = 'DEFAULT'

    def nameList(self):
        """Return list of names of format items"""
        return [form.name for form in self.data]

    def findFormat(self, name):
        """Return format matching name or None"""
        try:
            return self.data[self.nameList().index(name)]
        except ValueError:
            return None

    def findDefaultFormat(self):
        """Return default or non-root format"""
        dflt = self.findFormat('DEFAULT')
        if not dflt:
            for format in self.data:
                if format.name != 'ROOT':
                    return format
            return self.data[0]
        return dflt

    def addIfMissing(self, format):
        """Add format to list if not a duplicate"""
        if format not in self.data:
            self.data.append(format)

    def removeQuiet(self, format):
        """Remove from list if there"""
        try:
            self.remove(format)
        except ValueError:
            pass

    def renameFields(self, nameDict, rootItem):
        """Rename data fields for all items, doesn't change node format"""
        for item in rootItem.descendantList(1):
            for key in item.data.keys():
                try:
                    item.data[nameDict[(item.nodeFormat.name, key)]] = \
                            item.data[key]
                    del item.data[key]
                except KeyError:
                    pass

    def updateAllLineFields(self):
        """Re-find fields to update format lines for any changes
           in the fieldLists"""
        for format in self.data:
            format.updateLineFields()
        docparam.fileInfoItem.nodeFormat.updateLineFields()

    def update(self, updateList, renameDict):
        """Update formats to be equiv to newList,
           changed names are in renameDict"""
        for key in renameDict.keys():
            format = self.findFormat(key)
            if format:
                format.name = renameDict[key]
        newList = []
        for updateFormat in updateList:
            format = self.findFormat(updateFormat.name)
            if format:
                format.fieldList = updateFormat.fieldList
                format.lineList = updateFormat.lineList
                format.childType = updateFormat.childType
                format.sibPrefix = updateFormat.sibPrefix
                format.sibSuffix = updateFormat.sibSuffix
                newList.append(format)
            else:
                newList.append(updateFormat)
        self.data = newList
