#!/usr/bin/env python

#****************************************************************************
# output.py, provides non-GUI base classes for output of node info
#
# TreeLine, an information storage program
# Copyright (C) 2004, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#****************************************************************************

from UserList import UserList
import docparam


class OutputItem:
    """Tree item rich text output class - stores text, level & height"""
    def __init__(self, textLines, level):
        self.textLines = textLines
        if docparam.spaceBetween:
            self.textLines.append(u'<nobr> </nobr>') # qt3 - ignores <br><br>
        self.level = level
        self.height = len(self.textLines)
        self.firstSibling = 0
        self.lastSibling = 0
        self.hasChildren = 0
        self.prefix = ''
        self.suffix = ''

    def addBreaks(self):
        """Add <br> elements to format lines"""
        self.textLines = [line + u'<br />' for line in self.textLines]

    def addIndents(self, prevLevel, nextLevel):
        """Add <div> elements to format indented lines, return this level"""
        for num in range(self.level - prevLevel):
            self.textLines[0] = u'<div>%s' % self.textLines[0]
        for num in range(self.level - nextLevel):
            self.textLines[-1] = u'%s</div>' % self.textLines[-1]
        return self.level

    def equalPrefix(self, other):
        """Return 1 if prefix and suffix are equivalent"""
        return self.prefix.strip().upper() == other.prefix.strip().upper() and \
               self.suffix.strip().upper() == other.suffix.strip().upper()

    def textList(self, withPrefix=0, withSuffix=0):
        """Return list of text lines, optionally with prefix/suffix"""
        textList = self.textLines[:]
        if not textList:
            textList = ['']
        if withPrefix:
            textList[0] = self.prefix + textList[0]
        if withSuffix:
            textList[-1] += self.suffix
        return textList


class OutputGroup(UserList):
    """List of OutputItems - splits into segments & adds indentation"""
    def __init__(self, initList=[]):
        UserList.__init__(self, initList)

    def addBreaks(self):
        """Add <br> elements to format lines"""
        for item in self.data:
            item.addBreaks()

    def addIndents(self, prevLevel=0):
        """Add <div> elements to indent items"""
        for num in range(len(self.data)):
            nextLevel = 0
            if num + 1 < len(self.data):
                nextLevel = self.data[num + 1].level
            prevLevel = self.data[num].addIndents(prevLevel, nextLevel)

    def addPrefix(self):
        """Add prefix and suffix text to group items"""
        closeRqd = None
        for item, next in map(None, self.data, self.data[1:]):
            if item.prefix and item.textLines and closeRqd == None:
                item.textLines[0] = item.prefix + item.textLines[0]
                closeRqd = item.suffix
            if closeRqd != None and (item.lastSibling or item.hasChildren or \
                                     next == None or \
                                     not item.equalPrefix(next)):
                if item.textLines:
                    item.textLines[-1] = item.textLines[-1] + closeRqd
                else:
                    item.textLines = [closeRqd]
                closeRqd = None

    def joinPrefixItems(self):
        """Merges adjacent items with same prefix and level for printing"""
        newList = []
        mergeList = OutputGroup()
        for item in self.data:
            if mergeList and (item.level != mergeList[0].level or \
                              not item.prefix or \
                              not item.equalPrefix(mergeList[0])):
                mergeList.mergeGroup()
                newList.append(mergeList[0])
                mergeList.data = []
            mergeList.append(item)
        mergeList.mergeGroup()
        newList.append(mergeList[0])
        self.data = newList

    def mergeGroup(self):
        """Merge self (group of adjacent items)"""
        if len(self.data) < 2:
            return
        mainItem = self.data[0]
        for item in self.data[1:]:
            mainItem.textLines.extend(item.textLines)
        mainItem.height = reduce(lambda x,y: x+y, [item.height for item in \
                                                   self.data])
        mainItem.lastSibling = self.data[-1].lastSibling
        mainItem.hasChildren = self.data[-1].hasChildren

    def splitColumns(self, numColumns):
        """Split output into even columns, return list with a group for each"""
        if numColumns < 2:
            return [self]
        if len(self.data) <= numColumns:
            return [OutputGroup([item]) for item in self]
        groupList = []
        numEach = len(self.data) / numColumns
        for colNum in range(numColumns - 1):
            groupList.append(OutputGroup(self.data[colNum * numEach : \
                                                   (colNum+1) * numEach]))
        groupList.append(OutputGroup(self.data[(numColumns-1) * numEach : ]))
        numChanges = 1
        while numChanges:
            numChanges = 0
            for colNum in range(numColumns - 1):
                if groupList[colNum].totalHeight() > groupList[colNum+1].\
                   totalHeight() + groupList[colNum][-1].height:
                    groupList[colNum+1].insert(0, groupList[colNum][-1])
                    groupList[colNum] = groupList[colNum][:-1]
                    numChanges += 1
                if groupList[colNum].totalHeight() + groupList[colNum+1][0].\
                   height <= groupList[colNum+1].totalHeight():
                    groupList[colNum].append(groupList[colNum+1][0])
                    groupList[colNum+1] = groupList[colNum+1][1:]
                    numChanges += 1
        return groupList

    def setHeights(self, textHtFunc, totalWidth, indent, fixList):
        """Set item heights after adding prefixes and applying fixes,
           textHtFunc is a callback to get height from text,
           fixList is a list of (regexp, replacement) tuples"""
        for prevItem, item in zip([None] + self.data, self.data):
            width = totalWidth - indent * item.level
            if not prevItem or item.level != prevItem.level or not item.prefix:
                textList = item.textList(1, 1)
                item.height = self.textHeight(textList, textHtFunc, width, \
                                              fixList)
            else:
                prevList = prevItem.textList(1, 1)
                prevHeight = self.textHeight(prevList, textHtFunc, width, \
                                             fixList)
                fullList = prevItem.textList(1, 0) + item.textList(0, 1)
                fullHeight = self.textHeight(fullList, textHtFunc, width, \
                                             fixList)
                item.height = fullHeight - prevHeight

    def textHeight(self, textList, textHtFunc, width, fixList):
        """Return text height after joining and applying fixes"""
        text = u'<br>\n'.join(textList)
        for exp, repl in fixList:
            text = exp.sub(repl, text)
        return textHtFunc(text, width)

    def totalHeight(self):
        """Return the combined height of items in the group"""
        return reduce(lambda x, y: x+y, [item.height for item in self])

    def splitPages(self, pageHeight, pageLevelBreaks=None):
        """Split output by pageHeight, return list with a group for each"""
        groupList = [OutputGroup()]
        height = 0
        for item in self.data:
            if height + item.height > pageHeight:
                height = 0
                groupList.append(OutputGroup())
            height += item.height
            groupList[-1].append(item)
        if pageLevelBreaks != None:
            activeLevels = {}
            for pageList in groupList:
                for item in pageList:
                    if item.firstSibling and item.level:
                        activeLevels[item.level] = 1
                    if item.lastSibling and item.level:
                        activeLevels[item.level] = 0
                pageLevelBreaks.append([key for key in activeLevels.keys() \
                                        if activeLevels[key]])
        return groupList

    def getLines(self):
        """Return full list of text lines from this group"""
        lines = []
        for item in self.data:
            lines.extend(item.textLines)
        return lines
