#!/usr/bin/env python

#****************************************************************************
# textedit2.py, provides classes for the qt2 text editors
#
# TreeLine, an information storage program
# Copyright (C) 2003, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#*****************************************************************************

from treedoc import TreeDoc
import docparam
from qt import *


class DataEditLine(QMultiLineEdit):
    """Line editor within data edit view"""
    tagMenuEntries = [('&Bold', '<b>', '</b>'), ('&Italics', '<i>', '</i>'), \
                      ('&Underline', '<u>', '</u>'), \
                      ('&Size...', '<font size="%+d">', '</font>'), \
                      ('&Color...', '<font color="%s">', '</font>')]
    tagMenuFirstId = 100
    def __init__(self, key, item, mainWin, labelRef, stdWidth, \
                 parent=None, name=None):
        QMultiLineEdit.__init__(self, parent, name)
        self.key = key
        self.item = item
        self.mainWin = mainWin
        self.labelRef = labelRef
        self.labelFont = labelRef.font()
        self.labelBoldFont = labelRef.font()
        self.labelBoldFont.setBold(1)
        self.stdWidth = stdWidth
        self.setMinimumWidth(stdWidth)
        self.setWordWrap(QMultiLineEdit.WidgetWidth)
        self.undoAvail = 0
        self.redoAvail = 0
        self.format = item.nodeFormat.findField(key)
        text = item.data.get(key, '')
        editText = self.format.editText(text, self.mainWin.editFormats)
        if not editText and text:
            editText = text
            self.labelRef.setFont(self.labelBoldFont)
            self.labelRef.update()
        self.setText(editText)
        maxNumLines = mainWin.option.intData('MaxEditLines', 1, \
                                             mainWin.__class__.maxNumLines)
        if self.format.typeName == 'LongText':
            lines = maxNumLines
        else:
            lines = min(maxNumLines, self.numLines())
        self.setFixedVisibleLines(lines + \
                                  (self.maxLineWidth() > stdWidth and 2 or 0))
        self.connect(self, SIGNAL('textChanged()'), self.readChange)
        self.connect(self, SIGNAL('undoAvailable(bool)'), self.setUndoAvail)
        self.connect(self, SIGNAL('redoAvailable(bool)'), self.setRedoAvail)

    def readChange(self):
        """Update variable from edit contents"""
        # text = u' '.join(unicode(self.text()).split())
        text = unicode(self.text()).strip()
        editText = self.format.editText(self.item.data.get(self.key, ''), \
                                        self.mainWin.editFormats)
        if not editText:
            editText = self.item.data.get(self.key, '')
        if text != editText:
            docparam.undoStore.addDataUndo(self.item, \
                                           self.mainWin.doc.selectedNodes, 1)
            newText = self.format.extractEditValue(text, \
                                                   self.mainWin.editFormats)
            if not newText and text:   # text not valid
                self.labelRef.setFont(self.labelBoldFont)
                self.item.data[self.key] = text
            else:                      # text valid
                self.labelRef.setFont(self.labelFont)
                self.item.data[self.key] = newText
            self.labelRef.update()
            docparam.modified = 1
            self.emit(PYSIGNAL('entryChanged'), ())

    def pasteText(self, text):
        """Paste text given in param"""
        self.insert(text)
        self.readChange()

    def copyAvail(self):
        """Return 1 if there is selected text"""
        return self.hasMarkedText()

    def sizeHint(self):
        """Set prefered size"""
        return QSize(self.stdWidth, QMultiLineEdit.sizeHint(self).height())

    def setUndoAvail(self, avail):
        """Set undo availability based on signal"""
        self.undoAvail = avail

    def setRedoAvail(self, avail):
        """Set redo availability based on signal"""
        self.redoAvail = avail

    def tagSubMenu(self):
        """Return menu for html tag additions"""
        menu = QPopupMenu(self)
        index = 0
        for text, open, close in DataEditLine.tagMenuEntries:
            menu.insertItem(text, DataEditLine.tagMenuFirstId + index)
            menu.setItemEnabled(DataEditLine.tagMenuFirstId + index, \
                                self.hasMarkedText())
            index += 1
        self.connect(menu, SIGNAL('activated(int)'), self.addTag)
        return menu

    def addTag(self, num):
        """Add HTML tag based on popup menu"""
        label, openTag, closeTag = DataEditLine.\
                              tagMenuEntries[num - DataEditLine.tagMenuFirstId]
        text = unicode(self.markedText())
        if label == '&Size...':
            num, ok = QInputDialog.getInteger('Font Size', \
                                              'Enter size factor (-6 to +6)', \
                                              1, -6, 6, 1, self)
            if not ok or num == 0:
                return
            openTag = openTag % num
        elif label == '&Color...':
            color = QColorDialog.getColor(QColor(), self)
            if not color.isValid():
                return
            openTag = openTag % color.name()
        self.insert('%s%s%s' % (openTag, text, closeTag))
        self.readChange()

    def mousePressEvent(self, event):
        """Mouse press down event for custom popup menu"""
        if event.button() == Qt.RightButton:
            popup = QPopupMenu(self)
            popup.insertItem('&Add Font Tags', self.tagSubMenu())
            popup.insertSeparator()
            id = popup.insertItem('&Undo', self, SLOT('undo()'), \
                                  Qt.CTRL+Qt.Key_Z)
            popup.setItemEnabled(id, self.undoAvail)
            id = popup.insertItem('&Redo', self, SLOT('redo()'), \
                                  Qt.CTRL+Qt.Key_Y)
            popup.setItemEnabled(id, self.redoAvail)
            popup.insertSeparator()
            id = popup.insertItem('C&ut', self, SLOT('cut()'), \
                                  Qt.CTRL+Qt.Key_X)
            popup.setItemEnabled(id, self.copyAvail())
            id = popup.insertItem('&Copy', self, SLOT('copy()'), \
                                  Qt.CTRL+Qt.Key_C)
            popup.setItemEnabled(id, self.copyAvail())
            id = popup.insertItem('&Paste', self, SLOT('paste()'), \
                                  Qt.CTRL+Qt.Key_V)
            try:
                text = unicode(QApplication.clipboard().text())
            except UnicodeError:
                text = ''
            popup.setItemEnabled(id, len(text))
            id = popup.insertItem('C&lear', self, SLOT('clear()'))
            text = unicode(self.text())
            popup.setItemEnabled(id, len(text))
            popup.insertSeparator()
            id = popup.insertItem('&Select All', self, SLOT('selectAll()'))
            popup.setItemEnabled(id, len(text))
            popup.popup(event.globalPos())
        else:
            QMultiLineEdit.mousePressEvent(self, event)

    def keyPressEvent(self, event):
        """Bind keys to functions"""
        if event.key() == Qt.Key_V and event.state() == Qt.ControlButton:
            self.mainWin.editPaste()  # override normal paste
            event.accept()
        elif event.key() == Qt.Key_Tab:
            self.focusNextPrevChild(1)
            event.accept()
        else:
            QMultiLineEdit.keyPressEvent(self, event)


class TitleListView(QMultiLineEdit):
    """Right pane list edit view, children of currently selected parent"""
    def __init__(self, mainWin, showChildren=1, parent=None, name=None):
        QMultiLineEdit.__init__(self, parent, name)
        self.mainWin = mainWin
        self.showChildren = showChildren
        self.connect(self, SIGNAL('textChanged()'), self.readChange)

    def updateView(self):
        """Replace contents with selected item child list"""
        self.blockSignals(1)
        self.clear()
        item = self.mainWin.treeView.currentTreeItem()
        if item:
            if not self.showChildren:
                self.setText(item.title())
            elif self.mainWin.showItemChildren:
                self.setText(u'\n'.join(item.childText()))
            else:
                self.setText(u'')
        self.blockSignals(0)

    def readChange(self):
        """Update doc from edit view contents"""
        item = self.mainWin.treeView.currentTreeItem()
        if item:
            if self.showChildren:
                item.editChildList(unicode(self.text()).split('\n'))
            else:
                docparam.undoStore.addDataUndo(item, \
                                             self.mainWin.doc.selectedNodes, 1)
                if not item.setTitle(unicode(self.text())):
                    docparam.undoStore.removeLastUndo()
                    return
            self.mainWin.treeView.updateTree(item)
            self.mainWin.updateCmdAvail()

    def copyAvail(self):
        """Return 1 if there is selected text"""
        return self.hasMarkedText()

    def pasteText(self, text):
        """Paste text given in param"""
        self.insert(text)
        self.emit(SIGNAL('textChanged()'), ())

    def scrollPage(self, numPages=1):
        """Scrolls down by numPages (negative for up)"""
        if numPages > 0:
            for num in range(numPages):
                self.pageDown()
        else:
            for num in range(-numPages):
                self.pageUp()

    def dropEvent(self, event):
        """Force update after text drop"""
        QMultiLineEdit.dropEvent(self, event)
        self.emit(SIGNAL('textChanged()'), ())

    def keyPressEvent(self, event):
        """Bind keys to functions"""
        if event.key() == Qt.Key_V and event.state() == Qt.ControlButton:
            self.mainWin.editPaste()  # override normal paste
            event.accept()
        elif event.key() == Qt.Key_Tab:
            self.focusNextPrevChild(1)
            event.accept()
        else:
            QMultiLineEdit.keyPressEvent(self, event)


class FormatEdit(QMultiLineEdit):
    """Editor signals cursor movement"""
    def __init__(self, parent=None, name=None):
        QMultiLineEdit.__init__(self, parent, name)
        self.setMinimumSize(QMultiLineEdit.minimumSize(self).width(), \
                            self.fontMetrics().lineSpacing() * 4)

    def event(self, event):
        """Signal cursor movement if text didn't also change"""
        pos = self.getCursorPosition()
        self.setEdited(0)
        result = QMultiLineEdit.event(self, event)
        if not self.edited() and pos != self.getCursorPosition():
            self.emit(PYSIGNAL('cursorMove'), ())
        return result

    def keyPressEvent(self, event):
        """Ignore tab key to allow focus change"""
        if event.key() == Qt.Key_Tab:
            self.focusNextPrevChild(1)
            event.accept()
        else:
            QMultiLineEdit.keyPressEvent(self, event)


class SpellContextEdit(QMultiLineEdit):
    """Editor for spell check word context"""
    def __init__(self, parent=None, name=None):
        QMultiLineEdit.__init__(self, parent, name)
        self.setWordWrap(QMultiLineEdit.WidgetWidth)

    def sizeHint(self):
        """Set prefered size"""
        fontHeight = QFontMetrics(self.font()).lineSpacing()
        return QSize(QMultiLineEdit.sizeHint(self).width(), fontHeight * 3)

    def setSelection(self, fromPos, toPos):
        """Select given range"""
        fromLine, fromCol = self.posToLinePos(fromPos)
        self.setCursorPosition(fromLine, fromCol, 0)
        toLine, toCol = self.posToLinePos(toPos)
        self.setCursorPosition(toLine, toCol, 1)

    def posToLinePos(self, pos):
        """Return line number and linePos tuple for absolute position pos"""
        beginPos = 0
        for lineNum in range(self.numLines()):
            lineLen = len(unicode(self.textLine(lineNum)))
            if pos <= beginPos + lineLen:
                return (lineNum, pos - beginPos)
            beginPos += lineLen
        return (-1, 0)  # error value
