#!/usr/bin/env python

#****************************************************************************
# textedit3.py, provides classes for the qt3 text editors
#
# TreeLine, an information storage program
# Copyright (C) 2003, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#*****************************************************************************

from treedoc import TreeDoc
import docparam
from qt import *


class DataEditLine(QTextEdit):
    """Line editor within data edit view"""
    tagMenuEntries = [('&Bold', '<b>', '</b>'), ('&Italics', '<i>', '</i>'), \
                      ('&Underline', '<u>', '</u>'), \
                      ('&Size...', '<font size="%s">', '</font>'), \
                      ('&Color...', '<font color="%s">', '</font>')]
    tagMenuFirstId = 100
    def __init__(self, key, item, mainWin, labelRef, stdWidth, \
                 parent=None, name=None):
        QTextEdit.__init__(self, parent, name)
        self.key = key
        self.item = item
        self.mainWin = mainWin
        self.labelRef = labelRef
        self.labelFont = labelRef.font()
        self.labelBoldFont = labelRef.font()
        self.labelBoldFont.setBold(1)
        self.setTextFormat(Qt.PlainText)
        self.setWordWrap(QTextEdit.FixedPixelWidth)
        self.setWrapColumnOrWidth(stdWidth - 5)
        # self.setTabChangesFocus(1)  # not in Qt 3.0.x
        self.format = item.nodeFormat.findField(key)
        text = item.data.get(key, '')
        editText = self.format.editText(text, self.mainWin.editFormats)
        if not editText and text:
            editText = text
            self.labelRef.setFont(self.labelBoldFont)
            self.labelRef.update()
        self.setText(editText)
        maxNumLines = mainWin.option.intData('MaxEditLines', 1, \
                                             mainWin.__class__.maxNumLines)
        if self.format.typeName == 'LongText':
            numLines = maxNumLines
        else:
            numLines = min(max(1, self.lines()), maxNumLines)
        lineHt = QFontMetrics(self.font()).lineSpacing()
        self.setFixedSize(stdWidth, numLines * lineHt + lineHt / 2)
        self.setWordWrap(QTextEdit.WidgetWidth)
        self.connect(self, SIGNAL('textChanged()'), self.readChange)

    def readChange(self):
        """Update variable from edit contents"""
        # text = u' '.join(unicode(self.text()).split())
        text = unicode(self.text()).strip()
        editText = self.format.editText(self.item.data.get(self.key, ''), \
                                        self.mainWin.editFormats)
        if not editText:
            editText = self.item.data.get(self.key, '')
        if text != editText:
            docparam.undoStore.addDataUndo(self.item, \
                                           self.mainWin.doc.selectedNodes, 1)
            newText = self.format.extractEditValue(text, \
                                                   self.mainWin.editFormats)
            if not newText and text:   # text not valid
                self.labelRef.setFont(self.labelBoldFont)
                self.item.data[self.key] = text
            else:                      # text valid
                self.labelRef.setFont(self.labelFont)
                self.item.data[self.key] = newText
            self.labelRef.update()
            docparam.modified = 1
            self.emit(PYSIGNAL('entryChanged'), ())

    def pasteText(self, text):
        """Paste text given in param"""
        self.insert(text)
        self.readChange()

    def copyAvail(self):
        """Return 1 if there is selected text"""
        return self.hasSelectedText()

    # removed - interfered with middle-button paste
    # def paste(self):
        # """Override normal paste"""
        # self.mainWin.editPaste()

    def keyPressEvent(self, event):
        """Bind keys to functions"""
        if event.key() == Qt.Key_V and event.state() == Qt.ControlButton:
            self.mainWin.editPaste()  # override normal paste
        elif event.key() == Qt.Key_Tab:
            self.parent().focusNextPrevChild(event.state() != Qt.ShiftButton)
        else:
            QTextEdit.keyPressEvent(self, event)

    def tagSubMenu(self):
        """Return menu for html tag additions"""
        menu = QPopupMenu(self)
        index = 0
        for text, open, close in DataEditLine.tagMenuEntries:
            menu.insertItem(text, DataEditLine.tagMenuFirstId + index)
            menu.setItemEnabled(DataEditLine.tagMenuFirstId + index, \
                                self.hasSelectedText())
            index += 1
        self.connect(menu, SIGNAL('activated(int)'), self.addTag)
        return menu

    def addTag(self, num):
        """Add HTML tag based on popup menu"""
        label, openTag, closeTag = DataEditLine.\
                              tagMenuEntries[num - DataEditLine.tagMenuFirstId]
        text = unicode(self.selectedText())
        if label == '&Size...':
            num, ok = QInputDialog.getInteger('Font Size', \
                                              'Enter size factor (-6 to +6)', \
                                              1, -6, 6, 1, self)
            if not ok or num == 0:
                return
            openTag = openTag % num
        elif label == '&Color...':
            color = QColorDialog.getColor(QColor(), self)
            if not color.isValid():
                return
            openTag = openTag % color.name()
        self.insert('%s%s%s' % (openTag, text, closeTag))

    def createPopupMenu(self, pos):
        """Returns popup menu, overridden to change paste command and add
           font tag menu item"""
        popup = QTextEdit.createPopupMenu(self, pos)
        popup.removeItemAt(5)
        popup.insertItem('&Paste\tCtrl+V', self.mainWin.editPaste, 0, -1, 5)
        try:
            text = unicode(QApplication.clipboard().text())
        except UnicodeError:
            text = ''
        popup.setItemEnabled(popup.idAt(5), len(text))
        popup.insertSeparator(0)
        popup.insertItem('&Add Font Tags', self.tagSubMenu(), -1, 0)
        return popup



class TitleListView(QTextEdit):
    """Right pane list edit view, children of currently selected parent"""
    def __init__(self, mainWin, showChildren=1, parent=None, name=None):
        QTextEdit.__init__(self, parent, name)
        self.mainWin = mainWin
        self.showChildren = showChildren
        self.oldItem = None
        self.setTextFormat(Qt.PlainText)
        self.setWordWrap(QTextEdit.NoWrap)
        # self.setTabChangesFocus(1)  # not in Qt 3.0.x
        self.connect(self, SIGNAL('textChanged()'), self.readChange)

    def updateView(self):
        """Replace contents with selected item child list"""
        self.blockSignals(1)
        # self.clear()
        item = self.mainWin.treeView.currentTreeItem()
        if item:
            if not self.showChildren:
                self.setText(item.title())
            elif self.mainWin.showItemChildren:
                self.setText(u'\n'.join(item.childText()))
            else:
                self.setText(u'')
            if item is not self.oldItem:
                self.ensureVisible(0, 0)  # reset scroll if root changed
            self.oldItem = item
        else:
            self.clear()
        self.blockSignals(0)

    def readChange(self):
        """Update doc from edit view contents"""
        item = self.mainWin.treeView.currentTreeItem()
        if item:
            if self.showChildren:
                item.editChildList(unicode(self.text()).split('\n'))
            else:
                docparam.undoStore.addDataUndo(item, \
                                             self.mainWin.doc.selectedNodes, 1)
                if not item.setTitle(unicode(self.text())):
                    docparam.undoStore.removeLastUndo()
                    return
            self.mainWin.treeView.updateTree(item)
            self.mainWin.updateCmdAvail()

    def copyAvail(self):
        """Return 1 if there is selected text"""
        return self.hasSelectedText()

    def pasteText(self, text):
        """Paste text given in param"""
        self.insert(text)
        self.emit(SIGNAL('textChanged()'), ())

    def scrollPage(self, numPages=1):
        """Scrolls down by numPages (negative for up)"""
        self.scrollBy(0, numPages * self.visibleHeight())

    def keyPressEvent(self, event):
        """Bind keys to functions"""
        if event.key() == Qt.Key_V and event.state() == Qt.ControlButton:
            self.mainWin.editPaste()  # override normal paste
        elif event.key() == Qt.Key_Tab:
            self.parent().focusNextPrevChild(event.state() != Qt.ShiftButton)
        else:
            QTextEdit.keyPressEvent(self, event)

    def createPopupMenu(self, pos):
        """Returns popup menu, overridden to change paste command"""
        popup = QTextEdit.createPopupMenu(self, pos)
        popup.removeItemAt(5)
        popup.insertItem('&Paste\tCtrl+V', self.mainWin.editPaste, 0, -1, 5)
        try:
            text = unicode(QApplication.clipboard().text())
        except UnicodeError:
            text = ''
        popup.setItemEnabled(popup.idAt(5), len(text))
        return popup


class FormatEdit(QTextEdit):
    """Editor signals cursor movement"""
    def __init__(self, parent=None, name=None):
        QTextEdit.__init__(self, parent, name)
        self.setTextFormat(Qt.PlainText)
        self.setWordWrap(QTextEdit.NoWrap)
        # self.setTabChangesFocus(1)  # not in Qt 3.0.x
        self.setMinimumSize(QTextEdit.minimumSize(self).width(), \
                            self.fontMetrics().lineSpacing() * 4)
        self.connect(self, SIGNAL('cursorPositionChanged(int, int)'), \
                     self.cursorChg)

    def cursorChg(self):
        """Signal cursor movement"""
        self.emit(PYSIGNAL('cursorMove'), ())

    def numLines(self):
        """Wrap multilinedit command"""
        return self.paragraphs()

    def textLine(self, lineNum):
        """Wrap multilinedit command"""
        return self.text(lineNum)

    def deselect(self):
        """Wrap multilinedit command"""
        self.removeSelection()

    def keyPressEvent(self, event):
        """Bind keys to functions"""
        if event.key() == Qt.Key_Tab:
            self.parent().focusNextPrevChild(event.state() != Qt.ShiftButton)
        else:
            QTextEdit.keyPressEvent(self, event)


class SpellContextEdit(QTextEdit):
    """Editor for spell check word context"""
    def __init__(self, parent=None, name=None):
        QTextEdit.__init__(self, parent, name)
        self.setTextFormat(Qt.PlainText)

    def sizeHint(self):
        """Set prefered size"""
        fontHeight = QFontMetrics(self.currentFont()).lineSpacing()
        return QSize(QTextEdit.sizeHint(self).width(), fontHeight * 3)

    def setSelection(self, fromPos, toPos):
        """Select given range in first paragraph"""
        QTextEdit.setSelection(self, 0, fromPos, 0, toPos)

    def keyPressEvent(self, event):
        """Bind keys to functions"""
        if event.key() == Qt.Key_Tab:
            self.parent().focusNextPrevChild(event.state() != Qt.ShiftButton)
        else:
            QTextEdit.keyPressEvent(self, event)
