#!/usr/bin/env python

#****************************************************************************
# treedialogs.py, provides many dialog interfaces
#
# TreeLine, an information storage program
# Copyright (C) 2004, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#*****************************************************************************

import sys, os.path, copy, re
from treedoc import TreeDoc
from treeformat import NodeFormat
from fieldformat import TextFormat, LongTextFormat, NumberFormat, \
                        ChoiceFormat, CombinationFormat, DateFormat, \
                        TimeFormat, BooleanFormat, URLFormat, PathFormat, \
                        EmailFormat, PictureFormat
from helpview import HelpView
import docparam
import treemainwin
from qt import *
if qVersion()[0] >= '3':
    from textedit3 import FormatEdit, SpellContextEdit
else:
    from textedit2 import FormatEdit, SpellContextEdit
from xpm import treeline_xpm

stdFlags = Qt.WStyle_Customize | Qt.WStyle_NormalBorder | Qt.WStyle_Title | \
           Qt.WStyle_SysMenu

class ConfigDlg(QDialog):
    """Dialog for configuring data types"""
    def __init__(self, mainWin, initType, parent=None, name=None, \
                 modal=0, flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.mainWin = mainWin

        self.formatList = copy.deepcopy(docparam.nodeFormats)
        self.fileInfoFormat = self.formatList.findFormat(docparam.fileInfoItem.\
                                                         nodeFormat.name)
        if not self.fileInfoFormat:
            self.fileInfoFormat = copy.deepcopy(docparam.fileInfoItem.\
                                                         nodeFormat)
        self.currentType = None
        self.fieldRenameDict = {}
        self.typeRenameDict = {}
        self.parentLevel = 0
        self.otherType = None
        self.setIcon(QPixmap(treeline_xpm))
        if sys.platform == 'win32':
            self.setCaption('Configure Data Types  (PyQt)')
        else:
            self.setCaption('Configure Data Types')

        topLayout = QVBoxLayout(self, 7)
        selectBox = QHGroupBox('&Data Type', self)
        topLayout.addWidget(selectBox)
        self.selectCombo = QComboBox(0, selectBox)
        modSelect = QPushButton('&Modify List', selectBox)
        modSelect.setMaximumWidth(modSelect.sizeHint().width())
        self.connect(modSelect, SIGNAL('clicked()'), self.modTypeList)
        self.connect(self.selectCombo, SIGNAL('activated(const QString&)'), \
                     self.changeType)

        midLayout = QGridLayout(topLayout, 5, 3)
        fieldBox = QGroupBox('Fields', self)
        midLayout.addMultiCellWidget(fieldBox, 0, 1, 0, 0)
        fieldLayout = QGridLayout(fieldBox, 8, 2, 10, 2)
        fieldLayout.addRowSpacing(0, self.fontMetrics().lineSpacing())
        fieldLayout.setRowStretch(0, 0)
        self.fieldListView = QListView(fieldBox)
        self.fieldListView.addColumn('Name')
        self.fieldListView.addColumn('Type')
        self.fieldListView.setSorting(-1)
        fieldLayout.addMultiCellWidget(self.fieldListView, 1, 7, 1, 1)
        fieldLayout.setRowStretch(1, 1)
        fieldLayout.setColStretch(1, 0)
        self.upFieldButton = QPushButton('Move &Up', fieldBox)
        fieldLayout.addWidget(self.upFieldButton, 1, 0)
        self.connect(self.upFieldButton, SIGNAL('clicked()'), self.fieldUp)
        self.downFieldButton = QPushButton('Move Do&wn', fieldBox)
        fieldLayout.addWidget(self.downFieldButton, 2, 0)
        self.connect(self.downFieldButton, SIGNAL('clicked()'), self.fieldDown)
        self.newFieldButton = QPushButton('&New Field...', fieldBox)
        fieldLayout.addWidget(self.newFieldButton, 3, 0)
        self.connect(self.newFieldButton, SIGNAL('clicked()'), self.newField)
        self.fieldTypeButton = QPushButton('F&ield Type...', fieldBox)
        fieldLayout.addWidget(self.fieldTypeButton, 4, 0)
        self.connect(self.fieldTypeButton, SIGNAL('clicked()'), self.fieldType)
        self.renameFieldButton = QPushButton('&Rename Field...', fieldBox)
        fieldLayout.addWidget(self.renameFieldButton, 5, 0)
        self.connect(self.renameFieldButton, SIGNAL('clicked()'), \
                     self.renameField)
        self.delFieldButton = QPushButton('D&elete Field', fieldBox)
        fieldLayout.addWidget(self.delFieldButton, 6, 0)
        self.connect(self.delFieldButton, SIGNAL('clicked()'), self.delField)
        otherfieldButton = QPushButton('&Other Fields...', fieldBox)
        fieldLayout.addWidget(otherfieldButton, 7, 0)
        self.connect(otherfieldButton, SIGNAL('clicked()'), self.otherfields)

        xferTitleLayout = QVBoxLayout(5)
        midLayout.addLayout(xferTitleLayout, 0, 1)
        self.toTitleButton = QPushButton('>>', self)
        self.toTitleButton.setMaximumWidth(self.toTitleButton.height())
        xferTitleLayout.addWidget(self.toTitleButton)
        self.connect(self.toTitleButton, SIGNAL('clicked()'), \
                     self.fieldToTitle)
        self.delTitleButton = QPushButton('<<', self)
        self.delTitleButton.setMaximumWidth(self.delTitleButton.height())
        xferTitleLayout.addWidget(self.delTitleButton)
        self.connect(self.delTitleButton, SIGNAL('clicked()'), \
                     self.delTitleField)

        titleBox = QGroupBox('&Title Format', self)
        midLayout.addWidget(titleBox, 0, 2)
        titleLayout = QVBoxLayout(titleBox, 10, 5)
        titleLayout.addSpacing(self.fontMetrics().lineSpacing())
        self.titleEdit = TitleEdit(titleBox)
        titleLayout.addWidget(self.titleEdit)
        self.connect(self.titleEdit, SIGNAL('textChanged(const QString&)'), \
                     self.changeTitle)
        self.connect(self.titleEdit, PYSIGNAL('cursorMove'), \
                     self.setButtonAvail)

        xferOutputLayout = QVBoxLayout(5)
        midLayout.addLayout(xferOutputLayout, 1, 1)
        self.toOutputButton = QPushButton('>>', self)
        self.toOutputButton.setMaximumWidth(self.toOutputButton.height())
        xferOutputLayout.addWidget(self.toOutputButton)
        self.connect(self.toOutputButton, SIGNAL('clicked()'), \
                     self.fieldToOutput)
        self.delOutputButton = QPushButton('<<', self)
        self.delOutputButton.setMaximumWidth(self.delOutputButton.height())
        xferOutputLayout.addWidget(self.delOutputButton)
        self.connect(self.delOutputButton, SIGNAL('clicked()'), \
                     self.delOutputField)

        outputBox = QGroupBox('Output &Format', self)
        midLayout.addWidget(outputBox, 1, 2)
        outputLayout = QVBoxLayout(outputBox, 10, 5)
        outputLayout.addSpacing(self.fontMetrics().lineSpacing())
        self.outputEdit = FormatEdit(outputBox)
        outputLayout.addWidget(self.outputEdit)
        self.connect(self.outputEdit, SIGNAL('textChanged()'), \
                     self.changeOutput)
        self.connect(self.outputEdit, PYSIGNAL('cursorMove'), \
                     self.setButtonAvail)

        prefixBox = QHGroupBox('Siblin&g Prefix Tags', self)
        midLayout.addWidget(prefixBox, 2, 0)
        self.prefixEdit = QLineEdit('', prefixBox)
        self.connect(self.prefixEdit, SIGNAL('textChanged(const QString&)'), \
                     self.changeSibText)

        suffixBox = QHGroupBox('Sibling Suffi&x Tags', self)
        midLayout.addMultiCellWidget(suffixBox, 3, 4, 0, 0)
        self.suffixEdit = QLineEdit('', suffixBox)
        self.connect(self.suffixEdit, SIGNAL('textChanged(const QString&)'), \
                     self.changeSibText)

        childBox = QHGroupBox('Default C&hild Type', self)
        midLayout.addWidget(childBox, 2, 2)
        self.childCombo = QComboBox(0, childBox)
        self.loadTypeNames(initType)
        self.connect(self.childCombo, SIGNAL('activated(const QString&)'), \
                     self.changeChildType)
        self.loadFields()
        self.connect(self.fieldListView, SIGNAL('selectionChanged()'), \
                     self.setButtonAvail)
        self.loadText()
        self.setButtonAvail()

        ctrlLayout = QHBoxLayout()
        midLayout.addLayout(ctrlLayout, 4, 2)
        ctrlLayout.insertStretch(0)
        okButton = QPushButton('&OK', self)
        ctrlLayout.addWidget(okButton)
        self.connect(okButton, SIGNAL('clicked()'), self, SLOT('accept()'))
        cancelButton = QPushButton('&Cancel', self)
        ctrlLayout.addWidget(cancelButton)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))
        self.selectCombo.setFocus()

    def loadTypeNames(self, initName=''):
        """Load combo box with data type names"""
        names = self.formatList.nameList()
        names.sort()
        self.selectCombo.clear()
        self.childCombo.clear()
        self.childCombo.insertItem('[None]')
        for name in names:
            self.selectCombo.insertItem(name)
            self.childCombo.insertItem(name)
        if initName:
            self.selectCombo.setCurrentItem(names.index(initName))
        self.currentType = self.formatList.\
                           findFormat(str(self.selectCombo.currentText()))
        self.updateChildType(names)

    def updateChildType(self, typeNames):
        """Set current type name in child type combo"""
        try:
            childItem = typeNames.index(self.currentType.childType) + 1
        except ValueError:
            childItem = 0
        self.childCombo.setCurrentItem(childItem)

    def changeType(self, text):
        """Change type name based on combo box signal"""
        self.currentType = self.formatList.findFormat(str(text))
        self.loadFields()
        self.loadText()
        names = self.formatList.nameList()
        names.sort()
        self.updateChildType(names)
        self.setButtonAvail()

    def changeChildType(self, text):
        """Change default child type based on combo box signal"""
        typeName = str(text)
        self.currentType.childType = typeName != '[None]' and typeName or ''

    def setButtonAvail(self):
        """Update button availability"""
        fieldList = self.currentType.fieldList
        num = self.selectedFieldNum()
        titlePos = self.titleFieldPos()
        outputPos = self.outputFieldPos()
        self.upFieldButton.setEnabled(not self.otherType and num > 0)
        self.downFieldButton.setEnabled(not self.otherType and \
                                        0 <= num < len(fieldList) - 1)
        self.newFieldButton.setEnabled(not self.otherType)
        self.fieldTypeButton.setEnabled(not self.otherType or \
                                        self.parentLevel == 0)
        self.renameFieldButton.setEnabled(not self.otherType)
        self.delFieldButton.setEnabled(not self.otherType and \
                                       num >= 0 and len(fieldList) > 1)
        self.toTitleButton.setEnabled(num >= 0 and not titlePos)
        self.delTitleButton.setEnabled(titlePos and 1 or 0)
        self.toOutputButton.setEnabled(num >= 0 and not outputPos)
        self.delOutputButton.setEnabled(outputPos and 1 or 0)

    def loadFields(self, selNum=0):
        """Load list with field names"""
        self.fieldListView.clear()
        items = []
        if self.otherType:
            fields = [field for field in self.otherType.fieldList if \
                      field.showInDialog]
        else:
            fields = self.currentType.fieldList[:]
        fields.reverse()
        for field in fields:
            items.insert(0, QListViewItem(self.fieldListView, field.name, \
                                          field.typeName))
        self.fieldListView.setSelected(items[selNum], 1)

    def selectedFieldName(self):
        """Return currently selected field name"""
        item = self.fieldListView.selectedItem()
        if not item:
            # select last item for unselecting click
            item = self.fieldListView.lastItem()
            self.fieldListView.setSelected(item, 1)
        return str(item.text(0))

    def selectedFieldNum(self):
        """Return currently selected field number"""
        if self.otherType:
            fields = [field.name for field in self.otherType.fieldList if \
                      field.showInDialog]
        else:
            fields = self.currentType.fieldNames()
        return fields.index(self.selectedFieldName())

    def fieldUp(self):
        """Move field upward in list"""
        fieldList = self.currentType.fieldList
        num = self.selectedFieldNum()
        if num > 0:
            fieldList[num-1], fieldList[num] = fieldList[num], fieldList[num-1]
            self.loadFields(num-1)

    def fieldDown(self):
        """Move field downward in list"""
        fieldList = self.currentType.fieldList
        num = self.selectedFieldNum()
        if 0 <= num < len(fieldList) - 1:
            fieldList[num], fieldList[num+1] = fieldList[num+1], fieldList[num]
            self.loadFields(num+1)

    def newField(self):
        """Add a new field to the list"""
        fieldNames = self.currentType.fieldNames()
        dlg = FieldEntry('Add Field', 'Enter new field name:', '', \
                         fieldNames, self, None, 1)
        if dlg.exec_loop() == QDialog.Accepted:
            htmlAttrs = self.mainWin.option.boolData('HtmlNewFields') and \
                        {'html': 'y'} or {}
            self.currentType.addNewField(dlg.text, htmlAttrs)
            self.loadFields(len(fieldNames))

    def fieldType(self):
        """Display the dialog for changing the field type settings"""
        fieldNum = self.selectedFieldNum()
        field = self.currentType.fieldList[fieldNum]
        allowTypeChg = 1
        if self.otherType:  # file info field
            fields = [field for field in self.otherType.fieldList if \
                      field.showInDialog]
            field = fields[fieldNum]
            allowTypeChg = 0
        dlg = FieldFormatDialog(field, allowTypeChg, self, None, 1)
        if dlg.exec_loop() == QDialog.Accepted:
            # field's class is changed to avoid having to change the ref's
            field.__class__ = eval(dlg.currentType + 'Format')
            field.format = dlg.format
            field.prefix = dlg.prefix
            field.suffix = dlg.suffix
            field.html = dlg.html
            field.initFormat()
            if self.otherType:  # file info field
                if not self.formatList.findFormat(docparam.fileInfoItem.\
                                                           nodeFormat.name):
                    self.formatList.append(self.fileInfoFormat)
            self.loadFields(fieldNum)

    def renameField(self):
        """Change the field name and save for item changes"""
        fieldNames = self.currentType.fieldNames()
        num = self.selectedFieldNum()
        if num >= 0:
            dlg = FieldEntry('Rename Field', 'Rename from "%s" to:' \
                             % fieldNames[num], fieldNames[num], fieldNames, \
                             self, None, 1)
            if dlg.exec_loop() == QDialog.Accepted:
                self.loadText()
                self.currentType.fieldList[num].name = dlg.text
                self.fieldRenameDict[(self.currentType.name, fieldNames[num])] \
                     = dlg.text
                self.loadFields(num)

    def delField(self):
        """Remove a field from the list"""
        fieldList = self.currentType.fieldList
        num = self.selectedFieldNum()
        if num >= 0:
            if self.currentType.removeField(fieldList[num]):
                self.loadText()
            del fieldList[num]
            if num:
                num -= 1
            self.loadFields(num)

    def otherfields(self):
        """Display dialog to allow setting of ancestor & file data field
           choices"""
        dlg = OtherFieldDlg(self.parentLevel, self.otherType, self.formatList, \
                            self.fileInfoFormat, self, None, 1)
        if dlg.exec_loop() == QDialog.Accepted:
            self.parentLevel = dlg.parentLevel
            self.otherType = dlg.otherType
        else:
            self.parentLevel = 0
            self.otherType = None
        self.loadFields()
        self.setButtonAvail()

    def loadText(self):
        """Load text into title and output format editors and
           sibling prefix/postfix"""
        lines = self.currentType.getLines()
        self.titleEdit.blockSignals(1)
        self.titleEdit.setText(lines[0])
        self.titleEdit.blockSignals(0)
        self.outputEdit.blockSignals(1)
        self.outputEdit.setText(u'\n'.join(lines[1:]))
        if lines[1:]:
            self.outputEdit.setCursorPosition(len(lines) - 2, len(lines[-1]))
        self.outputEdit.blockSignals(0)
        self.prefixEdit.blockSignals(1)
        self.prefixEdit.setText(self.currentType.sibPrefix)
        self.prefixEdit.blockSignals(0)
        self.suffixEdit.blockSignals(1)
        self.suffixEdit.setText(self.currentType.sibSuffix)
        self.suffixEdit.blockSignals(0)

    def changeTitle(self):
        """Update title format lines based on editor change"""
        self.currentType.changeTitleLine(unicode(self.titleEdit.text()))
        self.setButtonAvail()

    def changeOutput(self):
        """Update output format lines based on editor change"""
        self.currentType.lineList = self.currentType.lineList[:1]
        for lineNum in range(self.outputEdit.numLines()):
            self.currentType.addLine(unicode(self.outputEdit.textLine(lineNum)))
        self.setButtonAvail()

    def changeSibText(self):
        """Update sibling prefix and suffix text based on change"""
        self.currentType.sibPrefix = unicode(self.prefixEdit.text())
        self.currentType.sibSuffix = unicode(self.suffixEdit.text())

    def fieldSepName(self):
        """Return current field name with proper seperators"""
        name = self.selectedFieldName()
        if self.parentLevel < 0:
            return u'{*?%s*}' % name
        if not self.parentLevel and self.otherType:
            return u'{*!%s*}' % name
        return u'{*%s%s*}' % (self.parentLevel * '*', name)

    def fieldToTitle(self):
        """Add selected field to cursor pos in editor"""
        self.titleEdit.deselect()
        self.titleEdit.insert(self.fieldSepName())
        self.titleEdit.setFocus()

    def delTitleField(self):
        """Remove field from cursor pos in editor"""
        start, end = self.titleFieldPos()
        line = self.currentType.getLines()[0]
        line = line[:start] + line[end:]
        self.titleEdit.setText(line)
        self.titleEdit.setCursorPosition(start)

    def fieldToOutput(self):
        """Add selected field to cursor pos in editor"""
        name = self.selectedFieldName()
        self.outputEdit.deselect()
        self.outputEdit.insert(self.fieldSepName())
        self.changeOutput()
        self.outputEdit.setFocus()

    def delOutputField(self):
        """Remove field from cursor pos in editor"""
        num, start, end = self.outputFieldPos()
        lines = self.currentType.getLines()[1:]
        lines[num] = lines[num][:start] + lines[num][end:]
        self.outputEdit.setText(u'\n'.join(lines))
        self.outputEdit.setCursorPosition(num, start)

    def titleFieldPos(self):
        """Return tuple of start, end for title field at cursor or None"""
        line = self.currentType.getLines()[0]
        cursorPos = self.titleEdit.cursorPosition()
        return self.currentFieldPos(cursorPos, line)

    def outputFieldPos(self):
        """Return tuple of line num start, end for output field at cursor 
           or None"""
        lines = self.currentType.getLines()[1:]
        cursorPos = self.outputEdit.getCursorPosition()
        if len(lines) > cursorPos[0]:
            pos = self.currentFieldPos(cursorPos[1], lines[cursorPos[0]])
            if pos:
                return (cursorPos[0], pos[0], pos[1])
        return None

    def currentFieldPos(self, cursorPos, textLine):
        """Return tuple of start, end for field at cursorPos or None"""
        pattern = re.compile('{\*(.*?)\*}')
        match = pattern.search(textLine)
        while match:
            if match.start() < cursorPos < match.end():
                return (match.start(), match.end())
            match = pattern.search(textLine, match.end())
        return None

    def modTypeList(self):
        """Show dialog to add/remove from data type list"""
        dlg = ModTypeListDlg(self.mainWin, self.formatList, \
                             self.typeRenameDict, self, None, 1)
        if dlg.exec_loop() == QDialog.Accepted:
            self.formatList = dlg.formatList
            self.typeRenameDict = dlg.typeRenameDict
            self.loadTypeNames(str(dlg.listBox.currentText()))
            self.loadFields()
            self.loadText()
            self.setButtonAvail()


class TitleEdit(QLineEdit):
    """Editor signals cursor movement and focus-in events"""
    def __init__(self, parent=None, name=None):
        QLineEdit.__init__(self, parent, name)

    def event(self, event):
        """Signal cursor movement if text didn't also change"""
        pos = self.cursorPosition()
        self.setEdited(0)
        result = QLineEdit.event(self, event)
        if not self.edited() and pos != self.cursorPosition():
            self.emit(PYSIGNAL('cursorMove'), ())
        return result

    def focusInEvent(self, event):
        """Signal focus in events"""
        self.emit(PYSIGNAL('focusIn'), (self,))
        QLineEdit.focusInEvent(self, event)


class FieldEntry(QDialog):
    """Dialog for alpha-numeric and underscore text entry"""
    def __init__(self, caption, label, dfltText='', badStr=[], parent=None, \
                 name=None, modal=0, flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.text = ''
        self.badStr = badStr
        if sys.platform == 'win32':
            caption += '  (PyQt)'
        self.setCaption(caption)
        self.setIcon(QPixmap(treeline_xpm))
        topLayout = QVBoxLayout(self, 5)
        label = QLabel(label, self)
        topLayout.addWidget(label)
        self.entry = QLineEdit(dfltText, self)
        topLayout.addWidget(self.entry)
        self.entry.setFocus()
        self.connect(self.entry, SIGNAL('returnPressed()'), \
                     self, SLOT('accept()'))
        ctrlLayout = QHBoxLayout(topLayout)
        ctrlLayout.insertStretch(0)
        okButton = QPushButton('&OK', self)
        ctrlLayout.addWidget(okButton)
        self.connect(okButton, SIGNAL('clicked()'), self, SLOT('accept()'))
        cancelButton = QPushButton('&Cancel', self)
        ctrlLayout.addWidget(cancelButton)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))

    def accept(self):
        """Check for acceptable string before closing"""
        try:
            self.text = str(self.entry.text()).strip()
        except UnicodeError:
            self.text = ''
        if not self.text.replace('_', '').isalnum():
            QMessageBox.warning(self, 'TreeLine', \
               'Illegal characters (only alpa-numerics & underscores allowed)')
            return
        if self.text in self.badStr:
            QMessageBox.warning(self, 'TreeLine', \
                                'Entered name was already used')
            return
        return QDialog.accept(self)


class ModTypeListDlg(QDialog):
    """Dialog for adding and deleting data types"""
    def __init__(self, mainWin, formatList, renameDict, parent=None, \
                 name=None, modal=0, flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.mainWin = mainWin
        self.formatList = copy.deepcopy(formatList)
        self.typeRenameDict = copy.deepcopy(renameDict)
        self.setIcon(QPixmap(treeline_xpm))
        if sys.platform == 'win32':
            self.setCaption('Add/Remove Data Types  (PyQt)')
        else:
            self.setCaption('Add/Remove Data Types')

        topLayout = QVBoxLayout(self, 5)
        groupBox = QGroupBox('Data Types', self)
        topLayout.addWidget(groupBox)
        innerLayout = QGridLayout(groupBox, 5, 2, 10, 2)
        innerLayout.addRowSpacing(0, self.fontMetrics().lineSpacing())
        self.listBox = QListBox(groupBox)
        self.loadList()
        innerLayout.addMultiCellWidget(self.listBox, 1, 4, 0, 0)
        newButton = QPushButton('&New Type', groupBox)
        innerLayout.addWidget(newButton, 1, 1)
        self.connect(newButton, SIGNAL('clicked()'), self.newType)
        copyButton = QPushButton('Copy &Type', groupBox)
        innerLayout.addWidget(copyButton, 2, 1)
        self.connect(copyButton, SIGNAL('clicked()'), self.copyType)
        renameButton = QPushButton('&Rename Type', groupBox)
        innerLayout.addWidget(renameButton, 3, 1)
        self.connect(renameButton, SIGNAL('clicked()'), self.renameType)
        delButton = QPushButton('&Delete Type', groupBox)
        innerLayout.addWidget(delButton, 4, 1)
        self.connect(delButton, SIGNAL('clicked()'), self.delType)

        ctrlLayout = QHBoxLayout(topLayout)
        ctrlLayout.insertStretch(0)
        okButton = QPushButton('&OK', self)
        ctrlLayout.addWidget(okButton)
        self.connect(okButton, SIGNAL('clicked()'), self, SLOT('accept()'))
        cancelButton = QPushButton('&Cancel', self)
        ctrlLayout.addWidget(cancelButton)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))
        self.listBox.setFocus()

    def loadList(self, selName=''):
        """Load type names into list"""
        names = self.formatList.nameList()
        names.sort()
        self.listBox.clear()
        for name in names:
            self.listBox.insertItem(name)
        selNum = 0
        if selName:
            selNum = names.index(selName)
        self.listBox.setSelected(selNum, 1)

    def newType(self):
        """Add new type to the list"""
        dlg = FieldEntry('Add Type', 'Enter new type name:', '', \
                         self.formatList.nameList(), self, None, 1)
        if dlg.exec_loop() == QDialog.Accepted:
            self.formatList.append(NodeFormat(dlg.text, 'Name'))
            self.loadList(dlg.text)

    def copyType(self):
        """Copy selected type to a new type"""
        current = str(self.listBox.currentText())
        dlg = FieldEntry('Copy Type', 'Enter new type name:', current, \
                         self.formatList.nameList(), self, None, 1)
        if dlg.exec_loop() == QDialog.Accepted:
            newFormat = copy.deepcopy(self.formatList.findFormat(current))
            newFormat.name = dlg.text
            self.formatList.append(newFormat)
            self.loadList(dlg.text)

    def renameType(self):
        """Rename selected type"""
        current = str(self.listBox.currentText())
        dlg = FieldEntry('Rename Type', 'Rename from "%s" to:' \
                         % current, current, self.formatList.nameList(), \
                         self, None, 1)
        if dlg.exec_loop() == QDialog.Accepted:
            self.formatList.findFormat(current).name = dlg.text
            self.typeRenameDict[current] = dlg.text
            self.loadList(dlg.text)

    def delType(self):
        """Remove selected type from the list"""
        current = str(self.listBox.currentText())
        format = self.formatList.findFormat(current)
        if self.mainWin.doc.root.usesType(format):
            QMessageBox.warning(self, 'TreeLine', \
                                'Cannot delete data type being used by nodes')
            return
        self.formatList.removeQuiet(format)
        self.loadList()


class FieldFormatDialog(QDialog):
    """Dialog for changing the field type settings"""
    types = ['Text', 'LongText', 'Number', 'Choice', 'Combination', \
             'Date', 'Time', 'Boolean', 'URL', 'Path', 'Email', 'Picture']
    def __init__(self, field, allowTypeChg=1, parent=None, name=None, \
                 modal=0, flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.currentType = field.typeName
        self.format = field.format
        self.prefix = field.prefix
        self.suffix = field.suffix
        self.html = field.html
        caption = 'Field Format'
        if sys.platform == 'win32':
            caption += '  (PyQt)'
        self.setCaption(caption)
        self.formatHelpView = None
        topLayout = QGridLayout(self, 3, 2, 10, 5)

        typeBox = QVGroupBox('Field &Type', self)
        topLayout.addWidget(typeBox, 0, 0)
        self.typeCombo = QComboBox(0, typeBox)
        for type in FieldFormatDialog.types:
            self.typeCombo.insertItem(type)
        self.typeCombo.setCurrentItem(FieldFormatDialog.types.\
                                      index(self.currentType))
        self.connect(self.typeCombo, SIGNAL('activated(const QString&)'), \
                     self.changeType)
        typeBox.setEnabled(allowTypeChg)

        self.formatBox = QHGroupBox('Output &Format', self)
        topLayout.addWidget(self.formatBox, 0, 1)
        self.formatEdit = QLineEdit(self.formatBox)
        self.formatEdit.setText(self.format)
        self.formatBox.setEnabled(field.defaultFormat != '')
        self.helpButton = QPushButton('Format Help', self.formatBox)
        self.connect(self.helpButton, SIGNAL('clicked()'), self.formatHelp)

        extraBox = QHGroupBox('&Extra Text', self)
        topLayout.addWidget(extraBox, 1, 0)
        prefixBox = QVBox(extraBox)
        QLabel('Prefix', prefixBox)
        self.prefixEdit = QLineEdit(prefixBox)
        self.prefixEdit.setText(self.prefix)
        suffixBox = QVBox(extraBox)
        QLabel('Suffix', suffixBox)
        self.suffixEdit = QLineEdit(suffixBox)
        self.suffixEdit.setText(self.suffix)

        self.handleBox = QVButtonGroup('Content Text Handling', self)
        topLayout.addWidget(self.handleBox, 1, 1)
        self.htmlButton = QRadioButton('&Allow HTML rich text', self.handleBox)
        plainButton = QRadioButton('&Plain text with line breaks', \
                                   self.handleBox)
        self.htmlButton.setChecked(self.html)
        plainButton.setChecked(not self.html)
        self.handleBox.setEnabled(field.htmlOption)

        ctrlLayout = QHBoxLayout()
        topLayout.addMultiCellLayout(ctrlLayout, 2, 2, 0, 1)
        ctrlLayout.insertStretch(0)
        okButton = QPushButton('&OK', self)
        ctrlLayout.addWidget(okButton)
        self.connect(okButton, SIGNAL('clicked()'), self, SLOT('accept()'))
        cancelButton = QPushButton('&Cancel', self)
        ctrlLayout.addWidget(cancelButton)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))
        self.typeCombo.setFocus()

    def changeType(self, text):
        """Change type name based on combo box signal"""
        self.currentType = str(text)
        formatText = eval(self.currentType + 'Format').defaultFormat
        self.formatEdit.setText(formatText)
        self.formatBox.setEnabled(formatText != '')
        self.handleBox.setEnabled(eval(self.currentType + 'Format').htmlOption)

    def formatHelp(self):
        """Show help window for formats"""
        menu = self.formatMenu()
        menu.popup(self.helpButton.\
                   mapToGlobal(QPoint(0, self.helpButton.height())))
        self.connect(menu, SIGNAL('activated(int)'), self.insertFormat)

    def formatMenu(self):
        """Popup menu for format help"""
        menu = QPopupMenu(self)
        self.formatDict = {}
        index = 0
        for item in eval(self.currentType + 'Format').formatMenuList:
            if item:
                descr, key = item
                self.formatDict[index] = key
                menu.insertItem(descr, index)
                index += 1
            else:
                menu.insertSeparator()
        return menu

    def insertFormat(self, id):
        """Insert format text from id into edit box"""
        self.formatEdit.insert(self.formatDict[id])

    def accept(self):
        """Retrieve data settings before closing"""
        self.format = str(self.formatEdit.text())
        self.prefix = str(self.prefixEdit.text())
        self.suffix = str(self.suffixEdit.text())
        self.html = self.htmlButton.isChecked()
        return QDialog.accept(self)


class OtherFieldDlg(QDialog):
    """Dialog for changing to ancestor and file data fields in the ConfigDlg"""
    levelList = ['No Other Reference', 'File Info Reference', \
                 'Any Ancestor Reference', 'Parent Reference', \
                 'Grandparent Reference', 'Great Grandparent Reference']
    levelNums = [0, 0, -1, 1, 2, 3]   # correspond to levelList
    def __init__(self, parentLevel, otherType, formatList, fileInfoFormat, \
                 parent=None, name=None, modal=0, flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.parentLevel = parentLevel
        self.otherType = otherType
        self.formatList = formatList
        self.fileInfoFormat = fileInfoFormat
        self.setIcon(QPixmap(treeline_xpm))
        if sys.platform == 'win32':
            self.setCaption('Choose Other Field References  (PyQt)')
        else:
            self.setCaption('Choose Other Field References')

        topLayout = QVBoxLayout(self, 5)
        levelBox = QHGroupBox('Reference Level', self)
        topLayout.addWidget(levelBox)
        levelCombo = QComboBox(0, levelBox)
        for name in OtherFieldDlg.levelList:
            levelCombo.insertItem(name)
        levelCombo.setCurrentItem(OtherFieldDlg.levelNums.\
                                                index(self.parentLevel))
        if not self.parentLevel and otherType:
            levelCombo.setCurrentItem(1)
        self.connect(levelCombo, SIGNAL('activated(int)'), self.changeLevel)

        typeBox =  QHGroupBox('Parent Type', self)
        topLayout.addWidget(typeBox)
        self.typeCombo =  QComboBox(0, typeBox)
        names = self.formatList.nameList()
        names.sort()
        for name in names:
            self.typeCombo.insertItem(name)
        if otherType and self.parentLevel:
            self.typeCombo.setCurrentItem(names.index(otherType.name))
        self.typeCombo.setEnabled(self.parentLevel)
        self.connect(self.typeCombo, SIGNAL('activated(const QString&)') , \
                     self.changeType)

        ctrlLayout = QHBoxLayout(topLayout)
        ctrlLayout.insertStretch(0)
        okButton = QPushButton('&OK', self)
        ctrlLayout.addWidget(okButton)
        self.connect(okButton, SIGNAL('clicked()'), self, SLOT('accept()'))
        cancelButton = QPushButton('&Cancel', self)
        ctrlLayout.addWidget(cancelButton)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))

    def changeLevel(self, index):
        """Change level in response to combo box change"""
        self.parentLevel = OtherFieldDlg.levelNums[index]
        if index and not self.parentLevel:
            self.otherType = self.fileInfoFormat
        elif self.parentLevel:
            self.otherType = self.formatList.findFormat(str(self.typeCombo.\
                                                             currentText()))
        else:
            self.otherType = None
        self.typeCombo.setEnabled(self.parentLevel)

    def changeType(self, name):
        """Change type in response to combo box change"""
        self.otherType = self.formatList.findFormat(str(name))


class TypeSetDlg(QDialog):
    """Dialog for setting items to a type"""
    def __init__(self, mainWin, parent=None, name=None, modal=0, \
                 flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.mainWin = mainWin
        self.setIcon(QPixmap(treeline_xpm))
        if sys.platform == 'win32':
            self.setCaption('Set Data Types  (PyQt)')
        else:
            self.setCaption('Set Data Types')
        topLayout = QVBoxLayout(self, 5)
        self.groupBox = QGroupBox('', self)
        topLayout.addWidget(self.groupBox)
        innerLayout = QGridLayout(self.groupBox, 5, 2, 10, 5)
        innerLayout.addRowSpacing(0, self.fontMetrics().lineSpacing())
        self.listBox = QListBox(self.groupBox)
        innerLayout.addMultiCellWidget(self.listBox, 1, 4, 0, 0)
        self.itemButton = QPushButton('Set &Selection', self.groupBox)
        innerLayout.addWidget(self.itemButton, 1, 1)
        self.connect(self.itemButton, SIGNAL('clicked()'), self.setItem)
        self.childButton = QPushButton('Set S&election\'s Children', \
                                       self.groupBox)
        innerLayout.addWidget(self.childButton, 2, 1)
        self.connect(self.childButton, SIGNAL('clicked()'), self.setChild)
        self.descendButton = QPushButton('Set All &Descendants', self.groupBox)
        innerLayout.addWidget(self.descendButton, 3, 1)
        self.connect(self.descendButton, SIGNAL('clicked()'), self.setDescend)
        self.conditionButton = QPushButton('Set Descendants C&ondtionally...', \
                                           self.groupBox)
        innerLayout.addWidget(self.conditionButton, 4, 1)
        self.connect(self.conditionButton, SIGNAL('clicked()'), \
                     self.setCondition)

        ctrlLayout = QHBoxLayout(topLayout)
        ctrlLayout.insertStretch(0)
        closeButton = QPushButton('&Close', self)
        ctrlLayout.addWidget(closeButton)
        self.connect(closeButton, SIGNAL('clicked()'), self, SLOT('close()'))
        self.loadList()
        self.connect(self.listBox, SIGNAL('selectionChanged()'), \
                     self.updateDlg)
        self.listBox.setFocus()

    def loadList(self):
        """Load types into list box"""
        names = docparam.nodeFormats.nameList()
        names.sort()
        self.listBox.clear()
        for name in names:
            self.listBox.insertItem(name)
        self.listBox.setSelected(0, 1)
        self.updateDlg()

    def updateDlg(self):
        """Update label text & button availability"""
        selList = self.mainWin.doc.selectedNodes
        selTypes = [item.nodeFormat.name for item in selList]
        childTypes = []
        descendTypes = []
        for item in selList:
            childTypes.extend([i.nodeFormat.name for i in item.childList])
            descendTypes.extend([i.nodeFormat.name for i in \
                                 item.descendantList(1)][1:])
        if len(selList) == 1:
            self.groupBox.setTitle('Selection = "%s"' % selList[0].title())
        elif len(selList) > 1:
            self.groupBox.setTitle('Multiple Selection')
        else:
            self.groupBox.setTitle('No Selection')
        currentType = str(self.listBox.currentText())
        self.itemButton.setEnabled(len(selTypes) and \
                                   (min(selTypes) != max(selTypes) or \
                                   selTypes[0] != currentType))
        self.childButton.setEnabled(len(childTypes) and \
                                    (min(childTypes) != max(childTypes) or \
                                     childTypes[0] != currentType))
        self.descendButton.setEnabled(len(descendTypes) and \
                                      (min(descendTypes) != max(descendTypes) \
                                       or descendTypes[0] != currentType))
        self.conditionButton.setEnabled(len(descendTypes) and \
                                      (min(descendTypes) != max(descendTypes) \
                                       or descendTypes[0] != currentType))

    def updateViews(self):
        """Update main views due to type setting changes"""
        docparam.modified = 1
        self.mainWin.updateViews(self.mainWin.doc.selectedNodes)

    def setCurrentSel(self):
        """Set type list selection to current item on initial open"""
        self.blockSignals(1)
        selTypes = [item.nodeFormat.name for item \
                    in self.mainWin.doc.selectedNodes]
        names = docparam.nodeFormats.nameList()
        names.sort()
        if selTypes and min(selTypes) == max(selTypes):
            self.listBox.setSelected(names.index(selTypes[0]), 1)
        self.blockSignals(0)

    def setItem(self):
        """Set types for selected item"""
        newFormat = docparam.nodeFormats.\
                    findFormat(str(self.listBox.currentText()))
        docparam.undoStore.addTypeUndo(self.mainWin.doc.selectedNodes, \
                                       self.mainWin.doc.selectedNodes)
        for item in self.mainWin.doc.selectedNodes:
            item.nodeFormat = newFormat
        self.updateDlg()
        self.updateViews()

    def setChild(self):
        """Set types for selected item's children"""
        newFormat = docparam.nodeFormats.\
                    findFormat(str(self.listBox.currentText()))
        childList = []
        for parent in self.mainWin.doc.selectedNodes:
            for child in parent.childList:
                childList.append(child)
        docparam.undoStore.addTypeUndo(childList, \
                                       self.mainWin.doc.selectedNodes)
        for item in childList:
            item.nodeFormat = newFormat
        self.updateDlg()
        self.updateViews()

    def setDescend(self):
        """Set types for selected item's descendants"""
        newFormat = docparam.nodeFormats.\
                    findFormat(str(self.listBox.currentText()))
        itemList = []
        for parent in self.mainWin.doc.selectedNodes:
            for item in parent.descendantList(1)[1:]:
                itemList.append(item)
        docparam.undoStore.addTypeUndo(itemList, self.mainWin.doc.selectedNodes)
        for item in itemList:
            item.nodeFormat = newFormat
        self.updateDlg()
        self.updateViews()

    def setCondition(self):
        """Set types for selected item's descendants conditionally"""
        typeList = []
        for item in self.mainWin.doc.selectedNodes:
            for type in item.descendTypes():
                if type.name not in typeList:
                    typeList.append(type.name)
        if len(typeList) > 1:
            typeList = QStringList.split(' ', ' '.join(typeList))
            caption = 'Select Type'
            if sys.platform == 'win32':
                caption += '  (PyQt)'
            type, ok = QInputDialog.getItem(caption, 'Change from data type', \
                                            typeList, 0, 0, self)
            if not ok:
                return
            type = str(type)
        else:
            type = typeList[0]
        format = docparam.nodeFormats.findFormat(type)
        dlg = FilterDlg(format.fieldNames(), self, None, 1)
        if dlg.exec_loop() != QDialog.Accepted:
            return
        expr = dlg.exprText()
        if expr:
            newFormat = docparam.nodeFormats.\
                        findFormat(str(self.listBox.currentText()))
            itemList = []
            for parent in self.mainWin.doc.selectedNodes:
                for item in parent.descendantList(1)[1:]:
                    data = item.editDataDict(self.mainWin.editFormats)
                    if item.nodeFormat == format and eval(expr):
                        itemList.append(item)
            docparam.undoStore.addTypeUndo(itemList, \
                                           self.mainWin.doc.selectedNodes)
            for item in itemList:
                item.nodeFormat = newFormat
            self.updateDlg()
            self.updateViews()

    def keyPressEvent(self, event):
        """Close on escape key"""
        if event.key() == Qt.Key_Escape:
            self.close()
        QDialog.keyPressEvent(self, event)

    def closeEvent(self, event):
        """Signal that view is closing"""
        self.emit(PYSIGNAL('viewClosed'), (0,))
        event.accept()


class FieldSelectList(QListView):
    """List view that shows direction of sort, changes with right-click
       or left/right arrows"""
    def __init__(self, parent=None, name=None):
        QListView.__init__(self, parent, name)

    def contentsMousePressEvent(self, event):
        """Signal right-click"""
        if event.button() == Qt.RightButton:
            item = self.itemAt(self.contentsToViewport(event.pos()))
            if item:
                self.emit(PYSIGNAL('directionToggle'), (item,))
        else:
            QListView.contentsMousePressEvent(self, event)

    def keyPressEvent(self, event):
        if event.key() in (Qt.Key_Left, Qt.Key_Right):
            item = self.currentItem()
            if item:
                self.emit(PYSIGNAL('directionToggle'), (item,))
        else:
            QListView.keyPressEvent(self, event)


class FieldSelectDlg(QDialog):
    """Dialog for selecting from a field list in order"""
    directionText = ['descend', 'ascend']
    def __init__(self, fieldList, caption, label, directional=0, \
                 parent=None, name=None, modal=0, flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.availFields = fieldList[:]
        self.availFields.reverse()
        self.selFields = []
        self.directional = directional
        self.setIcon(QPixmap(treeline_xpm))
        if sys.platform == 'win32':
            caption += '  (PyQt)'
        self.setCaption(caption)
        topLayout = QGridLayout(self, 2, 3, 10, 5)
        groupBox = QVGroupBox(label, self)
        topLayout.addMultiCellWidget(groupBox, 0, 0, 0, 2)
        self.listView = directional and FieldSelectList(groupBox) or \
                        QListView(groupBox)
        self.listView.addColumn('#')
        self.listView.addColumn('Fields')
        if directional:
            self.listView.addColumn('Direction')
            self.connect(self.listView, PYSIGNAL('directionToggle'), \
                         self.toggleDirection)
        self.listView.setSorting(-1)
        self.listView.setAllColumnsShowFocus(1)
        self.listView.setSelectionMode(QListView.Multi)
        self.listView.setFocus()
        self.connect(self.listView, SIGNAL('returnPressed(QListViewItem*)'), \
                     self.selectCurrent)
        self.connect(self.listView, SIGNAL('selectionChanged()'), \
                     self.updateSelFields)
        QLabel('(Use right mouse click or\nleft/right keys to change\n'\
               'direction)', groupBox)

        self.okButton = QPushButton('&OK', self)
        topLayout.addWidget(self.okButton, 1, 1)
        self.okButton.setEnabled(0)
        self.connect(self.okButton, SIGNAL('clicked()'), \
                     self, SLOT('accept()'))
        cancelButton = QPushButton('&Cancel', self)
        topLayout.addWidget(cancelButton, 1, 2)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))
        self.loadFields()

    def loadFields(self):
        """Load fields into list view"""
        for field in self.availFields:
            QListViewItem(self.listView, '', field)

    def getSelList(self):
        """Return sort list, a list of tuples if directional"""
        result = self.selFields[:]
        if self.directional:
            item = self.listView.firstChild()
            while item:
                num = str(item.text(0))
                if num:
                    num = int(num) - 1
                    result[num] = (result[num], FieldSelectDlg.directionText.\
                                                index(str(item.text(2))))
                item = item.nextSibling()
        return result

    def selectCurrent(self, viewItem):
        """Select current item on return pressed signal"""
        self.listView.setSelected(viewItem, not viewItem.isSelected())

    def toggleDirection(self, item):
        """Toggle sort direction for viewItem"""
        origText = str(item.text(2))
        if origText:
            item.setText(2, FieldSelectDlg.directionText[not FieldSelectDlg.\
                                                         directionText.\
                                                         index(origText)])

    def updateSelFields(self):
        """Update selected fields based on current selections"""
        item = self.listView.firstChild()
        while item:
            if item.isSelected():
                if str(item.text(1)) not in self.selFields:
                    self.selFields.append(str(item.text(1)))
                    if self.directional:
                        item.setText(2, FieldSelectDlg.directionText[1])
            elif str(item.text(1)) in self.selFields:
                self.selFields.remove(str(item.text(1)))
                if self.directional:
                    item.setText(2, '')
            item = item.nextSibling()
        item = self.listView.firstChild()
        while item:
            if item.isSelected():
                item.setText(0, `self.selFields.index(str(item.text(1))) \
                                 + 1`)
            else:
                item.setText(0, '')
            item = item.nextSibling()
        self.okButton.setEnabled(len(self.selFields))


class EditFieldsDlg(QDialog):
    """Dialog for editing all instances of the selected nodes field"""
    def __init__(self, fieldList, parent=None, name=None, modal=0, \
                 flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.resultDict = {}
        self.setIcon(QPixmap(treeline_xpm))
        if sys.platform == 'win32':
            self.setCaption('Change Selecttion  (PyQt)')
        else:
            self.setCaption('Change Selection')
        topLayout = QGridLayout(self, 3, 3, 10, 5)
        groupBox = QVGroupBox('&Field', self)
        topLayout.addMultiCellWidget(groupBox, 0, 0, 0, 2)
        self.fieldBox = QComboBox(0, groupBox)
        for field in fieldList:
            self.fieldBox.insertItem(field)
        self.connect(self.fieldBox, SIGNAL('activated(const QString &)'), \
                     self.changeField)

        groupBox = QVGroupBox('&New Value', self)
        topLayout.addMultiCellWidget(groupBox, 1, 1, 0, 2)
        self.editBox = QLineEdit(groupBox)
        self.connect(self.editBox, SIGNAL('textChanged(const QString &)'), \
                     self.updateText)

        self.okButton = QPushButton('&OK', self)
        topLayout.addWidget(self.okButton, 2, 1)
        self.okButton.setEnabled(0)
        self.connect(self.okButton, SIGNAL('clicked()'), \
                     self, SLOT('accept()'))
        cancelButton = QPushButton('&Cancel', self)
        topLayout.addWidget(cancelButton, 2, 2)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))
        self.fieldBox.setFocus()

    def changeField(self, newField):
        """Update the value editor based on new field name"""
        self.editBox.blockSignals(1)
        self.editBox.setText(self.resultDict.get(str(newField), ''))
        self.editBox.blockSignals(0)

    def updateText(self, newText):
        """Update the stored values based on editor change"""
        self.resultDict[str(self.fieldBox.currentText())] \
            = unicode(newText)
        self.okButton.setEnabled(1)


class ExportDlg(QDialog):
    """Dialog for selecting type of file export"""
    htmlType = 0
    dirType = 1
    xsltType = 2
    trlType = 3
    textType = 4
    tableType = 5
    xbelType = 6
    mozType = 7
    def __init__(self, itemText, parent=None, name=None, modal=0, \
                 flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.exportType = ExportDlg.htmlType
        self.setIcon(QPixmap(treeline_xpm))
        if sys.platform == 'win32':
            self.setCaption('Export File  (PyQt)')
        else:
            self.setCaption('Export File')
        topLayout = QVBoxLayout(self, 5)
        label = QLabel(self)
        label.setFrameStyle(QFrame.Panel | QFrame.Sunken)
        label.setTextFormat(Qt.PlainText)
        label.setText('Exporting from "%s"' % itemText)
        topLayout.addWidget(label)
        groupBox = QVButtonGroup('Export Type', self)
        topLayout.addWidget(groupBox)
        QRadioButton('&HTML single file output', groupBox)
        QRadioButton('HTML &directories output', groupBox)
        QRadioButton('&XSLT output', groupBox)
        QRadioButton('TreeLine &subtree', groupBox)
        QRadioButton('&Tabbed title text', groupBox)
        QRadioButton('Table &of child data', groupBox)
        QRadioButton('XBEL &bookmarks', groupBox)
        QRadioButton('&Mozilla HTML bookmarks', groupBox)
        groupBox.setButton(0)
        self.connect(groupBox, SIGNAL('clicked(int)'), self.updateAvail)

        optionBox = QVGroupBox('Export Options', self)
        topLayout.addWidget(optionBox)
        self.rootButton = QCheckBox('&Include root node', optionBox)
        self.openOnlyButton = QCheckBox('Only open &node children', optionBox)
        self.headerButton = QCheckBox('Include print header && &footer', \
                                      optionBox)
        columnBox = QHBox(optionBox)
        columnBox.setSpacing(5)
        self.numColSpin = QSpinBox(1, treemainwin.TreeMainWin.maxNumCol, \
                                   1, columnBox)
        self.numColSpin.setMaximumWidth(40)
        self.colLabel = QLabel('&Columns', columnBox)
        self.colLabel.setBuddy(self.numColSpin)

        ctrlLayout = QHBoxLayout(topLayout)
        ctrlLayout.insertStretch(0)
        okButton = QPushButton('&OK', self)
        ctrlLayout.addWidget(okButton)
        self.connect(okButton, SIGNAL('clicked()'), self, SLOT('accept()'))
        cancelButton = QPushButton('&Cancel', self)
        ctrlLayout.addWidget(cancelButton)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))
        groupBox.setFocus()

    def updateAvail(self, numClicked):
        """Update options available based on export type change"""
        self.exportType = numClicked
        if numClicked in (ExportDlg.htmlType, ExportDlg.xsltType, \
                          ExportDlg.textType):
            self.rootButton.setEnabled(1)
        elif numClicked == ExportDlg.tableType:
            self.rootButton.setChecked(0)
            self.rootButton.setEnabled(0)
        else:
            self.rootButton.setChecked(1)
            self.rootButton.setEnabled(0)
        if numClicked in (ExportDlg.htmlType, ExportDlg.textType):
            self.openOnlyButton.setEnabled(1)
        else:
            self.openOnlyButton.setChecked(0)
            self.openOnlyButton.setEnabled(0)
        if numClicked in (ExportDlg.htmlType, ExportDlg.dirType):
            self.headerButton.setEnabled(1)
        else:
            self.headerButton.setChecked(0)
            self.headerButton.setEnabled(0)
        if numClicked == ExportDlg.htmlType:
            self.colLabel.setEnabled(1)
        else:
            self.numColSpin.setValue(1)
            self.numColSpin.setEnabled(0)
            self.colLabel.setEnabled(0)

    def isRootIncluded(self):
        """Return true if root include box is checked"""
        return self.rootButton.isChecked()

    def openOnly(self):
        """Return true if open only box is checked"""
        return self.openOnlyButton.isChecked()

    def headerIncluded(self):
        """Return true if header/footer include box is checked"""
        return self.headerButton.isChecked()
 
    def numColumns(self):
        """Return number of columns in spin box"""
        return self.numColSpin.value()


class FilterDlg(QDialog):
    """Dialog for selecting filter rules"""
    def __init__(self, fieldList, parent=None, name=None, \
                 modal=0, flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.fieldList = fieldList
        self.setIcon(QPixmap(treeline_xpm))
        if sys.platform == 'win32':
            self.setCaption('Filter Data  (PyQt)')
        else:
            self.setCaption('Filter Data')
        self.topLayout = QVBoxLayout(self, 5)
        
        self.ruleList = [FilterRule(1, self.fieldList, self)]
        self.topLayout.addWidget(self.ruleList[-1])
        self.connect(self.ruleList[-1], PYSIGNAL('opChange'), self.setValid)
        self.validRuleCount = 0
        self.boolList = []

        self.ctrlLayout = QHBoxLayout(self.topLayout)
        self.ctrlLayout.insertStretch(0)
        addButton = QPushButton('&Add Rule', self)
        self.ctrlLayout.addWidget(addButton)
        self.connect(addButton, SIGNAL('clicked()'), self.addRule)
        self.okButton = QPushButton('&OK', self)
        self.ctrlLayout.addWidget(self.okButton)
        self.okButton.setEnabled(0)
        self.connect(self.okButton, SIGNAL('clicked()'), \
                     self, SLOT('accept()'))
        cancelButton = QPushButton('&Cancel', self)
        self.ctrlLayout.addWidget(cancelButton)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))

    def addRule(self):
        """Add new rule to dialog"""
        self.boolList.append(QComboBox(0, self))
        self.boolList[-1].insertItem('and')
        self.boolList[-1].insertItem('or')
        self.topLayout.insertWidget(len(self.ruleList) * 2 - 1, \
                                    self.boolList[-1], 0, Qt.AlignHCenter )
        self.ruleList.append(FilterRule(len(self.ruleList) + 1, \
                                        self.fieldList, self))
        self.topLayout.insertWidget(len(self.ruleList) * 2 - 2, \
                                    self.ruleList[-1])
        self.connect(self.ruleList[-1], PYSIGNAL('opChange'), self.setValid)
        self.boolList[-1].show()
        self.ruleList[-1].show()

    def setValid(self, valid):
        """Update ok button status if a rule is valid"""
        if valid:
            self.validRuleCount += 1
        else:
            self.validRuleCount -= 1
        self.okButton.setEnabled(self.validRuleCount)

    def exprText(self):
        """Return current expression text"""
        if not self.validRuleCount:
            return ''
        ruleTextList = [rule.ruleText() for rule in self.ruleList]
        boolTextList = [box.currentText() for box in self.boolList]
        while not ruleTextList[0]:
            del ruleTextList[0]
            del boolTextList[0]
        expr = ruleTextList.pop(0)
        for rule in ruleTextList:
            bool = boolTextList.pop(0)
            if rule:
                expr = '%s %s %s' % (expr, bool, rule)
        return expr


class FilterRule(QHGroupBox):
    """Saves rules for filtering items"""
    oper = ('None', 'All', '==', '<', '<=', '>', '>=', '!=', \
            'starts with', 'ends with')
    def __init__(self, num, fieldList, parent=None, name=None):
        QHGroupBox.__init__(self, parent, name)
        self.setTitle('Rule ' + `num`)
        self.fieldBox = QComboBox(0, self)
        for field in fieldList:
            self.fieldBox.insertItem(field)
        self.opBox = QComboBox(0, self)
        for op in FilterRule.oper:
            self.opBox.insertItem(op)
        self.connect(self.opBox, SIGNAL('activated(const QString &)'), \
                     self.changeOp)
        self.edit = QLineEdit(self)
        self.edit.setMinimumWidth(80)
        self.edit.setEnabled(0)
        self.valid = 0
        self.fieldBox.setFocus()

    def changeOp(self, newOp):
        """Update the filter based on type selection change"""
        valid = (str(newOp) != 'None')
        self.edit.setEnabled(valid and str(newOp) != 'All')
        if self.valid != valid:
            self.emit(PYSIGNAL('opChange'), (valid,))
            self.valid = valid

    def ruleText(self):
        """Return text expression for rule"""
        if not self.valid:
            return ''
        op = str(self.opBox.currentText())
        field = str(self.fieldBox.currentText())
        text = unicode(self.edit.text()).lower()
        if len(op) <= 2:
            return "data.get('%s', '').lower() %s '''%s'''" % (field, op, text)
        if op == 'All':
            return '1'
        if op == 'starts with':
            return "data.get('%s', '').lower().startswith('''%s''')" \
                   % (field, text)
        return "data.get('%s', '').lower().endswith('''%s''')" % (field, text)


class NumberingDlg(QDialog):
    """Dialog for adding numbering to nodes"""
    outlineType = 0
    sectionType = 1
    singleType = 2
    outlineFormat = ['I.', 'A.', '1.', 'a)', '(1)', '(a)', '(i)']
    sectionFormat = ['1', '.1', '.1']
    singleFormat = ['1.']
    def __init__(self, fieldList, maxLevels, parent=None, name=None, \
                 modal=0, flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.maxLevels = maxLevels
        self.currentStyle = NumberingDlg.outlineType
        self.currentFormat = []
        self.setIcon(QPixmap(treeline_xpm))
        if sys.platform == 'win32':
            self.setCaption('Data Numbering  (PyQt)')
        else:
            self.setCaption('Data Numbering')
        topLayout = QVBoxLayout(self, 5)
        groupBox = QVGroupBox('&Number Field', self)
        topLayout.addWidget(groupBox)
        self.fieldBox = QComboBox(1, groupBox)
        for field in fieldList:
            self.fieldBox.insertItem(field)
        self.fieldBox.clearEdit()
        self.fieldBox.setFocus()
        self.connect(self.fieldBox, SIGNAL('textChanged(const QString&)'), \
                     self.updateField)

        groupBox = QVGroupBox('Root Node', self)
        topLayout.addWidget(groupBox)
        self.inclRootButton = QCheckBox('&Include root node', groupBox)
        self.inclRootButton.setChecked(1)
        self.connect(self.inclRootButton, SIGNAL('toggled(bool)'), \
                     self.updateRoot)

        self.styleBox = QVButtonGroup('Number Style', self)
        topLayout.addWidget(self.styleBox)
        QRadioButton('Outline (&discrete numbers)', self.styleBox)
        QRadioButton('&Section (append to parent number)', self.styleBox)
        QRadioButton('Single &level (children only)', self.styleBox)
        self.styleBox.setButton(self.currentStyle)
        self.connect(self.styleBox, SIGNAL('clicked(int)'), self.updateStyle)

        groupBox = QHGroupBox('Number &Format', self)
        topLayout.addWidget(groupBox)
        self.formatEdit = QLineEdit(groupBox)
        self.connect(self.formatEdit, SIGNAL('textChanged(const QString&)'), \
                     self.updateFormat)
        QLabel('for Level', groupBox)
        self.levelBox = QSpinBox(groupBox)
        self.connect(self.levelBox, SIGNAL('valueChanged(int)'), \
                     self.updateLevel)

        groupBox = QHGroupBox('Initial N&umber', self)
        topLayout.addWidget(groupBox)
        QLabel('Start first level at number', groupBox)
        self.startBox = QSpinBox(1, 1000000, 1, groupBox)
        self.loadFormat()

        ctrlLayout = QHBoxLayout(topLayout)
        ctrlLayout.insertStretch(0)
        self.okButton = QPushButton('&OK', self)
        ctrlLayout.addWidget(self.okButton)
        self.okButton.setEnabled(0)
        self.connect(self.okButton, SIGNAL('clicked()'), \
                     self, SLOT('accept()'))
        cancelButton = QPushButton('&Cancel', self)
        ctrlLayout.addWidget(cancelButton)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))

    def updateField(self, text):
        """Update OK button based on combo changes"""
        self.okButton.setEnabled(len(str(text).strip()))

    def updateRoot(self, on):
        """Update styles based on include root change"""
        if self.maxLevels <= 1:
            if not on:
                self.currentStyle = NumberingDlg.singleType
                self.styleBox.setButton(self.currentStyle)
                self.styleBox.find(NumberingDlg.outlineType).setEnabled(0)
                self.styleBox.find(NumberingDlg.sectionType).setEnabled(0)
            else:
                self.styleBox.find(NumberingDlg.outlineType).setEnabled(1)
                self.styleBox.find(NumberingDlg.sectionType).setEnabled(1)
        if self.currentStyle == NumberingDlg.singleType and on:
            self.currentStyle = NumberingDlg.outlineType
            self.styleBox.find(NumberingDlg.outlineType).setChecked(1)
        self.loadFormat()

    def updateStyle(self, style):
        """Update dialog based on style selection change"""
        self.currentStyle = style
        if style == NumberingDlg.singleType:
            self.inclRootButton.setChecked(0)
        self.loadFormat()

    def loadFormat(self):
        """Load a default format and level numbers into dialog"""
        numLevels = self.maxLevels
        startLevel = 1
        if self.inclRootButton.isOn():
            numLevels += 1
            startLevel = 0
        if self.currentStyle == NumberingDlg.singleType:
            self.currentFormat = copy.deepcopy(NumberingDlg.singleFormat)
            numLevels = 1
        elif self.currentStyle == NumberingDlg.outlineType:
            self.currentFormat = copy.deepcopy(NumberingDlg.outlineFormat)
        else:
            self.currentFormat = copy.deepcopy(NumberingDlg.sectionFormat)
        while len(self.currentFormat) < numLevels:
            self.currentFormat.extend(self.currentFormat[-2:])
        self.currentFormat = self.currentFormat[:numLevels]
        self.levelBox.setMinValue(startLevel)
        self.levelBox.setMaxValue(startLevel + numLevels - 1)
        self.levelBox.setValue(startLevel)
        self.updateLevel(startLevel)

    def updateLevel(self, levelNum):
        """Update dialog based on a level change"""
        self.formatEdit.blockSignals(1)
        self.formatEdit.setText(self.currentFormat[levelNum \
                                                   - self.levelBox.minValue()])
        self.formatEdit.blockSignals(0)

    def updateFormat(self, text):
        """Update dialog based on a format string change"""
        self.currentFormat[self.levelBox.value() \
                           - self.levelBox.minValue()] = unicode(text).strip()

    def getField(self):
        """Return adjusted field name"""
        return str(self.fieldBox.currentText()).strip()

    def accept(self):
        """Check for acceptable field string before closing"""
        try:
            text = str(self.fieldBox.currentText()).strip()
        except UnicodeError:
            text = ''
        if not text.replace('_', '').isalnum():
            QMessageBox.warning(self, 'TreeLine', \
                                'Illegal characters in field '
                                '(only alpa-numerics & underscores allowed)')
            return
        return QDialog.accept(self)

    def includeRoot(self):
        """Return 1 if root include box is checked"""
        return self.inclRootButton.isOn()

    def startNumber(self):
        """Return value from start number box"""
        return self.startBox.value()


class FindTextEntry(QDialog):
    """Dialog for find string text entry"""
    def __init__(self, mainWin, parent=None, name=None, modal=0, \
                 flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.mainWin = mainWin
        caption = 'Find'
        if sys.platform == 'win32':
            caption += '  (PyQt)'
        self.setCaption(caption)
        self.setIcon(QPixmap(treeline_xpm))
        topLayout = QVBoxLayout(self, 5)
        label = QLabel('Enter key words', self)
        topLayout.addWidget(label)
        self.entry = QLineEdit(self)
        topLayout.addWidget(self.entry)
        self.entry.setFocus()
        # self.connect(self.entry, SIGNAL('returnPressed()'), self.findNext)
        ctrlLayout = QHBoxLayout(topLayout)
        ctrlLayout.insertStretch(0)
        prevButton = QPushButton('Find &Previous', self)
        ctrlLayout.addWidget(prevButton)
        self.connect(prevButton, SIGNAL('clicked()'), self.findPrev)
        nextButton = QPushButton('Find &Next', self)
        ctrlLayout.addWidget(nextButton)
        nextButton.setDefault(1)
        self.connect(nextButton, SIGNAL('clicked()'), self.findNext)
        closeButton = QPushButton('&Close', self)
        ctrlLayout.addWidget(closeButton)
        self.connect(closeButton, SIGNAL('clicked()'), self, SLOT('close()'))

    def find(self, forward=1):
        """Find match in direction"""
        text = unicode(self.entry.text()).strip()
        if text:
            item = self.mainWin.doc.\
                        findText(self.mainWin.treeView.currentTreeItem(), \
                                 text, forward)
            if item:
                self.mainWin.treeView.changeSelect(item, 1)
            else:
                self.mainWin.statusBar().message('Text string not found', 4000)

    def findNext(self):
        """Find next match"""
        self.find(1)

    def findPrev(self):
        """Find previous match"""
        self.find(0)
    
    def keyPressEvent(self, event):
        """Close on escape key"""
        if event.key() == Qt.Key_Escape:
            self.close()
        QDialog.keyPressEvent(self, event)

    def closeEvent(self, event):
        """Signal that view is closing"""
        self.emit(PYSIGNAL('viewClosed'), (0,))
        event.accept()


class SpellCheckDlg(QDialog):
    """Dialog for the spell check interface"""
    def __init__(self, spCheck, parent=None, name=None, modal=0, \
                 flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.setIcon(QPixmap(treeline_xpm))
        if sys.platform == 'win32':
            self.setCaption('Spell Check  (PyQt)')
        else:
            self.setCaption('Spell Check')
        self.spCheck = spCheck
        self.replaceAllDict = {}
        self.word = ''
        self.textLine = ''
        self.postion = 0
        self.ignoreWord = ''
        self.newLine = ''

        topLayout = QHBoxLayout(self, 5, 5)
        leftLayout = QVBoxLayout(topLayout)
        wordBox = QVGroupBox('Not in Dictionary', self)
        leftLayout.addWidget(wordBox)
        QLabel('Word:', wordBox)
        self.wordEdit = QLineEdit('', wordBox)
        self.connect(self.wordEdit, SIGNAL('textChanged(const QString&)'), \
                     self.updateFromWord)
        wordBox.addSpace(5)
        QLabel('Context:', wordBox)
        self.contextEdit = SpellContextEdit(wordBox)
        self.connect(self.contextEdit, SIGNAL('textChanged()'), \
                     self.updateFromContext)

        suggestBox = QVGroupBox('Suggestions', self)
        leftLayout.addWidget(suggestBox)
        self.suggestList = QListBox(suggestBox)
        self.connect(self.suggestList, SIGNAL('doubleClicked(QListBoxItem*)'), \
                     self.replace)

        rightLayout = QVBoxLayout(topLayout)
        ignoreButton = QPushButton('Ignor&e', self)
        rightLayout.addWidget(ignoreButton)
        self.connect(ignoreButton, SIGNAL('clicked()'), self.ignore)
        ignoreAllButton = QPushButton('&Ignore All', self)
        rightLayout.addWidget(ignoreAllButton)
        self.connect(ignoreAllButton, SIGNAL('clicked()'), self.ignoreAll)
        rightLayout.addStretch()
        addButton = QPushButton('&Add', self)
        rightLayout.addWidget(addButton)
        self.connect(addButton, SIGNAL('clicked()'), self.add)
        addLowerButton = QPushButton('Add &Lowercase', self)
        rightLayout.addWidget(addLowerButton)
        self.connect(addLowerButton, SIGNAL('clicked()'), self.addLower)
        rightLayout.addStretch()
        replaceButton = QPushButton('&Replace', self)
        rightLayout.addWidget(replaceButton)
        self.connect(replaceButton, SIGNAL('clicked()'), self.replace)
        self.replaceAllButton = QPushButton('Re&place All', self)
        rightLayout.addWidget(self.replaceAllButton)
        self.connect(self.replaceAllButton, SIGNAL('clicked()'), \
                     self.replaceAll)
        rightLayout.addStretch()
        cancelButton = QPushButton('&Cancel', self)
        rightLayout.addWidget(cancelButton)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))
        self.widgetDisableList = [ignoreButton, ignoreAllButton, addButton, \
                                  addLowerButton, self.suggestList]
        self.fullDisableList = self.widgetDisableList + \
                               [self.replaceAllButton, self.wordEdit]

    def setWord(self, textLine, results):
        """Set dialog contents from the checked line and the spell check
           results"""
        self.textLine = textLine
        self.word, self.position, suggestions = results[0]
        self.wordEdit.blockSignals(1)
        self.wordEdit.setText(self.word)
        self.wordEdit.blockSignals(0)
        self.contextEdit.blockSignals(1)
        self.contextEdit.setText(self.textLine)
        self.contextEdit.setSelection(self.position, \
                                      self.position + len(self.word))
        self.contextEdit.blockSignals(0)
        self.suggestList.clear()
        self.suggestList.insertStrList(suggestions)
        self.suggestList.setCurrentItem(0)
        for widget in self.fullDisableList:
            widget.setEnabled(1)
        self.ignoreWord = ''
        self.newLine = ''
        newWord = self.replaceAllDict.get(self.word, '')
        if newWord:
            self.newLine = self.replaceWord(newWord)

    def replaceWord(self, newWord):
        """Return textLine with word replaced to newWord"""
        return self.textLine[:self.position] + newWord + \
                       self.textLine[self.position+len(self.word):]

    def ignore(self):
        """Set word to ignored and close dialog"""
        self.ignoreWord = self.word
        self.accept()

    def ignoreAll(self):
        """Add to dictionary's ignore list and close dialog"""
        self.spCheck.acceptWord(self.word)
        self.accept()

    def add(self):
        """Add to dictionary and close dialog"""
        self.spCheck.addToDict(self.word, 0)
        self.accept()

    def addLower(self):
        """Add to dictionary as lowercase and close dialog"""
        self.spCheck.addToDict(self.word, 1)
        self.accept()

    def replace(self):
        """Replace with current suggestion or contents from word or context
           edit boxes and close dialog"""
        if self.widgetDisableList[0].isEnabled():
            newWord = unicode(self.suggestList.currentText())
            self.newLine = self.replaceWord(newWord)
        else:
            self.newLine = unicode(self.contextEdit.text())
        self.accept()

    def replaceAll(self):
        """Replace with current suggestion (in future too) and close dialog"""
        if self.widgetDisableList[0].isEnabled():
            newWord = unicode(self.suggestList.currentText())
        else:
            newWord = unicode(self.wordEdit.text())
        self.newLine = self.replaceWord(newWord)
        self.replaceAllDict[self.word] = newWord
        self.accept()

    def updateFromWord(self):
        """Update dialog after word editor change"""
        for widget in self.widgetDisableList:
            widget.setEnabled(0)
        newWord = unicode(self.wordEdit.text())
        self.suggestList.clearSelection()
        self.contextEdit.blockSignals(1)
        self.contextEdit.setText(self.replaceWord(newWord))
        self.contextEdit.setSelection(self.position, \
                                      self.position + len(newWord))
        self.contextEdit.blockSignals(0)

    def updateFromContext(self):
        """Update dialog after context editor change"""
        for widget in self.fullDisableList:
            widget.setEnabled(0)
        self.suggestList.clearSelection()


class RadioChoiceDlg(QDialog):
    """Dialog for choosing between a list of text items (radio buttons)
       choiceList contains tuples of item text and return values"""
    def __init__(self, caption, heading, choiceList, parent=None, \
                 name=None, modal=0, flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.setIcon(QPixmap(treeline_xpm))
        if sys.platform == 'win32':
            self.setCaption('%s  (PyQt)' % caption)
        else:
            self.setCaption(caption)
        topLayout = QVBoxLayout(self, 5)
        self.groupBox = QVButtonGroup(heading, self)
        topLayout.addWidget(self.groupBox)
        for text, value in choiceList:
            button = QRadioButton(text, self.groupBox)
            button.returnValue = value
        self.groupBox.setButton(0)

        ctrlLayout = QHBoxLayout(topLayout)
        ctrlLayout.insertStretch(0)
        okButton = QPushButton('&OK', self)
        ctrlLayout.addWidget(okButton)
        self.connect(okButton, SIGNAL('clicked()'), self, SLOT('accept()'))
        cancelButton = QPushButton('&Cancel', self)
        ctrlLayout.addWidget(cancelButton)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))
        self.groupBox.setFocus()

    def getResult(self):
        """Return value of selected button"""
        return self.groupBox.selected().returnValue


class PrintHeaderDlg(QDialog):
    """Dialog for configuring data types"""
    def __init__(self, mainWin, parent=None, name=None, \
                 modal=0, flags=stdFlags):
        QDialog.__init__(self, parent, name, modal, flags)
        self.mainWin = mainWin
        self.setIcon(QPixmap(treeline_xpm))
        if sys.platform == 'win32':
            self.setCaption('Set Print Header & Footer  (PyQt)')
        else:
            self.setCaption('Set Print Header & Footer')
        self.fileInfoFormat = copy.deepcopy(docparam.fileInfoItem.nodeFormat)

        topLayout = QVBoxLayout(self, 7)
        mainLayout = QGridLayout(topLayout, 3, 3)
        fieldBox = QVGroupBox('Fiel&ds', self)
        mainLayout.addMultiCellWidget(fieldBox, 0, 2, 0, 0)
        self.fieldListView = QListView(fieldBox)
        self.fieldListView.addColumn('Name')
        self.fieldListView.addColumn('Type')
        self.fieldListView.setSorting(-1)
        fieldFormatButton = QPushButton('Field Forma&t', fieldBox)
        self.connect(fieldFormatButton, SIGNAL('clicked()'), self.fieldFormat)

        xferLayout = QVBoxLayout(5)
        mainLayout.addLayout(xferLayout, 0, 1)
        self.addFieldButton = QPushButton('>>', self)
        self.addFieldButton.setMaximumWidth(self.addFieldButton.height())
        xferLayout.addWidget(self.addFieldButton)
        self.connect(self.addFieldButton, SIGNAL('clicked()'), self.addField)
        self.delFieldButton = QPushButton('<<', self)
        self.delFieldButton.setMaximumWidth(self.delFieldButton.height())
        xferLayout.addWidget(self.delFieldButton)
        self.connect(self.delFieldButton, SIGNAL('clicked()'), self.delField)

        headerFooterBox = QGroupBox('Header and Footer', self)
        mainLayout.addWidget(headerFooterBox, 0, 2)
        headerFooterLayout = QGridLayout(headerFooterBox, 6, 3, 10, 5)
        headerFooterLayout.addRowSpacing(0, self.fontMetrics().lineSpacing())
        headerFooterLayout.setRowStretch(0, 0)
        headerFooterLayout.addRowSpacing(5, self.fontMetrics().lineSpacing())
        headerFooterLayout.setRowStretch(5, 1)
        self.textEdits = []
        names = ['&Header Left', 'Header C&enter', 'Header &Right', \
                 '&Footer Left', 'Footer Ce&nter', 'Footer R&ight']
        for num, name in zip(range(6), names):
            if num < 3:
                row = 2
                col = num
            else:
                row = 4
                col = num - 3
            lineEdit = TitleEdit(headerFooterBox)
            headerFooterLayout.addWidget(QLabel(lineEdit, name, \
                                                headerFooterBox), \
                                         row - 1, col, \
                                         Qt.AlignLeft | Qt.AlignBottom)
            headerFooterLayout.setRowStretch(row - 1, 1)
            self.textEdits.append(lineEdit)
            headerFooterLayout.addWidget(lineEdit, row, col)
            self.connect(lineEdit, PYSIGNAL('cursorMove'), self.setButtonAvail)
            self.connect(lineEdit, SIGNAL('textChanged(const QString&)'), \
                         self.setButtonAvail)
            self.connect(lineEdit, PYSIGNAL('focusIn'), self.setCurrentEditor)

        ctrlLayout = QHBoxLayout(5)
        mainLayout.addMultiCellLayout(ctrlLayout, 2, 2, 1, 2)
        mainLayout.addRowSpacing(1, 30)
        mainLayout.setRowStretch(1, 2)
        ctrlLayout.insertStretch(0)
        okButton = QPushButton('&OK', self)
        ctrlLayout.addWidget(okButton)
        self.connect(okButton, SIGNAL('clicked()'), self, SLOT('accept()'))
        cancelButton = QPushButton('&Cancel', self)
        ctrlLayout.addWidget(cancelButton)
        self.connect(cancelButton, SIGNAL('clicked()'), self, SLOT('reject()'))

        self.loadFields()
        self.loadText()
        self.focusedEditor = self.textEdits[0]
        self.setButtonAvail()

    def setButtonAvail(self):
        """Update button availability"""
        self.delFieldButton.setEnabled(self.currentFieldPos() and 1 or 0)

    def setCurrentEditor(self, sender):
        """Set focusedEditor based on editor focus change signal"""
        self.focusedEditor = sender
        self.setButtonAvail()

    def loadFields(self, selNum=0):
        """Load list with field names"""
        self.fieldListView.clear()
        items = []
        fields = self.fileInfoFormat.fieldList[:]
        fields.reverse()
        for field in fields:
            items.insert(0, QListViewItem(self.fieldListView, field.name, \
                                          field.typeName))
        self.fieldListView.setSelected(items[selNum], 1)

    def loadText(self):
        """Load text into editors"""
        lines = self.fileInfoFormat.getLines()
        lines.extend([''] * (6 - len(lines)))
        for editor, line in zip(self.textEdits, lines):
            editor.blockSignals(1)
            editor.setText(line)
            editor.blockSignals(0)

    def getTextLines(self):
        """Return a list of six text lines of the header & footer"""
        return [unicode(editor.text()) for editor in self.textEdits]

    def addField(self):
        """Add selected field to cursor pos in active editor"""
        fieldName = str(self.fieldListView.selectedItem().text(0))
        text = u'{*!%s*}' % fieldName
        self.focusedEditor.deselect()
        self.focusedEditor.insert(text)
        self.focusedEditor.setFocus()

    def delField(self):
        """Remove field from cursor pos in editor"""
        start, end = self.currentFieldPos()
        text = unicode(self.focusedEditor.text())
        self.focusedEditor.setText(text[:start] + text[end:])
        self.focusedEditor.setCursorPosition(start)

    def currentFieldPos(self):
        """Return tuple of start, end for field at cursorPos in
           focusedEditor or None"""
        pattern = re.compile('{\*(.*?)\*}')
        text = unicode(self.focusedEditor.text())
        cursorPos = self.focusedEditor.cursorPosition()
        match = pattern.search(text)
        while match:
            if match.start() < cursorPos < match.end():
                return (match.start(), match.end())
            match = pattern.search(text, match.end())
        return None

    def fieldFormat(self):
        """Display the dialog for changing the field type settings"""
        fieldName = str(self.fieldListView.selectedItem().text(0))
        field = self.fileInfoFormat.findField(fieldName)
        dlg = FieldFormatDialog(field, 0, self, None, 1)
        if dlg.exec_loop() == QDialog.Accepted:
            field.format = dlg.format
            field.prefix = dlg.prefix
            field.suffix = dlg.suffix
            field.html = dlg.html

    def accept(self):
        """Set field lines after OK button is hit"""
        self.fileInfoFormat.lineList = []
        lines = self.getTextLines()
        for text, pos in zip(lines, range(len(lines))):
            if text:
                self.fileInfoFormat.insertLine(text, pos)
        QDialog.accept(self)
