#!/usr/bin/env python

#****************************************************************************
# treedoc.py, provides non-GUI base classes for document data
#
# TreeLine, an information storage program
# Copyright (C) 2004, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#****************************************************************************

import sys, os.path, stat, re, xml.sax, codecs, time
from gzip import GzipFile
if sys.platform != 'win32':
    import pwd
from xml.sax.saxutils import escape
from HTMLParser import HTMLParser
from treeformat import NodeFormat
from formatlist import FormatList
from fieldformat import LongTextFormat
from treeitem import TreeItem
from undo import UndoRedoStore
from gendate import GenDate
from gentime import GenTime
import docparam

escDict = {'"': '&quot;'}   # left out quotes

class TreeDoc:
    """Tree document class - stores root and has tree utilities"""
    def __init__(self, filePath=None, compressFile=0, defaultFieldAttrs={}):
        self.root = None
        docparam.nodeFormats = FormatList()
        # self.resetParam()
        self.fileName = ''
        if filePath:
            self.readFile(filePath)
        else:
            self.newFile(compressFile, defaultFieldAttrs)

    def resetParam(self):
        """Set doc setting values to defaults"""
        self.selectedNodes = []
        docparam.undoStore = UndoRedoStore()
        docparam.redoStore =  UndoRedoStore()
        docparam.modified = 0
        docparam.spaceBetween = 1
        docparam.formHtml = 1
        docparam.sortFields = ['']
        docparam.xslCssLink = ''
        self.xlstLink = ''
        self.createFileInfoItem()
        self.updateFileInfoItem()

    def createFileInfoItem(self):
        """Create new format & item to hold info about the current file"""
        formatName = 'INT_TL_FILE_DATA_FORM'
        format = NodeFormat(formatName)
        # docparam.nodeFormats.append(format) # don't add unless changed
        format.addNewField('File_Name')
        format.addNewField('File_Path')
        format.addNewField('File_Size', {'type': 'Number'})
        format.addNewField('File_Mod_Date', {'type': 'Date'})
        format.addNewField('File_Mod_Time', {'type': 'Time'})
        if sys.platform != 'win32':
            format.addNewField('File_Owner')
        format.addNewField('Page_Number')  # page info only for print header
        format.fieldList[-1].showInDialog = 0
        format.addNewField('Number_of_Pages')
        format.fieldList[-1].showInDialog = 0
        docparam.fileInfoItem = TreeItem(None, format)
        for field in format.fieldList:
            field.useFileInfo = 1

    def swapFileInfoFormat(self):
        """Copy the fileInfo field formats from the nodeFormats list to the
           doc's fileInfoItem fields and move it to the nodeFormats list"""
        docFormat = docparam.fileInfoItem.nodeFormat
        nodeFormat = docparam.nodeFormats.findFormat(docFormat.name)
        if not nodeFormat:
            return
        for docField, nodeField in zip(docFormat.fieldList, \
                                       nodeFormat.fieldList):
            docField.prefix = nodeField.prefix
            docField.suffix = nodeField.suffix
            docField.html = nodeField.html
            docField.format = nodeField.format
        docFormat.lineList = nodeFormat.lineList
        docparam.nodeFormats.remove(nodeFormat)
        docparam.nodeFormats.append(docFormat)

    def updateFileInfoItem(self):
        """Update contents of file info item"""
        try:
            status = os.stat(self.fileName)
        except:
            docparam.fileInfoItem.data = {}
            return
        docparam.fileInfoItem.data['File_Name'] = \
                              os.path.basename(self.fileName)
        docparam.fileInfoItem.data['File_Path'] = os.path.dirname(self.fileName)
        docparam.fileInfoItem.data['File_Size'] = str(status[stat.ST_SIZE])
        modTime = time.localtime(status[stat.ST_MTIME])
        docparam.fileInfoItem.data['File_Mod_Date'] = repr(GenDate(modTime))
        docparam.fileInfoItem.data['File_Mod_Time'] = repr(GenTime(modTime))
        if sys.platform != 'win32':
            docparam.fileInfoItem.data['File_Owner'] = \
                                  pwd.getpwuid(status[stat.ST_UID])[0]

    def getHeaderFooter(self, header=1):
        """Return formatted text for the header or the footer"""
        textLines = docparam.fileInfoItem.nodeFormat.\
                             formatAllTextLines(docparam.fileInfoItem)
        if header:
            textLines = textLines[:3]
        else:
            textLines = textLines[3:6]
        textLines.extend([''] * (3 - len(textLines)))
        aligns = ['left', 'center', 'right']
        result = ['<table width="100%"><tr>']
        count = 0
        for text, align in zip(textLines, aligns):
            if text:
                result.append('<td align="%s">%s</td>' % (align, text))
                count += 1
        result.append('</tr></table>')
        if count:
            return '\n'.join(result)
        return ''

    def getReadFileObj(self, filePath, errors='strict'):
        """Return file object and set self.compressFile to 0/1"""
        f = open(filePath, 'r')
        self.compressFile = f.read(2) == '\037\213'
        f.close()
        if self.compressFile:
            return codecs.EncodedFile(GzipFile(filePath, 'rb'), 'utf-8', \
                                      'utf-8', errors)
        return codecs.open(filePath, 'r', 'utf-8', errors)

    def getWriteFileObj(self, filePath, forceCompress):
        """Return write file object, compressed or not based on forceCompress,
           but always compress if has .gz extension"""
        if filePath.endswith('.gz') or forceCompress:
            return codecs.EncodedFile(GzipFile(filePath, 'wb'), 'utf-8')
        return codecs.open(filePath, 'w', 'utf-8')

    def newFile(self, compress=0, defaultFieldAttrs={}):
        """Start empty file"""
        self.compressFile = compress
        docparam.nodeFormats = FormatList([], 1, defaultFieldAttrs)
        self.root = TreeItem(None, docparam.nodeFormats[0])
        self.root.setTitle(u'Main')
        self.resetParam()
        self.fileName = ''

    def readFile(self, filePath, errors='strict'):
        """Open and read file - raise exception on failure"""
        origFileInfoItem = docparam.fileInfoItem
        self.createFileInfoItem()
        try:
            f = self.getReadFileObj(filePath, errors)
            handler = TreeSaxHandler()
            xml.sax.parse(f, handler)
        except IOError:
            print 'Error - could not read file', filePath
            docparam.fileInfoItem = origFileInfoItem
            raise
        except UnicodeError:
            print 'Warning - bad unicode characters were replaced'
            f.close()
            docparam.fileInfoItem = origFileInfoItem
            if errors == 'strict':
                errors = 'replace'
                self.readFile(filePath, errors)
            return
        except xml.sax.SAXException:
            f.close()
            docparam.fileInfoItem = origFileInfoItem
            raise ReadFileError('Could not open as treeline file')
        f.close()
        self.root = handler.rootItem
        self.fileName = filePath
        docparam.nodeFormats = FormatList(handler.formats.values())
        self.xlstLink = handler.xlstLink
        docparam.xslCssLink = handler.xslCssLink
        docparam.spaceBetween = handler.spaceBetween
        docparam.formHtml = handler.formHtml
        self.selectedNodes = []
        docparam.undoStore = UndoRedoStore()
        docparam.redoStore =  UndoRedoStore()
        docparam.modified = 0
        docparam.sortFields = ['']
        self.swapFileInfoFormat()
        self.updateFileInfoItem()

    def readTabbed(self, filePath, errors='strict'):
        """Import tabbed data into a flat tree - raise exception on failure"""
        try:
            f = self.getReadFileObj(filePath, errors)
            textList = f.readlines()
        except UnicodeError:
            print 'Warning - bad unicode characters were replaced'
            f.close()
            if errors == 'strict':
                errors = 'replace'
                self.readTabbed(filePath, errors)
            return
        f.close()
        bufList = [(text.count('\t', 0, len(text) - len(text.lstrip())), \
                    text.strip()) for text in textList if text.strip()]
        if bufList:
            buf = bufList.pop(0)
            if buf[0] == 0:
                nodeFormats = FormatList([], 1)  # set defaults ROOT & DEFAULT
                newRoot = TreeItem(None, nodeFormats[1])
                newRoot.setTitle(buf[1])
                if newRoot.loadTabbedChildren(bufList):
                    newRoot.nodeFormat = nodeFormats[0]
                    self.root = newRoot
                    self.fileName = filePath
                    docparam.nodeFormats = nodeFormats
                    self.resetParam()
                    return
        raise ReadFileError('Error in tabbed list')

    def readTable(self, filePath, errors='strict'):
        """Import table data into a flat tree - raise exception on failure"""
        try:
            f = self.getReadFileObj(filePath, errors)
            textList = f.readlines()
        except UnicodeError:
            print 'Warning - bad unicode characters were replaced'
            f.close()
            if errors == 'strict':
                errors = 'replace'
                self.readTable(filePath, errors)
            return
        f.close()
        nodeFormats = FormatList([], 1)  # set defaults ROOT & DEFAULT
        newRoot = TreeItem(None, nodeFormats[0])
        nodeFormats[1].fieldList = []
        nodeFormats[1].lineList = []
        nodeFormats[1].addTableFields(textList.pop(0).strip().split('\t'))
        newRoot.setTitle(u'Main')
        for line in textList:
            newItem = TreeItem(newRoot, nodeFormats[1])
            newRoot.childList.append(newItem)
            lineList = line.strip().split('\t')
            try:
                for num in range(len(lineList)):
                    newItem.data[nodeFormats[1].fieldList[num].name] \
                                = lineList[num].strip()
            except IndexError:
                docparam.modified = 0
                print 'Too few headings to read data as a table'
                raise ReadFileError('Too few headings to read data as a table')
        self.root = newRoot
        self.fileName = filePath
        docparam.nodeFormats = nodeFormats
        self.resetParam()

    def readLines(self, filePath, errors='strict'):
        """Import plain text, node per line"""
        try:
            f = self.getReadFileObj(filePath, errors)
            textList = f.readlines()
        except UnicodeError:
            print 'Warning - bad unicode characters were replaced'
            f.close()
            if errors == 'strict':
                errors = 'replace'
                self.readLines(filePath, errors)
            return
        f.close()
        nodeFormats = FormatList([], 1)  # set defaults ROOT & DEFAULT
        newRoot = TreeItem(None, nodeFormats[0])
        nodeFormats[1].fieldList = []
        nodeFormats[1].lineList = []
        nodeFormats[1].addTableFields(['TEXT'])
        newRoot.setTitle(u'Main')
        for line in textList:
            line = line.strip()
            if line:
                newItem = TreeItem(newRoot, nodeFormats[1])
                newRoot.childList.append(newItem)
                newItem.data['TEXT'] = line
        self.root = newRoot
        self.fileName = filePath
        docparam.nodeFormats = nodeFormats
        self.resetParam()

    def readPara(self, filePath, errors='strict'):
        """Import plain text, blank line delimitted"""
        try:
            f = self.getReadFileObj(filePath, errors)
            fullText = f.read().replace('\r', '')
        except UnicodeError:
            print 'Warning - bad unicode characters were replaced'
            f.close()
            if errors == 'strict':
                errors = 'replace'
                self.readPara(filePath, errors)
            return
        textList = fullText.split('\n\n')
        f.close()
        nodeFormats = FormatList([], 1)  # set defaults ROOT & DEFAULT
        newRoot = TreeItem(None, nodeFormats[0])
        nodeFormats[1].fieldList = []
        nodeFormats[1].lineList = []
        nodeFormats[1].addTableFields(['TEXT'])
        nodeFormats[1].fieldList[0].__class__ = LongTextFormat
        newRoot.setTitle(u'Main')
        for line in textList:
            line = line.strip()
            if line:
                newItem = TreeItem(newRoot, nodeFormats[1])
                newRoot.childList.append(newItem)
                newItem.data['TEXT'] = line
        self.root = newRoot
        self.fileName = filePath
        docparam.nodeFormats = nodeFormats
        self.resetParam()

    def readTreepad(self, filePath, errors='strict'):
        """Read Treepad text-node file"""
        try:
            f = self.getReadFileObj(filePath, errors)
            textList = f.read().split('<end node> 5P9i0s8y19Z')
            f.close()
        except UnicodeError:  # error common - broken unicode on windows
            print 'Warning - bad unicode characters were replaced'
            f.close()
            if errors == 'strict':
                errors = 'replace'
                self.readTreepad(filePath, errors)
            return
        nodeFormats = FormatList()
        format = NodeFormat('DEFAULT')
        format.addNewField('Title')
        format.addLine(u'{*Title*}')
        format.addNewField('Text', {'type': 'LongText'})
        format.addLine(u'{*Text*}')
        nodeFormats.append(format)
        itemList = []
        for text in textList:
            text = text.strip()
            if text:
                try:
                    text = text.split('<node>', 1)[1].lstrip()
                    lines = text.split('\n')
                    title = lines[0]
                    level = int(lines[1])
                    lines = lines[2:]
                except (ValueError, IndexError):
                    print 'Error - bad file format in %s' % fileName
                    raise ReadFileError('Error - bad file format in %s' % \
                                        fileName)
                item = TreeItem(None, format)
                item.data['Title'] = title
                item.data['Text'] = '\n'.join(lines)
                item.level = level
                itemList.append(item)
        self.root = itemList[0]
        parentList = []
        for item in itemList:
            if item.level != 0:
                parentList = parentList[:item.level]
                item.parent = parentList[-1]
                parentList[-1].childList.append(item)
            parentList.append(item)
        self.root = itemList[0]
        self.fileName = filePath
        docparam.nodeFormats = nodeFormats
        self.resetParam()

    def createBookmarkFormat(self):
        """Return a set of formats for bookmark imports"""
        nodeFormats = FormatList()
        format = NodeFormat('FOLDER')
        format.addNewField('Name')
        format.addLine(u'{*Name*}')
        format.addLine(u'{*Name*}')
        nodeFormats.append(format)
        format = NodeFormat('BOOKMARK')
        format.addNewField('Name')
        format.addLine(u'{*Name*}')
        format.addLine(u'{*Name*}')
        format.addNewField('Link', {'type': 'URL'})
        format.addLine(u'{*Link*}')
        nodeFormats.append(format)
        format = NodeFormat('SEPARATOR')
        format.addLine(u'------------------')
        format.addLine(u'<hr>')
        nodeFormats.append(format)
        return nodeFormats

    def readXbel(self, filePath, errors='strict'):
        """Read XBEL format bookmarks"""
        nodeFormats = self.createBookmarkFormat()
        try:
            f = self.getReadFileObj(filePath, errors)
            handler = XbelSaxHandler(nodeFormats)
            xml.sax.parse(f, handler)
        except UnicodeError:
            print 'Warning - bad unicode characters were replaced'
            f.close()
            if errors == 'strict':
                errors = 'replace'
                self.readXbel(filePath, errors)
            return
        except xml.sax.SAXException:
            f.close()
            raise ReadFileError('Could not open as XBEL file')
        f.close()
        if not handler.rootItem:
            raise ReadFileError('Could not open as XBEL file')
        self.root = handler.rootItem
        if not self.root.data.get('Name', ''):
            self.root.data['Name'] = 'Bookmarks'
        self.fileName = filePath
        docparam.nodeFormats = nodeFormats
        self.resetParam()

    def readMozilla(self, filePath, errors='strict'):
        """Read Mozilla HTML format bookmarks"""
        nodeFormats = self.createBookmarkFormat()
        try:
            f = self.getReadFileObj(filePath, errors)
            fullText = f.read()
        except UnicodeError:
            print 'Warning - bad unicode characters were replaced'
            f.close()
            if errors == 'strict':
                errors = 'replace'
                self.readMozilla(filePath, errors)
            return
        try:
            handler = HtmlBookmarkHandler(nodeFormats)
            handler.feed(fullText)
            handler.close()
        except ReadFileError:
            raise ReadFileError('Could not open as HTML bookmark file')
        if not handler.rootItem:
            raise ReadFileError('Could not open as XBEL file')
        self.root = handler.rootItem
        if not self.root.data.get('Name', ''):
            self.root.data['Name'] = 'Bookmarks'
        self.fileName = filePath
        docparam.nodeFormats = nodeFormats
        self.resetParam()

    def readXmlString(self, string, mergeFormat=0):
        """Read xml string and return top item or None"""
        try:
            handler = TreeSaxHandler()
            xml.sax.parseString(string.encode('utf-8'), handler)
        except xml.sax.SAXException:
            return None
        if mergeFormat:
            for val in handler.formats.values():
                docparam.nodeFormats.addIfMissing(val)
            for item in handler.rootItem.descendantList(1):
                item.nodeFormat = docparam.nodeFormats.\
                                       findFormat(item.nodeFormat.name)
        return handler.rootItem

    def writeFile(self, filePath, updateInfo=1):
        """Write file - returns 0/1 on fail/success"""
        lines = [u'<?xml version="1.0" encoding="utf-8" ?>']
        if self.xlstLink:
            lines.append(u'<?%s?>' % self.xlstLink)
        lines.extend(self.root.branchXml([], 1))
        try:
            f = self.getWriteFileObj(filePath, self.compressFile)
            f.writelines([line + '\n' for line in lines])
        except IOError:
            print 'Error - could not write file', filePath
            return 0
        f.close()
        if filePath.endswith('.gz'):
            self.compressFile = 1
        if updateInfo:
            docparam.modified = 0
            self.fileName = filePath
            self.updateFileInfoItem()
        return 1

    def exportHtml(self, filePath, item, includeRoot, openOnly=0, indent=20, \
                   addHeader=0):
        """Save contents of outGroup as html to file w/o columns,
           returns 1 on success"""
        outGroup = item.outputItemList(includeRoot, openOnly)
        return self.exportHtmlColumns(filePath, outGroup, 1, indent, addHeader)

    def exportHtmlColumns(self, filePath, outGroup, numCol=1, indent=20, \
                          addHeader=0):
        """Save contents of outGroup as html to file in columns,
           returns 1 on success"""
        outGroup.addBreaks()
        outGroups = outGroup.splitColumns(numCol)
        for group in outGroups:
            group.addPrefix()
            group.addIndents()
        htmlTitle = os.path.splitext(os.path.basename(filePath))[0]
        lines = [u'<html>', u'<meta http-equiv="Content-Type" ' \
                 'content="text/html; charset=utf-8">', u'<head>', \
                 u'<title>%s</title>' % htmlTitle, \
                 u'<style type="text/css"><!--', u'div {margin-left: %dpx}' \
                 % indent, u'td {padding: 10px}', \
                 u'tr {vertical-align: top}', u'--></style>', \
                 u'</head>', u'<body>']
        if addHeader:
            header = self.getHeaderFooter(1)
            if header:
                lines.append(header)
        lines.extend([u'<table>', u'<tr><td>'])
        for item in outGroups[0]:
            lines.extend(item.textLines)
        for group in outGroups[1:]:
            lines.append(u'</td><td>')
            for item in group:
                lines.extend(item.textLines)
        lines.extend([u'</td></tr>', u'</table>'])
        if addHeader:
            footer = self.getHeaderFooter(0)
            if footer:
                lines.append(footer)
        lines.extend([u'</body>', u'</html>'])
        try:
            f = self.getWriteFileObj(filePath, 0)
            f.writelines([line + '\n' for line in lines])
        except IOError:
            print 'Error - could not write file', filePath
            return 0
        f.close()
        return 1

    def exportDir(self, dirName, item, addHeader=0):
        """Write tree to nested directory struct with html tables"""
        oldDir = os.getcwd()
        os.chdir(str(dirName))
        if addHeader:
            header = self.getHeaderFooter(1)
            footer = self.getHeaderFooter(0)
            item.exportDir(None, header, footer)
        else:
            item.exportDir()
        os.chdir(oldDir)

    def exportXslt(self, filePath, includeRoot, indent=20):
        """Write XSLT file and add link in treeline file"""
        title = os.path.splitext(os.path.basename(filePath))[0]
        lines = [u'<xsl:stylesheet version="1.0" '\
                 u'xmlns:xsl="http://www.w3.org/1999/XSL/Transform">', \
                 u"<xsl:output method='html'/>", u'<xsl:template match="/">', \
                 u'<html>', u'<head>']
        if docparam.xslCssLink:
            lines.append('<link rel="stylesheet" href="%s" type="text/css"/>' \
                         % docparam.xslCssLink)
        lines.extend([u'<title>%s</title>' % title, u'</head>', u'<body>', \
                     u'<xsl:apply-templates/>', u'</body>', u'</html>', \
                     u'</xsl:template>'])
        if not includeRoot:
            lines.extend([u'', u'<xsl:template match="/%s">' \
                          % self.root.nodeFormat.name, \
                          u'<xsl:apply-templates/>', u'</xsl:template>'])
        docparam.nodeFormats.sort()
        for format in docparam.nodeFormats:
            lines.extend(format.xsltTemplate(indent))
        lines.extend([u'', u'<xsl:template match="*" />', u'', \
                      u'</xsl:stylesheet>'])
        try:
            f = self.getWriteFileObj(filePath, 0)
            f.writelines([line + '\n' for line in lines])
        except IOError:
            print 'Error - could not write file', filePath
            return 0
        f.close()
        link = u'xml-stylesheet type="text/xsl" href="%s"' \
               % os.path.basename(filePath)
        if self.xlstLink != link:
            self.xlstLink = link
            docparam.modified = 1
        return 1

    def exportTrlSubtree(self, filePath, item):
        """Write subtree TRL file starting form item"""
        lines = [u'<?xml version="1.0" encoding="utf-8" ?>']
        if self.xlstLink:
            lines.append(u'<?%s?>' % self.xlstLink)
        lines.extend(item.branchXml([], 1))
        try:
            f = self.getWriteFileObj(filePath, self.compressFile)
            f.writelines([line + '\n' for line in lines])
        except IOError:
            print 'Error - could not write file', filePath
            return 0
        f.close()
        return 1

    def exportTable(self, filePath, item):
        """Write data to table for children of item"""
        typeList = []
        headings = []
        tableList = []
        for child in item.childList:
            if child.nodeFormat not in typeList:
                for field in child.nodeFormat.fieldNames():
                    if field not in headings:
                        headings.append(field)
                typeList.append(child.nodeFormat)
            tableList.append(u'\t'.join([child.data.get(head, '') for \
                                        head in headings]))
        tableList.insert(0, u'\t'.join([head for head in headings]))
        try:
            f = self.getWriteFileObj(filePath, 0)
            f.writelines([line + os.linesep for line in tableList])
        except IOError:
            print 'Error - could not write file', filePath
            return 0
        f.close()
        return 1

    def exportTabbedTitles(self, filePath, item, includeRoot, openOnly=0):
        """Write tabbed titles for descendants of item"""
        if includeRoot:
            titleList = item.exportToText(0, openOnly)
        else:
            titleList = item.exportToText(-1, openOnly)
            del titleList[0]
        try:
            f = self.getWriteFileObj(filePath, 0)
            f.writelines([line + os.linesep for line in titleList])
        except IOError:
            print 'Error - could not write file', filePath
            return 0
        f.close()
        return 1

    def exportXbel(self, filePath, item):
        """Export XBEL bookmarks"""
        lines = [u'<!DOCTYPE xbel>']
        lines.extend(item.exportXbelBookmarks())
        try:
            f = self.getWriteFileObj(filePath, 0)
            f.writelines([line + '\n' for line in lines])
        except IOError:
            print 'Error - could not write file', filePath
            return 0
        f.close()
        return 1

    def exportHtmlBookmarks(self, filePath, item):
        """Export HTML bookmarks"""
        lines = [u'<!DOCTYPE NETSCAPE-Bookmark-file-1>', \
                 u'<META HTTP-EQUIV="Content-Type" CONTENT="text/html; '\
                  'charset=UTF-8">']
        lines.extend(item.exportHtmlBookmarks())
        try:
            f = self.getWriteFileObj(filePath, 0)
            f.writelines([line + '\n' for line in lines])
        except IOError:
            print 'Error - could not write file', filePath
            return 0
        f.close()
        return 1

    def updateSelection(self, item, select):
        """Add or remove item from selection list"""
        if select:
            self.selectedNodes.append(item)
        else:
            try:
                self.selectedNodes.remove(item)
            except ValueError:
                pass

    def findText(self, startItem, searchStr, forward=1):
        """Return item containing words in searchStr in any field or None"""
        wordList = [text.lower() for text in searchStr.split()]
        fullList = self.root.descendantList(1)
        currentPos = fullList.index(startItem)
        fullList = fullList[currentPos+1:] + fullList[:currentPos]
        if not forward:
            fullList.reverse()
        for item in fullList:
            if item.matchWords(wordList):
                return item
        return None

    def findTitleText(self, startItem, searchStr):
        """Return item containing searchStr in the title or None, 
           include startItem in search"""
        searchStr = searchStr.lower()
        next = startItem
        while 1:
            if next.title().lower().find(searchStr) >= 0:
                return next
            next = next.nextItem(1)
            if not next:
                next = self.root
            if next is startItem:
                return None

    def findNextTitle(self, startItem, searchStr, forward=1):
        """Return item containing searchStr in the title or None, 
           do not include startItem in search"""
        searchStr = searchStr.lower()
        next = startItem
        while 1:
            if forward:
                next = next.nextItem(1)
                if not next:
                    next = self.root
            else:
                next = next.prevItem(1)
                if not next:
                    next = self.root.descendantList()[-1]
            if next is startItem:
                return None
            if next.title().lower().find(searchStr) >= 0:
                return next

    def cmpItems(self, item1, item2):
        """Compare function for sorting, not case sensitive"""
        for field in docparam.sortFields:
            result = cmp(item1.data.get(field, '').upper(), \
                         item2.data.get(field, '').upper())
            if result != 0:
                return result
        return 0


class ReadFileError(Exception):
    """Exception class for errors on reading file content"""
    pass


class TreeSaxHandler(xml.sax.ContentHandler):
    """Handler to read xml thru sax"""
    lineAttrRe = re.compile('line\d+$')
    def __init__(self):
        xml.sax.ContentHandler.__init__(self)
        self.currentItem = None
        self.bareFormat = None
        self.rootItem = None
        self.formats = {}
        self.dataEntry = ''
        self.text = ''
        self.xlstLink = ''
        self.xslCssLink = ''
        self.spaceBetween = 1
        self.formHtml = 1

    def startElement(self, name, attrs):
        """Called by the reader at the open tag of each element"""
        if attrs.get(u'item', ''):  # reading TreeItem or bare format
            format = self.formats.get(name, None)
            if not format:
                format = NodeFormat(name)
                self.formats[name] = format
            for key in attrs.keys():
                if TreeSaxHandler.lineAttrRe.match(key):
                    format.insertLine(attrs.get(key, ''), int(key[4:]))
            childType = attrs.get(u'childtype', None)
            if childType:
                format.childType = childType
            sibPrefix = attrs.get(u'sibprefix', None)
            if sibPrefix:
                format.sibPrefix = sibPrefix
            sibSuffix = attrs.get(u'sibsuffix', None)
            if sibSuffix:
                format.sibSuffix = sibSuffix

            if attrs.get(u'item', '') == 'y':  # reading TreeItem
                newItem = TreeItem(self.currentItem, format)
                if self.currentItem:
                    self.currentItem.childList.append(newItem)
                else:
                    self.rootItem = newItem
                    if attrs.get(u'nospace', '').startswith('y'):
                        self.spaceBetween = 0
                    if attrs.get(u'nohtml', '').startswith('y'):
                        self.formHtml = 0
                    self.xslCssLink = attrs.get(u'xslcss', '')
                self.currentItem = newItem
            else:                    # reading bare format
                self.bareFormat = format

        else:                        # reading data
            self.text = ''
            self.dataEntry = name
            if not self.currentItem:
                raise xml.sax.SAXException, 'No valid item'
            currentFormat = self.bareFormat
            if not currentFormat:
                currentFormat = self.currentItem.nodeFormat
            if name not in currentFormat.fieldNames(): # add new field to format
                currentFormat.addNewField(name, attrs)

    def endElement(self, name):
        """Called by the reader at the close tag of each element"""
        if not self.currentItem:
            raise xml.sax.SAXException, 'No valid item'
        if self.dataEntry == name:   # finish data
            # self.simplifyText()
            if self.text:
                self.currentItem.data[name] = self.text
            self.dataEntry = ''
        else:                         # finish TreeItem
            if self.bareFormat:
                assert(self.bareFormat.name == name)
                self.bareFormat = None
            else:
                assert(self.currentItem.nodeFormat.name == name)
                self.currentItem = self.currentItem.parent

    def characters(self, content):
        """Called by the reader to process text"""
        self.text += content

    def simplifyText(self):
        """Remove extra whitespace from text"""
        self.text = u' '.join(self.text.split())

    def processingInstruction(self, target, data):
        """Recover xlst link from stylesheet instruction"""
        self.xlstLink = u' '.join((target, data))


class XbelSaxHandler(xml.sax.ContentHandler):
    """Handler to parse XBEL bookmark format"""
    def __init__(self, nodeFormats):
        xml.sax.ContentHandler.__init__(self)
        self.folderFormat = nodeFormats.findFormat('FOLDER')
        self.bookmarkFormat = nodeFormats.findFormat('BOOKMARK')
        self.separatorFormat = nodeFormats.findFormat('SEPARATOR')
        self.currentItem = None
        self.rootItem = None
        self.text = ''

    def startElement(self, name, attrs):
        """Called by the reader at the open tag of each element"""
        if name == 'folder' or name == 'xbel':
            newItem = TreeItem(self.currentItem, self.folderFormat)
            if self.currentItem:
                self.currentItem.childList.append(newItem)
            else:
                self.rootItem = newItem
            if attrs.get(u'folded', '') == 'no':
                newItem.open = 1
            self.currentItem = newItem
        elif name == 'bookmark':
            newItem = TreeItem(self.currentItem, self.bookmarkFormat)
            if self.currentItem:
                self.currentItem.childList.append(newItem)
            else:
                raise xml.sax.SAXException, 'No valid parent folder'
            newItem.data['Link'] = attrs.get(u'href', '')
            self.currentItem = newItem
        elif name == 'title':
            self.text = ''
        elif name == 'separator':
            newItem = TreeItem(self.currentItem, self.separatorFormat)
            if self.currentItem:
                self.currentItem.childList.append(newItem)
            else:
                raise xml.sax.SAXException, 'No valid parent folder'
            self.currentItem = newItem
        else:   # unsupported tags
            pass

    def endElement(self, name):
        """Called by the reader at the close tag of each element"""
        if not self.currentItem:
            raise xml.sax.SAXException, 'No valid item'
        if name in ('folder', 'xbel', 'bookmark', 'separator'):
            self.currentItem = self.currentItem.parent
        elif name == 'title':
            self.currentItem.data['Name'] = u' '.join(self.text.split())
        else:   # unsupported tags
            pass

    def characters(self, content):
        """Called by the reader to process text"""
        self.text += content


class HtmlBookmarkHandler(HTMLParser):
    """Handler to parse XBEL bookmark format"""
    escDict = {'amp': '&', 'lt': '<', 'gt': '>', 'quot': '"'}
    def __init__(self, nodeFormats):
        HTMLParser.__init__(self)
        self.folderFormat = nodeFormats.findFormat('FOLDER')
        self.bookmarkFormat = nodeFormats.findFormat('BOOKMARK')
        self.separatorFormat = nodeFormats.findFormat('SEPARATOR')
        self.rootItem = TreeItem(None, self.folderFormat)
        self.rootItem.data['Name'] = u'Bookmarks'
        self.currentItem = self.rootItem
        self.currentParent = None
        self.text = ''

    def handle_starttag(self, tag, attrs):
        """Called by the reader at each open tag"""
        if tag == 'dt' or tag == 'h1':      # start any entry
            self.text = ''
        elif tag == 'dl':    # start indent
            self.currentParent = self.currentItem
            self.currentItem = None
        elif tag == 'h3':    # start folder
            if not self.currentParent:
                raise ReadFileError, 'No valid parent folder'
            self.currentItem = TreeItem(self.currentParent, self.folderFormat)
            self.currentParent.childList.append(self.currentItem)
        elif tag == 'a':     # start link
            if not self.currentParent:
                raise ReadFileError, 'No valid parent folder'
            self.currentItem = TreeItem(self.currentParent, self.bookmarkFormat)
            self.currentParent.childList.append(self.currentItem)
            for name, value in attrs:
                if name == 'href':
                    self.currentItem.data['Link'] = value
        elif tag == 'hr':     # separator
            if not self.currentParent:
                raise ReadFileError, 'No valid parent folder'
            item = TreeItem(self.currentParent, self.separatorFormat)
            self.currentParent.childList.append(item)
            self.currentItem = None

    def handle_endtag(self, tag):
        """Called by the reader at each end tag"""
        if tag == 'dl':      # end indented section
            self.currentParent = self.currentParent.parent
            self.currentItem = None
        elif tag == 'h3' or tag == 'a':    # end folder or link
            if not self.currentItem:
                raise ReadFileError, 'No valid item'
            self.currentItem.data['Name'] = self.text
        elif tag == 'h1':    # end main title
            self.rootItem.data['Name'] = self.text

    def handle_data(self, data):
        """Called by the reader to process text"""
        self.text += data

    def handle_entityref(self, name):
        """Convert escaped entity ref to char"""
        self.text += HtmlBookmarkHandler.escDict.get(name, '')


def testXmlParser():
    """Return 1 if parser works correctly"""
    try:
        handler = xml.sax.ContentHandler()
        xml.sax.parseString('<XML>test</XML>', handler)
    except xml.sax.SAXException:
        return 0
    return 1


if __name__ == '__main__':
    doc = TreeDoc()
    if len(sys.argv) > 1:
        doc.readFile(sys.argv[1])
    print '\n'.join(doc.root.exportToText())
    print
    print '\n'.join(doc.root.childList[0].formatChildText())
