#!/usr/bin/env python

#****************************************************************************
# treeformat.py, provides non-GUI base classes for node formating info
#
# TreeLine, an information storage program
# Copyright (C) 2004, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#****************************************************************************

import re, copy, types
from xml.sax.saxutils import escape
from fieldformat import TextFormat, LongTextFormat, NumberFormat, \
                        ChoiceFormat, CombinationFormat, \
                        DateFormat, TimeFormat, BooleanFormat, \
                        URLFormat, PathFormat, EmailFormat, PictureFormat, \
                        ParentFormat, AncestorFormat
import treedoc
import docparam

types.StringTypes = (types.StringType, types.UnicodeType)  # Python <2.2 compat

class NodeFormat:
    """Stores node type data field list and line formatting"""
    fieldSplitRe = re.compile('({\*(?:\**|\?|!)\w+\*})')
    fieldPartRe = re.compile('{\*(\**|\?|!)(\w+)\*}')
    def __init__(self, name, defaultField='', defaultFieldAttrs={}):
        self.name = name
        self.fieldList = []
        self.lineList = []
        self.numEmptyFields = 0
        self.numFullFields = 0
        self.childType = ''
        self.sibPrefix = ''
        self.sibSuffix = ''
        if defaultField:
            self.addNewField(defaultField, defaultFieldAttrs)
            self.addLine(u'{*%s*}' % defaultField)
            self.addLine(u'{*%s*}' % defaultField)

    def __cmp__(self, other):
        """Comparison for equality and sorting"""
        return cmp(self.name, other.name)

    def fieldNames(self):
        """Return list of names of fields"""
        return [field.name for field in self.fieldList]

    def lineFields(self):
        """Return list of first fieldname in each line"""
        result = []
        for line in self.lineList[1:]:
            strippedLine = [part for part in line \
                            if type(part) not in types.StringTypes]
            result.append(strippedLine and strippedLine[0].name or '')
        return result

    def findField(self, name):
        """Return field matching name or None"""
        try:
            return self.fieldList[self.fieldNames().index(name)]
        except ValueError:
            return None

    def addNewField(self, name, attrs={}):
        """Add new field with type, format, extra text in attrs dict"""
        field = eval(attrs.get(u'type', 'Text') + 'Format')(name, attrs)
        self.fieldList.append(field)
        sepName = field.sepName()
        # replace field name with the field
        self.lineList = [[part == sepName and field or part for part in line] \
                         for line in self.lineList]

    def addFieldIfNew(self, name, attrs={}):
        """"Add new field if it doesn't exist"""
        if name not in self.fieldNames():
            self.addNewField(name, attrs)

    def addLine(self, text):
        """Add format line to end of lineList, line 0 is the title"""
        newLine = self.parseLine(text)
        if newLine:
            self.lineList.append(newLine)

    def insertLine(self, text, pos):
        """Change line at pos of lineList,
           add blanks in earlier positions if req'd, line 0 is the title"""
        newLine = self.parseLine(text)
        if not newLine:
            newLine = ['']
        self.lineList.extend([''] * (pos - len(self.lineList) + 1))
        self.lineList[pos] = newLine

    def changeTitleLine(self, text):
        """Change the title format line"""
        self.insertLine(text, 0)

    def updateLineFields(self):
        """Re-find fields to update for any changes in the fieldList"""
        self.lineList = [self.parseLine(line) for line in self.getLines()]

    def parseLine(self, text):
        """Parse text format line, return list of field types and text"""
        text = u' '.join(text.split())
        lineList = filter(None, NodeFormat.fieldSplitRe.split(text))
        newLine = []
        if lineList:
            for part in lineList:
                fieldMatch = NodeFormat.fieldPartRe.match(part)
                if fieldMatch:
                    modifier = fieldMatch.group(1)
                    fieldName = fieldMatch.group(2)
                    if not modifier:
                        field = self.findField(fieldName)
                        newLine.append(field and field or part)
                    elif modifier[0] == '*':
                        newLine.append(ParentFormat(fieldName, len(modifier)))
                    elif modifier == '?':
                        newLine.append(AncestorFormat(fieldName))
                    else:  # == '!' file field
                        field = docparam.fileInfoItem.nodeFormat.\
                                   findField(fieldName)
                        newLine.append(field and field or part)
                else:
                    newLine.append(part)
        return newLine

    def equalPrefix(self, other):
        """Return 1 if prefix and suffix are equivalent"""
        if self is other:
            return 1
        return self.sibPrefix.strip().upper() == \
               other.sibPrefix.strip().upper() and \
               self.sibSuffix.strip().upper() == \
               other.sibSuffix.strip().upper()

    def formatTitle(self, item):
        """Return string with formatted title data"""
        line = u''.join([self.fieldText(text, item, 1) for text in \
                         self.lineList[0]])
        return line.strip()

    def formatText(self, item, skipEmpty=1, addPrefix=0, addSuffix=0):
        """Return list of output strings from formatting data"""
        tagExp = re.compile('.*(<br[ /]*?>|<BR[ /]*?>|<hr[ /]*?>|<HR[ /]*?>)$')
        result = []
        for lineData in self.lineList[1:]:
            self.numEmptyFields = 0
            self.numFullFields = 0
            line = u''.join([self.fieldText(text, item, 0) for text \
                             in lineData])
            if self.numFullFields or not self.numEmptyFields or not skipEmpty:
                result.append(line.strip())   # add if tags are not empty
            else:
                tagMatch = tagExp.match(line)
                if tagMatch and result and docparam.formHtml:
                    # html tag at removed line end
                    result[-1] += tagMatch.group(1) # add to prev
        if addPrefix and self.sibPrefix:
            if result:
                result[0] = self.sibPrefix + result[0]
            else:
                result = [self.sibPrefix]
        if addSuffix and self.sibSuffix:
            if result:
                result[-1] += self.sibSuffix
            else:
                result = [self.sibSuffix]
        return result

    def formatAllTextLines(self, item):
        """Return a list of all text lines (title & output), add an empty
           string if a line has empty tags"""
        result = []
        for lineData in self.lineList:
            self.numEmptyFields = 0
            self.numFullFields = 0
            line = u''.join([self.fieldText(text, item, 0) for text \
                             in lineData])
            if self.numFullFields or not self.numEmptyFields:
                result.append(line.strip())   # add if tags are not empty
            else:
                result.append('')
        return result

    def fieldText(self, field, item, titleMode=0):
        """Return formatted text for field"""
        if type(field) in types.StringTypes:
            text = field
            if not titleMode and not docparam.formHtml:
                text = escape(text)
        else:
            if field.parentLevel:
                for num in range(field.parentLevel):
                    item = item.parent
                    if not item:
                        self.numEmptyFields += 1
                        return ''
                if field.parentLevel < 0:
                    item = item.parent
                    while item and not item.nodeFormat.findField(field.name):
                        item = item.parent
                    if not item:
                        self.numEmptyFields += 1
                        return ''
                field = item.nodeFormat.findField(field.name)
                if not field:
                    self.numEmptyFields += 1
                    return ''
            elif field.useFileInfo:
                item = docparam.fileInfoItem
            text = item.data.get(field.name, '')
            if text:
                text = field.outputText(text, titleMode)
                self.numFullFields += 1
            else:
                self.numEmptyFields += 1
        return text

    def setTitle(self, title, item):
        """Set data based on title string, return 1 if changed successfully"""
        fields = []
        pattern = u''
        extraText = u''
        for seg in self.lineList[0]:
            if type(seg) in types.StringTypes:
                pattern += re.escape(seg)
                extraText += seg
            elif seg.parentLevel:
                text = self.fieldText(seg, item, 1)
                pattern += re.escape(text)
                extraText += text
            else:
                fields.append(seg)
                pattern += '(.*)'
        match = re.match(pattern, title)
        if match:
            for num in range(len(fields)):
                item.data[fields[num].name] = match.group(num+1)
            return 1
        if not extraText.strip():  # assign to 1st field if sep is only spaces
            item.data[fields[0].name] = title
            for field in fields[1:]:
                item.data[field.name] = ''
            return 1
        return 0

    def formatXml(self):
        """Return text list for use in xml file"""
        result = []
        lines = self.getLines()
        for line, i in zip(lines, range(len(lines))):
            if line:
                result.append(u'line%d="%s"' % \
                              (i, escape(line, treedoc.escDict)))
        if self.childType:
            result.append(u'childtype="%s"' % self.childType)
        if self.sibPrefix:
            result.append(u'sibprefix="%s"' % escape(self.sibPrefix, \
                                                     treedoc.escDict))
        if self.sibSuffix:
            result.append(u'sibsuffix="%s"' % escape(self.sibSuffix, \
                                                     treedoc.escDict))
        return result

    def getLines(self):
        """Return text list of formatting lines"""
        return [u''.join([self.fieldName(part) for part in line]) \
                         for line in self.lineList]

    def fieldName(self, field):
        """Return field name with separators or line text part"""
        if type(field) in types.StringTypes:
            return field
        return field.sepName()

    def addTableFields(self, headingList):
        """Set fields based on import table headings"""
        headings = [self.correctFieldName(head) for head in headingList]
        self.addLine(u'{*%s*}' % headings[0])
        for text in headings:
            self.addNewField(text)
            self.addLine(u'{*%s*}' % text)

    def xsltTemplate(self, indent):
        """Return list of lines for xslt template"""
        tagExp = re.compile('.*(<br[ /]*>|<BR[ /]*>|<hr[ /]*>|<HR[ /]*>)$')
        brExp = re.compile('<[bB][rR][ /]*>')
        brFormat = u'<br />'
        hrExp = re.compile('<[hH][rR][ /]*>')
        hrFormat = u'<hr />'
        lineList = copy.deepcopy(self.lineList[1:])
        output = [u'', u'<xsl:template match="%s[@item=\'y\']">' % self.name]
        if self.sibPrefix:
            output.append(self.sibPrefix)
        for lineDataRaw in lineList:
            lineData = [(type(field) not in types.StringTypes) and field or \
                        hrFormat.join(hrExp.split(brFormat.\
                                                  join(brExp.split(field)))) \
                        for field in lineDataRaw]
            fields = [field for field in lineData if \
                      type(field) not in types.StringTypes]
            if fields:
                text = u'<xsl:if test="%s">' % \
                        ' or '.join([field.xslTestText() for field in fields])
                output.append(text)
                tagMatch = tagExp.match(type(lineData[-1]) in \
                                        types.StringTypes and \
                                        lineData[-1] or '')
                if tagMatch:
                    if len(tagMatch.group(1)) == len(lineData[-1]):
                        del lineData[-1]
                    else:
                        lineData[-1] = lineData[-1][:-len(tagMatch.group(1))]
                line = u''.join([(type(text) in types.StringTypes) \
                                 and (text==' ' and '<xsl:text> </xsl:text>' \
                                      or text) \
                                 or text.xslText() for text in lineData])
                output.extend([line + brFormat, u'</xsl:if>'])
                if tagMatch:
                    output.append(tagMatch.group(1))
            else:
                output.append(lineData[0] + brFormat)
        if self.sibSuffix:
            output.append(self.sibSuffix)
        if docparam.spaceBetween:
            output.append(u'<br/>')
        output.extend([u'<div style="margin-left:%ipx">' % indent, \
                       u'<xsl:apply-templates/>', u'</div>', \
                       u'</xsl:template>'])
        return output

    def correctFieldName(self, name):
        """Return name with only alphanumerics and underscores allowed"""
        return u''.join([char.isalnum() and char or '_' for \
                        char in name.strip()])

    def removeField(self, field):
        """Remove all occurances of field from lines
           return 0 if not found"""
        cnt = 0
        for lineData in self.lineList:
            while field in lineData:
                lineData.remove(field)
                cnt += 1
        self.lineList = filter(None, self.lineList)
        while len(self.lineList) < 2:
            self.lineList.append([''])
        return cnt

    def findLinkField(self):
        """Return most likely field containing a bookmark URL"""
        availFields = [field for field in self.fieldList if \
                       field.typeName == 'URL']
        if len(availFields) == 1:
            return availFields[0]
        if not availFields:
            return None
        for srchTerm in ['link', 'url', 'href']:
            for field in availFields:
                name = field.name.lower()
                if field.name.lower() == srchTerm:
                    return field
            for field in availFields:
                if field.name.lower().find(srchTerm) >= 0:
                    return field
        return availFields[0]
