#!/usr/bin/env python

#****************************************************************************
# treeitem.py, provides non-GUI base classes for tree items
#
# TreeLine, an information storage program
# Copyright (C) 2004, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#****************************************************************************

import copy, numbering, os, codecs
from xml.sax.saxutils import escape
from treeformat import NodeFormat
from output import OutputGroup, OutputItem
from formatlist import FormatList
import treedoc
import docparam

class TreeItem:
    """Data storage item for tree structure"""
    def __init__(self, parent, nodeFormat, initText=''):
        self.parent = parent
        self.nodeFormat = nodeFormat
        self.data = {}
        if initText:
            self.setTitle(initText)
        self.open = 0
        self.childList = []
        self.viewData = None
        self.level = 0    # updated only by numbered descend list

    def title(self):
        """Return title text"""
        text = self.nodeFormat.formatTitle(self)
        if not text:
            text = u'[BLANK TITLE]'
        return text

    def setTitle(self, title):
        """Set title text, return 1 if changed successfully"""
        title = u' '.join(title.split())
        if title:
            if self.nodeFormat.setTitle(title, self):
                docparam.modified = 1
                return 1
        return 0

    def formatText(self, skipEmpty=1, addPrefix=0, addSuffix=0):
        """Return list of formatted text lines"""
        return self.nodeFormat.formatText(self, skipEmpty, addPrefix, addSuffix)

    def formatChildText(self, skipEmpty=1, addExtra=0):
        """Return list of all children's formatted text lines"""
        if not self.childList:
            return []
        result = []
        sep = []
        prefixAdded = 0
        for child, next in map(None, self.childList, self.childList[1:]):
            addPrefix = not prefixAdded
            prefixAdded = 1
            addSuffix = 0
            if not next or not child.nodeFormat.equalPrefix(next.nodeFormat):
                addSuffix = 1
                prefixAdded = 0
            result.extend(sep + child.formatText(skipEmpty, addPrefix, \
                                                 addSuffix))
            sep = docparam.spaceBetween and [u''] or []
        return result

    def outputItemList(self, includeRoot=1, openOnly=0, level=0):
        """Return list of OutputItems"""
        outList = OutputGroup()
        if includeRoot:
            lines = self.formatText()
            outList.append(OutputItem(lines, level))
            if not self.prevSibling():
                outList[-1].firstSibling = 1
            if not self.nextSibling():
                outList[-1].lastSibling = 1
            if self.childList:
                outList[-1].hasChildren = 1
            outList[-1].prefix = self.nodeFormat.sibPrefix
            outList[-1].suffix = self.nodeFormat.sibSuffix
        else:
            level -= 1
        if self.open or not openOnly:
            for child in self.childList:
                outList.extend(child.outputItemList(1, openOnly, level + 1))
        return outList

    def branchXml(self, typeList=None, writeOptions=0):
        """Return list of xml lines, include format info if not in typeList"""
        if typeList == None:  # default writes no format info
            typeList = docparam.nodeFormats
        xmlList = [u'<%s item="y"' % escape(self.nodeFormat.name, \
                                            treedoc.escDict)]
        if writeOptions:
            if not docparam.spaceBetween:
                xmlList[0] += u' nospace="y"'
            if not docparam.formHtml:
                xmlList[0] += u' nohtml="y"'
            if docparam.xslCssLink:
                xmlList[0] += u' xslcss="%s"' % docparam.xslCssLink
        addFormat = self.nodeFormat not in typeList  # writes format on 1st only
        nodeFormat = self.nodeFormat
        if addFormat:
            xmlList.extend(nodeFormat.formatXml())
            typeList.append(self.nodeFormat)
        xmlList[-1] += u'>'
        for field in nodeFormat.fieldList:
            text = self.data.get(field.name, '')
            if text or addFormat:
                escKey = escape(field.name, treedoc.escDict)
                fieldFormat = ''
                if addFormat:
                    fieldFormat = field.writeXml()
                xmlList.append(u'<%s%s>%s</%s>' % \
                               (escKey, fieldFormat, \
                                escape(text, treedoc.escDict), escKey))
        for child in self.childList:
            xmlList.extend(child.branchXml(typeList))
        if writeOptions:      # write format info for any unused formats
            for format in docparam.nodeFormats:
                if format not in typeList:
                    name =  escape(format.name, treedoc.escDict)
                    xmlList.append(u'<%s item="n"' % name)
                    xmlList.extend(format.formatXml())
                    xmlList[-1] += u'>'
                    for field in format.fieldList:
                        escKey = escape(field.name, treedoc.escDict)
                        fieldFormat = field.writeXml()
                        xmlList.append(u'<%s%s></%s>' % (escKey, fieldFormat, \
                                                         escKey))
                    xmlList.append(u'</%s>' % name)
        xmlList.append(u'</%s>' % escape(self.nodeFormat.name, \
                                         treedoc.escDict))
        return xmlList

    def exportToText(self, level=0, openOnly=0):
        """Write tabbed list of descendants titles"""
        textList = [u'\t' * level + self.title()]
        if self.open or not openOnly:
            for child in self.childList:
                textList.extend(child.exportToText(level + 1, openOnly))
        return textList

    def exportDir(self, parentTitle=None, header='', footer=''):
        """Write dir structure with html tables"""
        if not self.childList:
            return
        dirName = str(self.data.get(self.nodeFormat.fieldList[0].name, ''))
        try:
            if not os.access(dirName, os.R_OK):
                os.mkdir(dirName, 0755)
            os.chdir(dirName)
        except OSError:
            print 'Error - cannot make directory', dirName
            return
        title = self.title()
        lines = [u'<html>', u'<meta http-equiv="Content-Type" ' \
                 'content="text/html; charset=utf-8">', u'<head>', \
                 u'<title>%s</title>' % title, u'</head>', u'<body>']
        if header:
            lines.append(header)
        lines.append(u'<h1 align="center">%s</h1>' % title)
        if parentTitle:
            lines.append('<p align="center">Parent: '
                         '<a href="../index.html">%s</a></p>' % parentTitle)
        lines.extend([u'<table cellpadding="10">', u'<tr>'])
        ### headings kludge????
        headings = self.childList[0].nodeFormat.lineFields()
        lines.extend([u'<th><u>%s</u></th>' % cell for cell in headings])
        lines.append(u'</tr><tr>')
        for child in self.childList:
            textList = child.formatText(0)
            childDir = child.data.get(child.nodeFormat.fieldList[0].name, '')
            if child.childList:
                textList[0] = '<a href="%s/index.html">%s</a>' % \
                              (childDir, textList[0])
            lines.extend([u'<td>%s</td>' % cell for cell in textList])
            lines.append(u'</tr><tr>')
        lines.extend([u'</tr>', u'</table>'])
        if footer:
            lines.append(footer)
        lines.extend([u'</body>', u'</html>'])
        try:
            f = codecs.open('index.html', 'w', 'utf-8')
            f.writelines([line + '\n' for line in lines])
        except IOError:
            print 'Error - could not write file to', dirname
            return
        f.close()
        for child in self.childList:
            child.exportDir(title, header, footer)
        os.chdir('..')

    def exportXbelBookmarks(self, level=0):
        """Return text list with descendant bookmarks in XBEL format"""
        indentsPerLevel = 3
        indent = ' ' * (indentsPerLevel * level)
        nextIndent = ' ' * (indentsPerLevel * (level + 1))
        title = escape(self.title(), treedoc.escDict)
        if not self.childList and level > 0:
            field = self.nodeFormat.findLinkField()
            if field:
                link = self.data.get(field.name, '')
                if link:
                    return [u'%s<bookmark href="%s">' % (indent, link), \
                            u'%s<title>%s</title>' % (nextIndent, title), \
                            u'%s</bookmark>' % indent]
            elif self.nodeFormat.name.lower() == 'separator':
                return [u'%s<separator/>' % indent]
        if level == 0:
            result = [u'<xbel>']
        else:
            result = [u'%s<folder>' % indent]
        result.append(u'%s<title>%s</title>' % (nextIndent, title))
        for child in self.childList:
            result.extend(child.exportXbelBookmarks(level + 1))
        if level == 0:
            result.append(u'</xbel>')
        else:
            result.append(u'%s</folder>' % indent)
        return result

    def exportHtmlBookmarks(self, level=0):
        """Return text list with descendant bookmarks in Mozilla format"""
        indentsPerLevel = 4
        indent = ' ' * (indentsPerLevel * level)
        title = escape(self.title())
        if not self.childList and level > 0:
            field = self.nodeFormat.findLinkField()
            if field:
                link = self.data.get(field.name, '')
                if link:
                    return [u'%s<DT><A HREF="%s">%s</A>' % (indent, link, \
                                                            title)]
            elif self.nodeFormat.name.lower() == 'separator':
                return [u'%s<HR>' % indent]
        if level == 0:
            result = [u'<TITLE>%s</TITLE>' % self.title(), \
                      u'<H1>%s</H1>' % title, '']
        else:
            result = [u'%s<DT><H3>%s</H3>' % (indent, title)]
        if self.childList:
            result.append(u'%s<DL><p>' % indent)
            for child in self.childList:
                result.extend(child.exportHtmlBookmarks(level + 1))
            result.append(u'%s</DL><p>' % indent)
        return result

    def loadTabbedChildren(self, bufList, level=0):
        """Recursive read of TreeItems from tabbed buffer"""
        while bufList:
            if bufList[0][0] == level + 1:
                buf = bufList.pop(0)
                child = TreeItem(self, self.nodeFormat, buf[1])
                self.childList.append(child)
                if not child.loadTabbedChildren(bufList, level + 1):
                    return 0
            elif 0 < bufList[0][0] <= level:
                return 1
            else:
                return 0
        return 1

    def editChildList(self, textList):
        """Update child names and order from textList"""
        stripList = filter(None, [u' '.join(text.split()) for text in textList])
        if len(stripList) == len(self.childList):
            for i in range(len(stripList)):  # assume rename if length is same
                if self.childList[i].title() != stripList[i]:
                    docparam.undoStore.addDataUndo(self.childList[i], [self], 1)
                    self.childList[i].setTitle(stripList[i])
                    docparam.modified = 1
        else:                # find new child positions if length differs
            oldLen = len(self.childList)
            newType = docparam.nodeFormats.findFormat(self.nodeFormat.childType)
            if not newType:
                newType = oldLen and self.childList[0].nodeFormat or \
                          self.nodeFormat
            docparam.undoStore.addChildListUndo(self, [self])
            newChildList = []
            for text in stripList:
                try:
                    newChildList.append(self.childList.pop(\
                     [child.title() for child in self.childList].index(text)))
                except ValueError:
                    newChildList.append(TreeItem(self, newType, text))
            if oldLen == 0 and newChildList:
                self.open = 1
            self.childList = newChildList
            docparam.modified = 1

    def descendantList(self, inclClosed=0, level=0):
        """Recursive list of TreeItems, default to open only, sets level num,
           returns list"""
        descendList = [self]
        self.level = level
        if self.open or inclClosed:
            for child in self.childList:
                descendList.extend(child.descendantList(inclClosed, level+1))
        return descendList

    def ancestorList(self):
        """Return list all parents and grandparents of self"""
        item = self.parent
        result = []
        while item:
            result.append(item)
            item = item.parent
        return result

    def allAncestorsOpen(self):
        """Returns 1 if all ancestors are set open"""
        closeList = [item for item in self.ancestorList() if not item.open]
        if closeList:
            return 0
        return 1

    def hasDescendant(self, child):
        """Return 1 if self has descendant child"""
        for item in self.descendantList(1):
            if item is child:
                return 1
        return 0

    def usesType(self, nodeFormat):
        """Return 1 if dataType is used by self or descendants"""
        for item in self.descendantList(1):
            if item.nodeFormat == nodeFormat:
                return 1
        return 0

    def numChildren(self):
        """Return number of children"""
        return len(self.childList)

    def childPos(self, child):
        """Return the number of the referenced child or -1"""
        for num in range(len(self.childList)):
            if self.childList[num] is child:
                return num
        return -1

    def childText(self):
        """Return list of child item strings (not recursive)"""
        return [child.title() for child in self.childList]

    def maxDescendLevel(self, thisLevel=0):
        """Return max number of levels below this node"""
        if not self.childList:
            return thisLevel
        return max([child.maxDescendLevel(thisLevel + 1) for child \
                    in self.childList])

    def prevSibling(self):
        """Return nearest older sibling or None"""
        if self.parent:
            i = self.parent.childPos(self)
            if i > 0:
                return self.parent.childList[i-1]
        return None

    def nextSibling(self):
        """Return next younger sibling or None"""
        if self.parent:
            i = self.parent.childPos(self) + 1
            if i < len(self.parent.childList):
                return self.parent.childList[i]
        return None

    def prevItem(self, inclClosed=0):
        """Return previous sibling or parent or None"""
        sib = self.prevSibling()
        if sib:
            while sib.numChildren() and (sib.open or inclClosed):
                sib = sib.childList[-1]
            return sib
        return self.parent

    def nextItem(self, inclClosed=0):
        """Return first child, next sibling or ancestors next sibling
           or None"""
        if self.childList and (self.open or inclClosed):
            return self.childList[0]
        ancestor = self
        while ancestor:
            sib = ancestor.nextSibling()
            if sib:
                return sib
            ancestor = ancestor.parent
        return None

    def delete(self):
        """Remove self from the tree structure - return parent on success"""
        parent = self.parent
        if not parent:
            return None
        parent.childList.remove(self)
        self.parent = None
        docparam.modified = 1
        return parent

    def addChild(self, text=u'New', pos=-1):
        """Add new child before position, -1 is at end - return new item"""
        if pos < 0:
            pos = len(self.childList)
        newFormat = docparam.nodeFormats.findFormat(self.nodeFormat.childType)
        if not newFormat:
            newFormat = self.childList and self.childList[0].nodeFormat or \
                        self.nodeFormat
        newItem = TreeItem(self, newFormat, text)
        self.childList.insert(pos, newItem)
        docparam.modified = 1
        return newItem

    def insertSibling(self, text=u'New', inAfter=0):
        """Add new sibling before or after self - return new item on success"""
        if not self.parent:
            return None
        pos = self.parent.childPos(self)
        if inAfter:
            pos += 1
        newFormat = docparam.nodeFormats.findFormat(self.parent.\
                                                    nodeFormat.childType)
        if not newFormat:
            newFormat = self.nodeFormat
        newItem = TreeItem(self.parent, newFormat, text)
        self.parent.childList.insert(pos, newItem)
        docparam.modified = 1
        return newItem

    def addTree(self, rootItem, pos=-1):
        """Add new tree as a child before position, -1 is at end
           return item"""
        if pos < 0:
            pos = len(self.childList)
        self.childList.insert(pos, rootItem)
        rootItem.parent = self
        docparam.modified = 1
        return rootItem

    def insertTree(self, rootItem, inAfter=0):
        """Add new tree before or after self - return item on success"""
        if not self.parent:
            return None
        pos = self.parent.childPos(self)
        if inAfter:
            pos += 1
        self.parent.childList.insert(pos, rootItem)
        rootItem.parent = self.parent
        docparam.modified = 1
        return rootItem

    def indent(self):
        """Becomes a child of the previous sibling - return self on success"""
        newParent = self.prevSibling()
        if not newParent:
            return None
        self.delete()
        newParent.addTree(self, -1)
        docparam.modified = 1
        return self

    def unindent(self):
        """Becomes its parents next sibling - return self on success"""
        sibling = self.parent
        if not sibling or not sibling.parent:
            return None
        self.delete()
        sibling.insertTree(self, 1)
        docparam.modified = 1
        return self

    def move(self, amount=-1):
        """Switch self with sibling, -1=up, 1=down, return self on success"""
        if self.parent:
            i = self.parent.childPos(self)
            j = i + amount
            if 0 <= j < len(self.parent.childList):
                self.parent.childList[i], self.parent.childList[j] = \
                                          self.parent.childList[j], self
                docparam.modified = 1
                return self
        return None

    def openBranch(self, setOpen=1):
        """Recursive open/close of all descendants, close if setOpen=0"""
        self.open = setOpen
        for child in self.childList:
            child.openBranch(setOpen)

    def openParents(self, openSelf=1):
        """Open self's ancestors and self if openSelf, 
           return list of changed items"""
        openList = []
        item = openSelf and self or self.parent
        while item:
            if not item.open and item.childList:
                item.open = 1
                openList.append(item)
            item = item.parent
        return openList

    def cmpItems(self, item1, item2):
        """Compare function for sorting, not case sensitive"""
        if not docparam.sortFields[0][0]:
            factor = docparam.sortFields[0][1] and 1 or -1
            return factor * cmp(item1.title().upper(), item2.title().upper())
        for field, direction in docparam.sortFields:
            field1 = item1.nodeFormat.findField(field)
            field2 = item2.nodeFormat.findField(field)
            factor = direction and 1 or -1
            if field1.sortSequence == field2.sortSequence:
                result = cmp(field1.sortValue(item1.data.get(field, '')), \
                             field2.sortValue(item2.data.get(field, '')))
                if result != 0:
                    return factor * result
            else:
                return factor * cmp(field1.sortSequence, field2.sortSequence)
        return 0

    def sortChildren(self):
        """Sort item children, not case sensitive"""
        self.childList.sort(self.cmpItems)
        docparam.modified = 1

    def sortType(self, nodeFormat):
        """Sort item descendants of a given format type"""
        childOfTypeList = [child for child in self.childList \
                           if child.nodeFormat == nodeFormat]
        if childOfTypeList:
            childOfTypeList.sort(self.cmpItems)
            if len(childOfTypeList) < len(self.childList):
                childOfTypeList.extend([child for child in self.childList \
                                        if child.nodeFormat != type])
            self.childList = childOfTypeList
        for child in self.childList:
            child.sortType(nodeFormat)
        docparam.modified = 1

    def sortBranch(self):
        """Sort item descendants, not case sensitive"""
        self.childList.sort(self.cmpItems)
        for child in self.childList:
            child.sortBranch()
        docparam.modified = 1

    def matchWords(self, wordList):
        """Return 1 if all words are in data fields, not case sensitive"""
        dataStr = u' '.join(self.data.values()).lower()
        for word in wordList:
            if dataStr.find(word) == -1:
                return 0
        return 1

    def cmpFields(self, fieldNames, item):
        """Return 1 if listed fields are the same in item and self"""
        for field in fieldNames:
            if self.data.get(field, '') != item.data.get(field, ''):
                return 0
        return 1

    def findEqivFields(self, fieldNames, itemList):
        """Return first item from list with same listed fields or None"""
        for item in itemList:
            if self.cmpFields(fieldNames, item):
                return item
        return None

    def commonFields(self, itemList):
        """Return names of fields that are common to all items in list"""
        if not itemList:
            return []
        typeList = []
        for item in itemList:
            if item.nodeFormat not in typeList:
                typeList.append(item.nodeFormat)
        fieldNames = typeList.pop(0).fieldNames()
        for type in typeList:
            typeFields = type.fieldNames()
            for field in fieldNames[:]:
                if field not in typeFields:
                    fieldNames.remove(field)
        return fieldNames

    def editFields(self, valueDict):
        """Set values for fields based on dictionary"""
        for field in valueDict.keys():
            self.data[field] = valueDict[field]
        docparam.modified = 1

    def descendTypes(self):
        """Return list of all types found in descendants"""
        types = []
        for item in self.descendantList(1)[1:]:
            if item.nodeFormat not in types:
                types.append(item.nodeFormat)
        return types

    def branchFields(self):
        """Return names of all fields found in self and descendents"""
        types = []
        for item in self.descendantList(1):
            if item.nodeFormat not in types:
                types.append(item.nodeFormat)
        fieldNames = []
        for type in types:
            for field in type.fieldNames():
                if field not in fieldNames:
                    fieldNames.append(field)
        return fieldNames

    def filterDescendants(self, type, expr, editFormats):
        """Remove children of given type recursively if expr is false"""
        for child in self.childList[:]:
            data = child.editDataDict(editFormats)
            if child.nodeFormat != type or eval(expr):
                child.filterDescendants(type, expr, editFormats)
            else:
                self.childList.remove(child)
                child.parent = None
                docparam.modified = 1

    def editDataDict(self, editFormats):
        """Return dict of data converted to edit format for filtering"""
        result = {}
        for field in self.nodeFormat.fieldList:
            text = self.data.get(field.name, '')
            result[field.name] = field.editText(text, editFormats)
        return result

    def addChildCat(self, catList):
        """Add child's category items as a new child level to expand data"""
        newType = '%s_TYPE' % catList[0]
        num = 1
        formatNames = docparam.nodeFormats.nameList()
        while newType in formatNames and \
              docparam.nodeFormats[formatNames.index(newType)].fieldNames() \
              != catList:
            newType = '%s_TYPE_%d' % (catList[0], num)
            num += 1
        newFormat = NodeFormat(newType, catList[0])
        docparam.nodeFormats.append(newFormat)
        for field in catList[1:]:
            newFormat.addNewField(field)
        newItems = []
        for child in self.childList:
            newParent = child.findEqivFields(catList, newItems)
            if not newParent:
                newParent = TreeItem(self, newFormat)
                for field in catList:
                    newParent.data[field] = child.data.get(field, '')
                newItems.append(newParent)
            newParent.childList.append(child)
            child.parent = newParent
        self.childList = newItems
        docparam.modified = 1

    def flatChildCat(self):
        """Collapse data by merging fields"""
        origFormatList = copy.deepcopy(docparam.nodeFormats)
        self.childList = [item for item in self.descendantList(1) if not \
                          item.childList]
        for item in self.childList:
            fieldList = item.nodeFormat.fieldNames()
            origFields = origFormatList.findFormat(item.nodeFormat.name)\
                                       .fieldNames()
            addedFields = []
            oldParent = item.parent
            while oldParent != self:
                for field in origFormatList.findFormat(oldParent.nodeFormat\
                                                       .name).fieldNames():
                    newField = field
                    num = 1
                    while newField in origFields or newField in addedFields:
                        newField = '%s_%d' % (field, num)
                        num += 1
                    item.data[newField] = oldParent.data.get(field, '')
                    addedFields.append(newField)
                    item.nodeFormat.addFieldIfNew(newField)
                oldParent = oldParent.parent
            item.parent = self
        docparam.modified = 1

    def arrangeByRef(self, refField):
        """Arrange data using parent references"""
        descendList = self.descendantList(1)[1:]
        for item in descendList:
            item.childList = []
        self.childList = []
        for item in descendList:
            refText = item.data.get(refField, '')
            parentList = [parent for parent in descendList if \
                       parent.data.get(parent.nodeFormat.fieldNames()[0], '') \
                       == refText]
            if len(parentList) > 1:   # pick nearest parent above the item
                itemPos = descendList.index(item)
                while len(parentList) > 1 and \
                      descendList.index(parentList[1]) < itemPos:
                    del parentList[0]
            if not parentList or parentList[0] == item:
                item.parent = self
            else:
                item.parent = parentList[0]
            item.parent.childList.append(item)
        docparam.modified = 1

    def flatByRef(self, refField):
        """Collapse data after adding references to parents"""
        descendList = self.descendantList(1)[1:]
        self.childList = descendList
        for item in descendList:
            item.childList = []
            item.data[refField] = item.parent.data.get(item.parent.nodeFormat.\
                                                       fieldNames()[0], '')
            item.parent = self
            item.nodeFormat.addFieldIfNew(refField)
        docparam.modified = 1

    def updateByRef(self, refRoot):
        """Update with new fields from reference file with matched 1st field
           Return a string describing changes"""
        refData = {}
        for item in refRoot.descendantList(1):
            refData[item.data.get(item.nodeFormat.fieldNames()[0], '')] = item
        formatChgs = {}
        numNewEntries = 0
        for item in self.descendantList(1):
            fields = item.nodeFormat.fieldNames()
            try:
                ref = refData[item.data[fields[0]]]
                for field in ref.data.keys():
                    if field not in fields:
                        item.data[field] = ref.data[field]
                        numNewEntries += 1
                        formatChgs.setdefault(item.nodeFormat.name, {}).\
                                 setdefault(field, \
                                            ref.nodeFormat.findField(field))
            except KeyError:
                pass
        numChgTypes = 0
        numNewFields = 0
        for typeName in formatChgs.keys():
            type = docparam.nodeFormats.findFormat(typeName)
            ref = formatChgs[typeName]
            numChgTypes += 1
            for field in ref.values():
                type.fieldList.append(field)
                numNewFields += 1
        if numNewEntries:
            docparam.modified = 1
        return (numNewEntries, numNewFields, numChgTypes)

    def addNumbering(self, field, format, rootIncluded, appendToParent, \
                     singleLevel=0, startNum=1, currentLevel=0):
        """Add number field to this node and descendants"""
        if rootIncluded:
            self.data[field] = numbering.numSeries(startNum, startNum + 1, \
                                                   format[currentLevel])[0]
            self.nodeFormat.addFieldIfNew(field)
            docparam.modified = 1
            currentLevel += 1
            startNum = 1
        if not self.childList:
            return
        docparam.modified = 1
        numList = numbering.numSeries(startNum, \
                                      len(self.childList) + startNum, \
                                      format[currentLevel])
        if appendToParent and currentLevel:
            numList = [self.data[field] + numText for numText in numList]
        for item in self.childList:
            item.data[field] = numList.pop(0)
            item.nodeFormat.addFieldIfNew(field)
        if not singleLevel:
            for item in self.childList:
                item.addNumbering(field, format, 0, appendToParent, \
                                   0, 1, currentLevel + 1)

