#!/usr/bin/env python

#****************************************************************************
# treemainwin.py, provides a class for the main window
#
# TreeLine, an information storage program
# Copyright (C) 2004, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#*****************************************************************************

try:
    from __main__ import __version__, __author__, helpFilePath
except ImportError:
    __version__ = __author__ = '??'
    helpFilePath = None
from treedoc import TreeDoc
from treedoc import ReadFileError
from treeitem import TreeItem
from treeformat import NodeFormat
from formatlist import FormatList
from treeview import TreeView
from treerightviews import DataOutView, DataEditView
from treedialogs import ConfigDlg, FieldEntry, TypeSetDlg, FieldSelectDlg, \
                        EditFieldsDlg, ExportDlg, FilterDlg, NumberingDlg, \
                        FindTextEntry, RadioChoiceDlg, SpellCheckDlg, \
                        PrintHeaderDlg
from option import Option
from optiondlg import OptionDlg, OptionDlgBool, OptionDlgInt, \
                      OptionDlgStr, OptionDlgDbl, OptionDlgPush, \
                      OptionDlgRadio
from recentfiles import RecentFiles
from treestates import TreeStates
from printpreview import PrintPrevDlg
from helpview import HelpView
from spellcheck import SpellCheck, SpellCheckError
import docparam
from qt import *
if qVersion()[0] >= '3':
    from textedit3 import TitleListView
else:
    from textedit2 import TitleListView
from xpm import *
import sys, os.path, copy, re, codecs


class TreeMainWin(QMainWindow):
    """Main window, menus, toolbar, and status"""
    copyFormat = NodeFormat('_DUMMY__ROOT_')
    minPrintMargin = 0.25
    maxPrintMargin = 3.0
    maxNumCol = 9
    maxNumLines = 50
    defaultList = [\
        "# Options for TreeLine, an information storage program",
        "#",
        "##############################################################",
        "#",
        "# Keyboard key bindings",
        "#",
        "# Note that key names are case-sensitive",
        "#",
        "# Available modifiers include SHIFT, CTRL, ALT, META",
        "#",
        "# Available keys include A-Z, 0-9, F1-F12, Insert, Delete,",
        "# Home, End, PageUp, PageDown, Left, Right, Up, Down, Asterisk,",
        "# Plus, Minus, Comma, Period, Slash, Apostrophe, Semicolon,",
        "# BracketLeft, BracketRight, Equal, Backslash, Space",
        "#",
        "##############################################################",
        "#",
        "FileNew             CTRL N",
        "FileOpen            CTRL O",
        "FileSave            CTRL S",
        "FileSaveAs          ",
        "FileExport          ",
        "FilePrint           CTRL P",
        "FilePrintOpt        ",
        "FileQuit            CTRL Q",
        "#",
        "EditUndo            CTRL Z",
        "EditRedo            CTRL Y",
        "EditCut             CTRL X",
        "EditCopy            CTRL C",
        "EditCopyText        ",
        "EditPaste           CTRL V",
        "EditPasteText       ",
        "EditRename          CTRL R",
        "EditInBefore        CTRL B",
        "EditInAfter         CTRL I",
        "EditAddChild        CTRL A",
        "EditDelete          Delete",
        "EditIndent          CTRL SHIFT Right",
        "EditUnindent        CTRL SHIFT Left",
        "EditMoveUp          CTRL SHIFT Up",
        "EditMoveDown        CTRL SHIFT Down",
        "#",
        "ViewDataOutput      CTRL D",
        "ViewDataEdit        CTRL E",
        "ViewTitleList       CTRL L",
        "ViewGoTree          CTRL G",
        "ViewDataPageUp      SHIFT PageUp",
        "ViewDataPageDown    SHIFT PageDown",
        "#",
        "DataSetItemType     CTRL T",
        "DataSetDescendType  ",
        "DataConfigType      ",
        "DataSortChildren    ",
        "DataSortType        ",
        "DataSortTitle       ",
        "DataFilter          ",
        "DataChange          ",
        "DataNumber          ",
        "DataCategoryAdd     ",
        "DataCategoryFlat    ",
        "DataRefArrange      ",
        "DataRefFlat         ",
        "DataRefUpdate       ",
        "#",
        "ToolsExpand         CTRL Right",
        "ToolsCollapse       CTRL Left",
        "ToolsFind           CTRL F",
        "ToolsSpellCheck     ",
        "ToolsRemXLST        ",
        "ToolsGenOptions     ",
        "ToolsFileOptions    ",
        "ToolsBackColor      ",
        "ToolsTextColor      ",
        "#",
        "HelpReadMe          F1",
        "HelpAbout           ",
        "#",
        "TreeSelectPrev      Up",
        "TreeSelectNext      Down",
        "TreeOpenItem        Right",
        "TreeCloseItem       Left",
        "TreePrevSibling     CTRL Up",
        "TreeNextSibling     CTRL Down",
        "TreeSelectParent    CTRL U",
        "TreeTop             Home",
        "TreeBottom          End",
        "TreePageUp          PageUp",
        "TreePageDown        PageDown",
        "TreeIncremSearch    Slash",
        "TreeIncremNext      F3",
        "TreeIncremPrev      SHIFT F3",
        "#",
        "##############################################################",
        "#",
        "# The following options are set from within the program,",
        "# editing here is not recommended",
        "#",
        "##############################################################",
        "#",
        "AutoFileOpen        no",
        "StartShowChildren   yes",
        "PersistTreeState    yes",
        "SaveWindowGeom      yes",
        "ClickRename         yes",
        "DragTree            yes",
        "InsertOnEnter       yes",
        "RenameNewNodes      yes",
        "OpenSearchNodes     yes",
        "CompressNewFiles    no",
        "HtmlNewFields       no",
        "#",
        "IndentOffset        20",
        "MaxEditLines        7",
        "UndoLevels          8",
        "AutoSaveMinutes     0",
        "#",
        "EditNumFormat       0.0000",
        "EditDateFormat      mm/dd/yyyy",
        "EditTimeFormat      H:MM",
        "#",
        "PrintPrev           yes",
        "PrintRoot           yes",
        "PrintLines          no",
        "PrintOpenOnly       no",
        "PrintUseOutputFont  yes",
        "PrintFont           ",
        "PrintFontSize       ",
        "PrintFontBold       no",
        "PrintFontItalic     no",
        "PrintFontUnderline  no",
        "PrintFontStrikeOut  no",
        "PrintPageSize       letter",
        "PrintNumCols        1",
        "PrintColSpace       0.30",
        "PrintOffset         20",
        "HorizMargin         0.50",
        "VertMargin          0.70",
        "#",
        "BackgroundR         255",
        "BackgroundG         255",
        "BackgroundB         255",
        "ForegroundR         0",
        "ForegroundG         0",
        "ForegroundB         0",
        "#",
        "OutputFont          ",
        "OutputFontSize      ",
        "OutputFontBold      no",
        "OutputFontItalic    no",
        "OutputFontUnderline no",
        "OutputFontStrikeOut no",
        "#",
        "ToolbarData         [1,2,0,0,0],[1,2,1,0,0]",
        "RecentFile1         ",
        "RecentFile2         ",
        "RecentFile3         ",
        "RecentFile4         ",
        "TreeState1          0:0:0:0",
        "TreeState2          0:0:0:0",
        "TreeState3          0:0:0:0",
        "TreeState4          0:0:0:0",
        "SpellCheckPath      ",
        "#",
        "WindowXSize         540",
        "WindowYSize         460",
        "WindowXPos          0",
        "WindowYPos          0",
        "TreeSplitPercent    40",
        "OutputSplitPercent  20",
        "EditorSplitPercent  25",
        "TitleSplitPercent   10",
        "PrintPrevXSize      300",
        "PrintPrevYSize      560",
        "PrintPrevXPos       0",
        "PrintPrevYPos       0"]

    def __init__(self, parent=None, name=None):
        QMainWindow.__init__(self, parent, name)
        self.setIcon(QPixmap(treeline_xpm))
        self.setAcceptDrops(1)
        self.option = Option('treeline', 20)
        self.option.loadAll(TreeMainWin.defaultList)
        self.resize(self.option.intData('WindowXSize', 10, 10000), \
                    self.option.intData('WindowYSize', 10, 10000))
        if self.option.boolData('SaveWindowGeom'):
            self.move(self.option.intData('WindowXPos', 0, 10000), \
                      self.option.intData('WindowYPos', 0, 10000))
        self.updateColors()
        self.editFormats = {}
        self.updateEditFormats()
        self.autoSaveTimer = QTimer(self)
        self.connect(self.autoSaveTimer, SIGNAL('timeout()'), self.autoSave)
        self.showItemChildren = self.option.boolData('StartShowChildren')

        split = QSplitter(self)
        self.setCentralWidget(split)
        self.treeView = TreeView(self, split)
        rightView = QWidget(split)
        rightLayout = QVBoxLayout(rightView)
        self.rightStack = QWidgetStack(rightView)
        rightLayout.addWidget(self.rightStack)
        rightTabs = QTabBar(rightView)
        rightTabs.setShape(QTabBar.RoundedBelow)
        rightTabs.setFocusPolicy(QWidget.NoFocus)

        dataOutId = rightTabs.addTab(QTab('&Data Output'))
        dataEditId = rightTabs.addTab(QTab('Data &Editor'))
        titleListId = rightTabs.addTab(QTab('Title &List'))
        rightLayout.addWidget(rightTabs)

        self.dataOutSplit = QSplitter(Qt.Vertical, self.rightStack)
        self.rightStack.addWidget(self.dataOutSplit, dataOutId)
        self.dataEditSplit = QSplitter(Qt.Vertical, self.rightStack)
        self.rightStack.addWidget(self.dataEditSplit, dataEditId)
        self.titleListSplit = QSplitter(Qt.Vertical, self.rightStack)
        self.rightStack.addWidget(self.titleListSplit, titleListId)

        parentView = DataOutView(self, 0, self.dataOutSplit)
        childView = DataOutView(self, 1, self.dataOutSplit)
        childList = [childView]
        DataEditView(self, 0, self.dataEditSplit)
        childList.append(DataEditView(self, 1, self.dataEditSplit))
        TitleListView(self, 0, self.titleListSplit)
        childList.append(TitleListView(self, 1, self.titleListSplit))
        if not self.showItemChildren:
            for child in childList:
                child.hide()
        font = self.getFontFromOptions('Output')
        if font:
            parentView.setFont(font)
            childView.setFont(font)

        outSplitPercent = self.option.intData('OutputSplitPercent', 1, 99)
        self.dataOutSplit.setSizes([outSplitPercent, 100 - outSplitPercent])
        editSplitPercent = self.option.intData('EditorSplitPercent', 1, 99)
        self.dataEditSplit.setSizes([editSplitPercent, 100 - editSplitPercent])
        titleSplitPercent = self.option.intData('TitleSplitPercent', 1, 99)
        self.titleListSplit.setSizes([titleSplitPercent, \
                                      100 - titleSplitPercent])

        self.connect(rightTabs, SIGNAL('selected(int)'), self.viewTabSelect)
        self.rightStack.raiseWidget(dataOutId)
        mainSplitPercent = self.option.intData('TreeSplitPercent', 1, 99)
        split.setSizes([mainSplitPercent, 100 - mainSplitPercent])

        self.doc = None
        self.setTypeDlg = None
        self.FindDlg = None
        self.helpView = None
        self.blockViewChg = 0
        self.fileImported = 0
        self.printer = QPrinter()
        try:
            pageSize = eval('QPrinter.%s' % self.option.\
                            strData('PrintPageSize', 0).capitalize())
            self.printer.setPageSize(pageSize)
        except AttributeError:
            self.printer.setPageSize(QPrinter.Letter)
        self.printFont = None   # set at printing
        self.printList = []
        self.pageLevelBreaks = []
        self.linesPerPage = 0

        self.setupMenus()
        toolbarData = eval('[' + self.option.strData('ToolbarData', 0) + ']')
        self.restoreToolbars(self.toolbars, toolbarData)
        self.statusBar().message('Ready', 2000)
        self.fileNew()

        self.connect(self.treeView, SIGNAL('selectionChanged()'), \
                     self.updateRightView)
        self.connect(QApplication.clipboard(), SIGNAL('dataChanged()'), \
                     self.setPasteAvail)
        self.connect(self.rightStack, SIGNAL('aboutToShow(QWidget*)'), \
                     self.updateRightView)
        self.connect(self.rightStack, SIGNAL('aboutToShow(int)'), \
                     rightTabs, SLOT('setCurrentTab(int)'))
        self.setPasteAvail()

    def getFontFromOptions(self, optionPrefix):
        """Return font if set in options or None"""
        fontName = self.option.strData('%sFont' % optionPrefix, 1)
        if fontName:
            try:
                fontSize = int(self.option.strData('%sFontSize' % \
                                                   optionPrefix, 1))
            except ValueError:
                fontSize = 10
            font = QFont(fontName, fontSize)
            font.setBold(self.option.boolData('%sFontBold' % optionPrefix))
            font.setItalic(self.option.boolData('%sFontItalic' % optionPrefix))
            font.setUnderline(self.option.boolData('%sFontUnderline' % \
                                                   optionPrefix))
            font.setStrikeOut(self.option.boolData('%sFontStrikeOut' % \
                                                   optionPrefix))
            return font
        return None

    def saveFontToOptions(self, font, optionPrefix):
        """Store font in option settings"""
        self.option.changeData('%sFont' % optionPrefix, str(font.family()), 1)
        self.option.changeData('%sFontSize' % optionPrefix, \
                               str(font.pointSize()), 1)
        self.option.changeData('%sFontBold' % optionPrefix, \
                               font.bold() and 'yes' or 'no', 1)
        self.option.changeData('%sFontItalic' % optionPrefix, \
                               font.italic() and 'yes' or 'no', 1)
        self.option.changeData('%sFontUnderline' % optionPrefix, \
                               font.underline() and 'yes' or 'no', 1)
        self.option.changeData('%sFontStrikeOut' % optionPrefix, \
                               font.strikeOut() and 'yes' or 'no', 1)

    def updateViews(self, select=None):
        """Update tree and child views"""
        QApplication.setOverrideCursor(Qt.waitCursor)
        self.treeView.updateTree(select)
        QApplication.restoreOverrideCursor()
        self.updateRightView()

    def updateRightView(self, split=None):
        """Update given right-hand view or the active one"""
        QApplication.setOverrideCursor(Qt.waitCursor)
        if not split:
            split = self.rightStack.visibleWidget()
        else:
            self.viewToAct[split].setOn(1)
        for view in split.children():
            if hasattr(view, 'updateView'):
                view.updateView()
        if self.setTypeDlg and self.setTypeDlg.isVisible():
            self.setTypeDlg.updateDlg()
        self.updateCmdAvail()
        QApplication.restoreOverrideCursor()

    def focusWidgetWithAttr(self, attr):
        """Return the focused widget or it's ancestor
           that has the given attr or None"""
        widget = qApp.focusWidget()
        while widget and not hasattr(widget, attr):
            widget = widget.parent()
        return widget

    def updateCmdAvail(self):
        """Update the enabled status of menus"""
        item = self.treeView.currentTreeItem()
        if not self.doc or not item:
            return
        self.fileSaveAct.setEnabled(docparam.modified)
        self.editInsertActGrp.setEnabled(item.parent != None)
        selParents = [node.parent for node in self.doc.selectedNodes]
        numChildren = [len(node.childList) for node in self.doc.selectedNodes]
        self.editUndoAct.setEnabled(len(docparam.undoStore.undoList))
        self.editRedoAct.setEnabled(len(docparam.redoStore.undoList))
        self.editDeleteAct.setEnabled(len(self.doc.selectedNodes) and \
                                      self.doc.root not in \
                                      self.doc.selectedNodes)
        self.editPrevSibActGrp.setEnabled(len(self.doc.selectedNodes) and \
                                          None not in \
                                          [node.prevSibling() for node in \
                                           self.doc.selectedNodes])
        self.editUnindentAct.setEnabled(len(self.doc.selectedNodes) and \
                                        None not in selParents and \
                                        None not in [node.parent for node in \
                                                     selParents])
        self.editMoveDownAct.setEnabled(len(self.doc.selectedNodes) and \
                                        None not in \
                                        [node.nextSibling() for node in \
                                         self.doc.selectedNodes])
        self.dataSelExistActGrp.setEnabled(len(self.doc.selectedNodes))
        self.dataSelParentActGrp.setEnabled(len(self.doc.selectedNodes) and \
                                            len(filter(None, numChildren)))
        allAncestors = []
        for node in self.doc.selectedNodes:
            allAncestors.extend(node.ancestorList())
        self.dataUnrelParentActGrp.setEnabled(len(self.doc.selectedNodes) and \
                                              0 not in numChildren and \
                                              0 not in [node not in \
                                                        allAncestors \
                                                        for node in \
                                                      self.doc.selectedNodes])
        self.dataSingleSelActGrp.setEnabled(len(self.doc.selectedNodes) == 1 \
                                            and numChildren[0])
        self.toolsRemXsltAct.setEnabled(len(self.doc.xlstLink))
        self.statusBar().clear()

    def setPasteAvail(self):
        """Check to see if text is available to paste"""
        try:
            text = unicode(QApplication.clipboard().text())
        except UnicodeError:
            text = ''
        self.editPasteActGrp.setEnabled(len(text))

    def disableAllControls(self):
        """Disable all menu and toolbar commands during tree renaming"""
        for group in self.actGrpList:
            group.setEnabled(0)

    def enableAllControls(self):
        """Enable controls after tree renaming"""
        for group in self.actGrpList:
            group.setEnabled(1)
        self.updateCmdAvail()

    def updateEditFormats(self):
        """Update dict of edit field formats"""
        self.editFormats['Number'] = self.option.strData('EditNumFormat', 1)
        self.editFormats['Date'] = self.option.strData('EditDateFormat', 1)
        self.editFormats['Time'] = self.option.strData('EditTimeFormat', 1)

    def updateColors(self):
        """Adjust the colors to the current option settings"""
        pal = QApplication.palette()
        background = QColor(self.option.intData('BackgroundR', 0, 255), \
                            self.option.intData('BackgroundG', 0, 255), \
                            self.option.intData('BackgroundB', 0, 255))
        foreground = QColor(self.option.intData('ForegroundR', 0, 255), \
                            self.option.intData('ForegroundG', 0, 255), \
                            self.option.intData('ForegroundB', 0, 255))
        pal.setColor(QColorGroup.Base, background)
        pal.setColor(QPalette.Active, QColorGroup.Text, foreground)
        pal.setColor(QPalette.Inactive, QColorGroup.Text, foreground)
        QApplication.setPalette(pal, 1)
        pal.setColor(QColorGroup.Text, QColor(0, 0, 0))
        self.statusBar().setPalette(pal)

    def setMainCaption(self):
        """Set main window caption using doc filename path"""
        caption = ''
        if self.doc.fileName:
            caption = '%s [%s] ' % (os.path.basename(self.doc.fileName), \
                                    os.path.dirname(self.doc.fileName))
        caption += '- TreeLine'
        if sys.platform == 'win32':
            caption += '  (PyQt)'
        self.setCaption(caption)

    def openFile(self, fileName, importOnFail=1):
        """Open given file, fail quietly if not importOnFail,
           return 0 if file should be removed from recent list, 1 otherwise"""
        autoSaveFile = self.autoSaveFilePath(fileName)
        if autoSaveFile and self.option.intData('AutoSaveMinutes', 0, 999):
            ans = QMessageBox.information(self, 'TreeLine', \
                                          'Backup file "%s" exists.\n'\
                                          'A previous session may have '\
                                          'crashed.' % autoSaveFile, \
                                          '&Restore Backup', '&Delete Backup', \
                                          '&Cancel File Open', 0, 2)
            if ans == 0:
                if not self.restoreAutoSaveFile(fileName):
                    QMessageBox.warning(self, 'TreeLine', \
                                        'Error - could not restore backup')
                    return 1
            elif ans == 1:
                self.delAutoSaveFile(fileName)
            else:
                return 1
        QApplication.setOverrideCursor(Qt.waitCursor)
        try:
            self.doc.readFile(fileName)
            self.fileImported = 0
        except IOError:
            QApplication.restoreOverrideCursor()
            QMessageBox.warning(self, 'TreeLine', \
                                'Error - could not read file "%s"' % fileName)
            return 0
        except ReadFileError, e:
            QApplication.restoreOverrideCursor()
            if not importOnFail:
                return 1
            # assume file is not a TreeLine file
            dlg = RadioChoiceDlg('Import Text', 'Choose Text Import Method', \
                        [('Tab indented text, one node per line', \
                          self.doc.readTabbed), \
                         ('Text table with header row, one node per line', \
                          self.doc.readTable), \
                         ('Plain text, one node per line (CR delimitted)', \
                          self.doc.readLines), \
                         ('Plain text paragraphs (blank line delimitted)', \
                          self.doc.readPara), \
                         ('Treepad file (text nodes only)', \
                          self.doc.readTreepad), \
                         ('XML bookmarks (XBEL format)', \
                          self.doc.readXbel), \
                         ('HTML bookmarks (Mozilla format)', \
                          self.doc.readMozilla)], self, None, 1)
            if dlg.exec_loop() != QDialog.Accepted:
                return 0
            try:
                QApplication.setOverrideCursor(Qt.waitCursor)
                apply(dlg.getResult(), (fileName,))
            except ReadFileError, e:
                QApplication.restoreOverrideCursor()
                QMessageBox.warning(self, 'TreeLine', 'Error - %s' % e)
                return 0
            self.fileImported = 1
        self.setMainCaption()
        self.treeView.setFocus()
        self.doc.root.open = 1
        docparam.undoStore.levels = self.option.intData('UndoLevels', 0, 99)
        docparam.redoStore.levels = self.option.intData('UndoLevels', 0, 99)
        QApplication.restoreOverrideCursor()
        self.updateViews(self.doc.root)
        self.recentFiles.addEntry(fileName)
        if self.option.boolData('PersistTreeState'):
            self.treeStates.loadCurrent(self.doc, self.treeView)
            self.updateViews(self.doc.selectedNodes)
        self.updateCmdAvail()
        self.loadTypeSubMenu()
        self.resetAutoSave()
        return 1

    def recentOpen(self, filePath):
        """Open from recentFiles signal"""
        if filePath and self.savePrompt():
            if not self.openFile(filePath):
                self.recentFiles.removeEntry(filePath)

    def autoOpen(self):
        """Open last used file"""
        if self.option.boolData('AutoFileOpen'):
            path = self.recentFiles.recentFileName(0)
            if path and not self.openFile(path, 0):
                self.recentFiles.removeEntry(path)

    def getFileName(self, caption, defaultExt, filterList, currentFilter=0):
        """Return user specified file name for save as & export"""
        dir, name = os.path.split(self.doc.fileName)
        dlg = QFileDialog(dir, ';;'.join(filterList), self, '', 1)
        dlg.setMode(QFileDialog.AnyFile)
        if name:
            dlg.setSelection(os.path.splitext(name)[0] + defaultExt)
        if qVersion()[0] >= '3':
            dlg.setSelectedFilter(currentFilter)
        elif currentFilter:     # can't setSelected, so change order
            filterList.insert(0, filterList.pop(currentFilter))
            dlg.setFilters(';;'.join(filterList))
        dlg.setCaption(caption)
        dlg.setIcon(QPixmap(treeline_xpm))
        result = dlg.exec_loop()
        fileName = str(dlg.selectedFile())
        if result == QDialog.Accepted and fileName:
            selectedFilter = str(dlg.selectedFilter())
            if '.' not in fileName and selectedFilter != 'All Files (*)':
                fileName += defaultExt
            if selectedFilter.find('Uncompressed') >= 0:
                self.doc.compressFile = 0
            elif selectedFilter.find('Compressed') >= 0:
                self.doc.compressFile = 1
            if os.access(fileName, os.F_OK):
                text = 'File "%s" exists.\nReplace existing file?' % fileName
                ans = QMessageBox.information(self, 'TreeLine', text, \
                                              '&Yes', '&No', '&Cancel', 0, 2)
                if ans != 0:
                    if ans == 1:
                        return self.getFileName(caption, defaultExt, \
                                                filterList, currentFilter)
                    return ''
            return fileName
        return ''

    def resetAutoSave(self):
        """Restart auto save timer if the option is enabled"""
        self.autoSaveTimer.stop()
        minutes = self.option.intData('AutoSaveMinutes', 0, 999)
        if minutes:
            self.autoSaveTimer.start(60000 * minutes)

    def autoSave(self):
        """Perform auto save if the option is enabled (called from timer)"""
        if docparam.modified and self.doc.fileName and not self.fileImported:
            self.doc.writeFile(self.doc.fileName + '~', 0)

    def autoSaveFilePath(self, baseName=''):
        """Return the path to a backup file if it exists"""
        filePath = baseName and baseName + '~' or self.doc.fileName + '~'
        if len(filePath) > 1 and os.access(filePath, os.R_OK):
            return filePath
        return ''

    def delAutoSaveFile(self, baseName=''):
        """Remove the backup auto save file if it exists"""
        filePath = self.autoSaveFilePath(baseName)
        if filePath:
            try:
                os.remove(filePath)
            except OSError:
                print 'Could not remove backup file %s' % filePath

    def restoreAutoSaveFile(self, baseName):
        """Move baseName~ to baseName by overwriting, return 1 on success"""
        try:
            os.remove(baseName)
        except OSError:
            print 'Could not remove file %s' % baseName
            return 0
        try:
            os.rename(baseName + '~', baseName)
        except OSError:
            print 'Could not rename file %s' % baseName + '~'
            return 0
        return 1

    def fileNew(self):
        """New file command"""
        if self.savePrompt():
            compress = self.option.boolData('CompressNewFiles')
            htmlAttrs = self.option.boolData('HtmlNewFields') and \
                        {'html': 'y'} or {}
            self.doc = TreeDoc(None, compress, htmlAttrs)
            docparam.undoStore.levels = self.option.intData('UndoLevels', 0, 99)
            docparam.redoStore.levels = self.option.intData('UndoLevels', 0, 99)
            self.setMainCaption()
            self.treeView.setFocus()
            self.updateViews(self.doc.root)
            self.updateCmdAvail()
            self.loadTypeSubMenu()
            self.resetAutoSave()

    def fileOpen(self):
        """Open a file"""
        if self.savePrompt():
            dfltPath = self.recentFiles.firstPath()
            if not dfltPath:
                dfltPath = os.path.expandvars('$HOME')
            if not dfltPath:
                dfltPath = '..'
            filters = ['TreeLine Files (*.trl *.xml)', 'Text Files (*.txt)', \
                       'Treepad Files (*.hjt)', 'XBEL Bookmarks (*.xml)', \
                       'Mozilla Bookmarks (*.html *.htm)', 'All Files (*)']
            fileName = str(QFileDialog.getOpenFileName(dfltPath, \
                                                       ';;'.join(filters), \
                                                       self))
            if fileName:
                self.openFile(fileName)

    def fileSave(self):
        """Save current file"""
        if self.doc.fileName and not self.fileImported:
            self.doc.writeFile(self.doc.fileName)
            self.updateCmdAvail()
            self.delAutoSaveFile()
            self.resetAutoSave()
        else:
            self.fileSaveAs()

    def fileSaveAs(self):
        """Save file with a new name"""
        oldFileName = self.doc.fileName
        filterList = ['TreeLine Files - Uncompressed (*.trl *.xml)', \
                      'TreeLine Files - Compressed (*.trl *.trl.gz)', \
                      'All Files (*)']
        fileName = self.getFileName('Save As', '.trl', filterList, \
                                    self.doc.compressFile)
        if fileName and self.doc.writeFile(fileName):
            self.setMainCaption()
            self.recentFiles.addEntry(fileName)
            self.fileImported = 0
            self.updateCmdAvail()
            self.delAutoSaveFile(oldFileName)
            self.delAutoSaveFile()
            self.resetAutoSave()

    def fileExport(self):
        """Export the file as html, a table or text"""
        item = self.treeView.currentTreeItem()
        dlg = ExportDlg(item.title(), self, None, 1)
        if dlg.exec_loop() != QDialog.Accepted:
            return
        indent = self.option.intData('IndentOffset', 0, TreeView.maxOffset)
        if dlg.exportType == ExportDlg.htmlType:
            fileName = self.getFileName('Export Html', '.html', \
                                        ['Html Files (*.html *.htm)', \
                                         'All Files (*)'])
            if not fileName:
                return
            if dlg.numColumns() == 1:
                self.doc.exportHtml(fileName, item, dlg.isRootIncluded(), \
                                    dlg.openOnly(), indent, \
                                    dlg.headerIncluded())
            else:
                outGroup = item.outputItemList(dlg.isRootIncluded(), \
                                               dlg.openOnly())
                for outItem in outGroup:
                    text = QSimpleRichText(u'<br/>'.join(outItem.textLines), \
                                           self.font())
                    # no <br> on last line - would add newline height
                    # bug in height() - '<br />' doesn't work
                    text.setWidth(5000)
                    outItem.height = text.height()
                self.doc.exportHtmlColumns(fileName, outGroup, \
                                           dlg.numColumns(), indent, \
                                           dlg.headerIncluded())
        elif dlg.exportType == ExportDlg.dirType:
            dirName = QFileDialog.getExistingDirectory(None, self, '', \
                                                       'Export to Directory', 1)
            if dirName:
                self.doc.exportDir(dirName, item, dlg.headerIncluded())
        elif dlg.exportType == ExportDlg.xsltType:
            link, ok = QInputDialog.getText('TreeLine', \
                                            'A link to a stylesheet can be '\
                                            'added to the XSL file\n'\
                                            'Enter a CSS filename '\
                                            '(blank for none)', \
                                            docparam.xslCssLink, self)
            if ok:
                fileName = self.getFileName('Export XSLT', '.xsl', \
                                             ['XSLT Files (*.xsl)', \
                                              'All Files (*)'])
                if fileName:
                    if docparam.xslCssLink != unicode(link):
                        docparam.xslCssLink = unicode(link)
                        docparam.modified = 1
                    self.doc.exportXslt(fileName, dlg.isRootIncluded(), indent)
                    self.fileSaveAct.setEnabled(docparam.modified)
        elif dlg.exportType == ExportDlg.trlType:
            filterList = ['TreeLine Files - Uncompressed (*.trl *.xml)', \
                          'TreeLine Files - Compressed (*.trl *.trl.gz)', \
                          'All Files (*)']
            origCompress = self.doc.compressFile
            fileName = self.getFileName('Export Subtree', '.trl', filterList, \
                                        self.doc.compressFile)
            if fileName:
                self.doc.exportTrlSubtree(fileName, item)
            self.doc.compressFile = origCompress
        elif dlg.exportType == ExportDlg.tableType:
            fileName = self.getFileName('Export Table', '.tbl', \
                                        ['Table Files (*.tbl *.txt)', \
                                         'All Files (*)'])
            if fileName:
                self.doc.exportTable(fileName, item)
        elif dlg.exportType == ExportDlg.textType:
            fileName = self.getFileName('Export Titles', '.txt', \
                                        ['Text Files (*.txt)', \
                                         'All Files (*)'])
            if fileName:
                self.doc.exportTabbedTitles(fileName, item, \
                                            dlg.isRootIncluded(), \
                                            dlg.openOnly())
        elif dlg.exportType == ExportDlg.xbelType:
            fileName = self.getFileName('Export XBEL Bookmarks', '.xml', \
                                        ['XML Files (*.xml)', \
                                         'All Files (*)'])
            if fileName:
                self.doc.exportXbel(fileName, item)
        else:    # mozilla bookmark type
            fileName = self.getFileName('Export Html Bookmarks', '.html', \
                                        ['Html Files (*.html *.htm)', \
                                         'All Files (*)'])
            if fileName:
                self.doc.exportHtmlBookmarks(fileName, item)

    def printPage(self, pageNum, painter):
        """Send pageNum to painter"""
        docparam.fileInfoItem.data['Page_Number'] = repr(pageNum)
        painter.setFont(self.printFont)
        addLines = self.option.boolData('PrintLines')
        indent = self.option.intData('PrintOffset', 0, \
                                     self.treeView.maxOffset)
        xLineDelta = indent / 2
        yLineDelta = 0
        if docparam.spaceBetween:
            yLineDelta = QFontMetrics(self.printFont).ascent()
        linePos = QFontMetrics(self.printFont).ascent() / 2 + 1
        numCols = self.option.intData('PrintNumCols', 1, 9)
        colSpacing = 72 * self.option.numData('PrintColSpace', 0, \
                                              TreeMainWin.maxPrintMargin)
        colWidth = (painter.window().width() - colSpacing * (numCols - 1)) \
                   / numCols
        if qVersion()[0] >= '3':   # qt2 to qt3 text draw API change
            drawRegion = painter.window()
        else:
            drawRegion = QRegion(painter.window())
        header = self.doc.getHeaderFooter(1)
        initY = 0
        if header:
            text = QSimpleRichText(header, self.printFont)
            text.setWidth(painter.window().width())
            text.draw(painter, 0, 0, drawRegion, QColorGroup())
            initY = 2 * QFontMetrics(self.printFont).lineSpacing()
        for colNum in range(numCols):
            colPos = colNum * (colSpacing + colWidth)
            yPos = initY
            if (pageNum - 1) * numCols + colNum >= len(self.printList):
                docparam.fileInfoItem.data['Page_Number'] = ''
                return
            lineStarts = {}
            for item in self.printList[(pageNum - 1) * numCols + colNum]:
                text = QSimpleRichText(''.join(item.textLines), \
                                       self.printFont)
                text.setWidth(colWidth - indent * item.level)
                xPos = colPos + indent * item.level
                text.draw(painter, xPos, yPos, drawRegion, QColorGroup())
                if addLines and item.level:
                    painter.drawLine(xPos - xLineDelta, yPos + linePos, \
                                     xPos - 2, yPos + linePos)
                    if item.firstSibling:
                        lineStarts[item.level] = yPos
                    if item.lastSibling:
                        painter.drawLine(xPos - xLineDelta, \
                                         lineStarts.get(item.level, initY), \
                                         xPos - xLineDelta, yPos + linePos)
                yPos += item.height
                if addLines and item.hasChildren and yLineDelta:
                    painter.drawLine(xPos + indent - xLineDelta, \
                                     yPos - yLineDelta, \
                                     xPos + indent - xLineDelta, yPos)
            if addLines:
                for level in self.pageLevelBreaks[(pageNum - 1) * numCols \
                                                  + colNum]:
                    xPos = colPos + indent * level - xLineDelta
                    painter.drawLine(xPos, lineStarts.get(level, initY), \
                                     xPos, yPos)
        footer = self.doc.getHeaderFooter(0)
        if footer:
            text = QSimpleRichText(footer, self.printFont)
            text.setWidth(painter.window().width())
            text.draw(painter, 0, painter.window().height() - 2 * \
                                  QFontMetrics(self.printFont).lineSpacing(), \
                      drawRegion, QColorGroup())
        docparam.fileInfoItem.data['Page_Number'] = ''

    def filePrint(self):
        """Print file starting from selected item"""
        self.printFont = self.getFontFromOptions('Print')
        if not self.printFont or self.option.boolData('PrintUseOutputFont'):
            outputViews = [view for view in self.dataOutSplit.children() if \
                           hasattr(view, 'updateView')]
            self.printFont = outputViews[0].font()
        useFile = self.printer.outputToFile()
        self.printer.setOutputFileName(os.path.splitext(self.doc.fileName)[0] +\
                                       '.ps')
        self.printer.setOutputToFile(useFile)
        self.printer.setFullPage(1)
        self.printer.setMinMax(1, 99)
        self.printer.setFromTo(1, 99)
        if not self.printer.setup(self):
            return
        paintMetrics = QPaintDeviceMetrics(self.printer)
        dpi = min(paintMetrics.logicalDpiX(), paintMetrics.logicalDpiY())
        scale = dpi / 100.0
        xMargin = int(max(self.printer.margins().width() / scale, \
                          72 * self.option.numData('HorizMargin', \
                                                TreeMainWin.minPrintMargin, \
                                                TreeMainWin.maxPrintMargin)))
        headerHt = self.doc.getHeaderFooter(1) and \
                   2 * QFontMetrics(self.printFont).lineSpacing() or 0
        footerHt = self.doc.getHeaderFooter(0) and \
                   2 * QFontMetrics(self.printFont).lineSpacing() or 0
        yMargin = int(max(self.printer.margins().height() / scale + \
                          headerHt + footerHt, \
                          72 * self.option.numData('VertMargin', \
                                                TreeMainWin.minPrintMargin, \
                                                TreeMainWin.maxPrintMargin)))
        pageSize = (int(paintMetrics.width() / scale), \
                    int(paintMetrics.height() / scale))
        pageFill = (pageSize[0] - 2 * xMargin, \
                    pageSize[1] - 2 * yMargin)
        # pageFill is size of tree content, pageAvail is content + header/footer
        pageAvail = (pageFill[0], pageFill[1] + headerHt + footerHt)
        yOffset = yMargin - headerHt
        numCols = self.option.intData('PrintNumCols', 1, 9)
        colSpacing = 72 * self.option.numData('PrintColSpace', 0, \
                                              TreeMainWin.maxPrintMargin)
        colWidth = (pageFill[0] - colSpacing * (numCols - 1)) / numCols

        item = self.treeView.currentTreeItem()
        includeRoot = self.option.boolData('PrintRoot')
        openOnly = self.option.boolData('PrintOpenOnly')
        indent = self.option.intData('PrintOffset', 0, \
                                     self.treeView.maxOffset)
        outGroup = item.outputItemList(includeRoot, openOnly)
        # work-arounds for height() problems with some tags
        brTagFix = (re.compile('(<br[ /]*?>|<BR[ /]*?>)'), u'<br>')
        hrTagFix = (re.compile('(<hr[ /]*?>|<HR[ /]*?>)'), u'<hr>')
        hTagFix = (re.compile('(</[hH][0-9] *>)'), u'\\1<br>')
        fixList = [brTagFix, hrTagFix, hTagFix]
        outGroup.setHeights(self.textHeight, colWidth, indent, fixList)
        outGroup.joinPrefixItems()
        # outGroup.addBreaks()
        if self.option.boolData('PrintLines'):
            self.pageLevelBreaks = []
        else:
            self.pageLevelBreaks = None
        self.printList = outGroup.splitPages(pageFill[1], self.pageLevelBreaks)
        for group in self.printList:
            group.addPrefix()
            group.addBreaks()

        minPg = max(self.printer.fromPage(), 1)
        maxPg = len(self.printList) / numCols + (len(self.printList) \
                                                 % numCols and 1 or 0)
        if 1 <= self.printer.toPage() < maxPg:
            maxPg = self.printer.toPage()
        docparam.fileInfoItem.data['Number_of_Pages'] = repr(maxPg)

        if self.option.boolData('PrintPrev'):
            dlg = PrintPrevDlg(minPg, maxPg, pageSize, pageAvail, \
                               (xMargin, yOffset), self.printPage, self, \
                               None, 1)
            dlg.setIcon(QPixmap(treeline_xpm))
            dlg.resize(self.option.intData('PrintPrevXSize', 10, 10000), \
                       self.option.intData('PrintPrevYSize', 10, 10000))
            if self.option.boolData('SaveWindowGeom'):
                dlg.move(self.option.intData('PrintPrevXPos', 0, 10000), \
                         self.option.intData('PrintPrevYPos', 0, 10000))
            result = dlg.exec_loop()
            if self.option.boolData('SaveWindowGeom'):
                self.option.changeData('PrintPrevXSize', dlg.width(), 1)
                self.option.changeData('PrintPrevYSize', dlg.height(), 1)
                self.option.changeData('PrintPrevXPos', dlg.x(), 1)
                self.option.changeData('PrintPrevYPos', dlg.y(), 1)
                self.option.writeChanges()
            if result != QDialog.Accepted:
                docparam.fileInfoItem.data['Number_of_Pages'] = ''
                return
        self.printer.setFullPage(0)    # fix MS margin problem?
        painter = QPainter()
        if not painter.begin(self.printer):
            docparam.fileInfoItem.data['Number_of_Pages'] = ''
            return
        if self.printer.pageOrder() == QPrinter.FirstPageFirst:
            seq = range(minPg, maxPg + 1)
        else:
            seq = range(maxPg, minPg - 1, -1)
        printMargin = ((pageSize[0] * scale - paintMetrics.width()) / 2, \
                       (pageSize[1] * scale - paintMetrics.height()) / 2)
        painter.setWindow(0, 0, pageAvail[0], pageAvail[1])
        painter.setViewport(xMargin * scale - printMargin[0], \
                            yOffset * scale - printMargin[1], \
                            pageAvail[0] * scale, pageAvail[1] * scale)
        self.printPage(seq.pop(0), painter)
        for num in seq:
            self.printer.newPage()
            self.printPage(num, painter)
        painter.end()
        docparam.fileInfoItem.data['Number_of_Pages'] = ''

    def textHeight(self, text, width):
        """Calculates height of rich text in printer font"""
        obj = QSimpleRichText(text, self.printFont)
        obj.setWidth(width)
        return obj.height()

    def filePrintOpt(self):
        """Set margins and page size for printing"""
        self.headerDlg = None
        self.printFont = None
        origPageSize = self.option.strData('PrintPageSize', 0)
        dlg = OptionDlg(self.option, self, None, 1)
        if sys.platform == 'win32':
            dlg.setCaption('Print Options  (PyQt)')
        else:
            dlg.setCaption('Print Options')
        dlg.setIcon(QPixmap(treeline_xpm))

        dlg.startGroupBox('Features')
        OptionDlgBool(dlg, 'PrintPrev', 'Preview before printing')
        OptionDlgBool(dlg, 'PrintRoot', 'Include root node')
        OptionDlgBool(dlg, 'PrintLines', 'Draw lines to children')
        OptionDlgBool(dlg, 'PrintOpenOnly', 'Print only open node children')

        dlg.startGroupBox('Printer Font')
        outputFontOpt = OptionDlgBool(dlg, 'PrintUseOutputFont', \
                                      'Use Data Output font')
        fontButtonOpt = OptionDlgPush(dlg, 'Set Printer Font', \
                                      self.setPrintFont)
        fontButtonOpt.control.setDisabled(outputFontOpt.control.isChecked())
        self.connect(outputFontOpt.control, SIGNAL('toggled(bool)'), \
                     fontButtonOpt.control, SLOT('setDisabled(bool)'))

        dlg.startGroupBox('Header/Footer')
        headerButtonOpt = OptionDlgPush(dlg, 'Set Header and Footer', \
                                        self.setHeader)
        dlg.endGroupBox()

        OptionDlgRadio(dlg, 'PrintPageSize', 'Default Page Size', \
                       [('letter', 'Letter'), ('a4', 'A4')])

        dlg.startGroupBox('Columns')
        OptionDlgInt(dlg, 'PrintNumCols', 'Number of columns', 1, \
                     TreeMainWin.maxNumCol)
        OptionDlgDbl(dlg, 'PrintColSpace', 'Space between columns (inch)', \
                     0, TreeMainWin.maxPrintMargin)

        dlg.startGroupBox('Offsets')
        OptionDlgInt(dlg, 'PrintOffset', 'Child indent offset (points)', \
                     0, TreeView.maxOffset)
        OptionDlgDbl(dlg, 'HorizMargin', 'Horizontal page margins (inch)', \
                     TreeMainWin.minPrintMargin, TreeMainWin.maxPrintMargin)
        OptionDlgDbl(dlg, 'VertMargin', 'Vertical page margins (inch)', \
                     TreeMainWin.minPrintMargin, TreeMainWin.maxPrintMargin)
        
        if dlg.exec_loop() == QDialog.Accepted:
            if self.printFont:
                self.saveFontToOptions(self.printFont, 'Print')
            pageSize = self.option.strData('PrintPageSize', 0)
            if pageSize != origPageSize:
                self.printer.setPageSize(eval('QPrinter.%s' % \
                                              pageSize.capitalize()))
            self.option.writeChanges()
            if self.headerDlg:
                docparam.undoStore.addFormatUndo(docparam.nodeFormats, \
                                                 docparam.fileInfoItem.\
                                                          nodeFormat, \
                                                 {}, {}, self.doc.root, \
                                                 self.doc.selectedNodes)
                docparam.nodeFormats.removeQuiet(docparam.fileInfoItem.\
                                                 nodeFormat)
                docparam.fileInfoItem.nodeFormat = self.headerDlg.fileInfoFormat
                docparam.nodeFormats.append(docparam.fileInfoItem.nodeFormat)
                docparam.nodeFormats.updateAllLineFields()
                docparam.modified = 1
                self.updateCmdAvail()
        self.headerDlg = None

    def setPrintFont(self):
        """Show dialog for setting custom printer font"""
        printFont = self.getFontFromOptions('Print')
        if not printFont:
            outputViews = [view for view in self.dataOutSplit.children() if \
                           hasattr(view, 'updateView')]
            printFont = outputViews[0].font()
        font, ok = QFontDialog.getFont(printFont, self)
        if ok and font != printFont:
            self.printFont = font

    def setHeader(self):
        """Show dialog for setting header and footer"""
        if not self.headerDlg:
            self.headerDlg = PrintHeaderDlg(self, self, None, 1)
        if not self.headerDlg.exec_loop() == QDialog.Accepted:
            self.headerDlg = None
        # else:
            # format = docparam.fileInfoItem.nodeFormat
            # format.lineList = []
            # lines = self.headerDlg.getTextLines()
            # for text, pos in zip(lines, range(len(lines))):
                # if text:
                    # format.insertLine(text, pos)
            # if not docparam.nodeFormats.findFormat(format.name):
                # docparam.nodeFormats.append(format)
            # docparam.modified = 1
            # self.updateCmdAvail()

    def editUndo(self):
        """Undo the previous action"""
        newSelect = docparam.undoStore.undo(docparam.redoStore, \
                                            self.doc.selectedNodes)
        self.updateViews(newSelect)
        if self.setTypeDlg and self.setTypeDlg.isVisible():
            self.setTypeDlg.loadList()
        self.loadTypeSubMenu()

    def editRedo(self):
        """Redo the previous undo"""
        newSelect = docparam.redoStore.undo(docparam.undoStore, \
                                            self.doc.selectedNodes)
        self.updateViews(newSelect)
        if self.setTypeDlg and self.setTypeDlg.isVisible():
            self.setTypeDlg.loadList()
        self.loadTypeSubMenu()

    def editCut(self):
        """Cut the branch or text to the clipboard"""
        widget = self.focusWidgetWithAttr('copyAvail')
        if self.treeView.hasFocus() or not widget or not widget.copyAvail():
            self.editCopyTree()
            self.editDelete()
        else:
            widget.cut()

    def editCopyTree(self):
        """Copy the tree branch to the clipboard"""
        if not self.doc.selectedNodes:
            item = self.treeView.currentTreeItem()
        elif len(self.doc.selectedNodes) > 1:
            docparam.nodeFormats.addIfMissing(TreeMainWin.copyFormat)
            item = TreeItem(None, TreeMainWin.copyFormat)
            for node in self.doc.selectedNodes:
                item.childList.append(copy.copy(node))
                item.childList[-1].parent = item
        else:
            item = self.doc.selectedNodes[0]
        clip = QApplication.clipboard()
        if qVersion()[0] >= '3' and clip.supportsSelection():
            clip.setSelectionMode(1)
            textList = []
            if len(self.doc.selectedNodes) > 1:
                for node in self.doc.selectedNodes:
                    textList.extend(node.exportToText())
            else:
                textList = item.exportToText()
            clip.setText(u'\n'.join(textList))
            clip.setSelectionMode(0)
        clip.setText(u'\n'.join(item.branchXml([TreeMainWin.copyFormat])))
        docparam.nodeFormats.removeQuiet(TreeMainWin.copyFormat)


    def editCopy(self):
        """Copy the branch or text to the clipboard"""
        split = self.rightStack.visibleWidget()
        if split == self.dataOutSplit:  # check for select in dataOut (no focus)
            views = [view for view in split.children() if \
                     hasattr(view, 'copyAvail') and view.copyAvail()]
            if views:
                views[0].copy()
                return
        widget = self.focusWidgetWithAttr('copyAvail')
        if self.treeView.hasFocus() or not widget or not widget.copyAvail():
            self.editCopyTree()
        else:
            widget.copy()

    def editCopyText(self):
        """Copy node title text to the clipboard"""
        if self.doc.selectedNodes:
            titles = [item.title() for item in self.doc.selectedNodes]
        else:
            titles = [self.treeView.currentTreeItem().title()]
        clip = QApplication.clipboard()
        if qVersion()[0] >= '3' and clip.supportsSelection():
            clip.setSelectionMode(1)
            clip.setText(u'\n'.join(titles))
            clip.setSelectionMode(0)
        clip.setText(u'\n'.join(titles))

    def editPaste(self):
        """Paste items or text from the clipboard"""
        try:
            text = unicode(QApplication.clipboard().text())
        except UnicodeError:
            return
        item = self.doc.readXmlString(text, self.treeView.hasFocus())
        if self.treeView.hasFocus():
            if item:
                if item.data:
                    itemList = [item]
                else:            # blank item is dummy root of multi-select
                    itemList = item.childList
                parent = self.treeView.currentTreeItem()
                docparam.undoStore.addChildListUndo(parent, \
                                                    self.doc.selectedNodes)
                for node in itemList:
                    parent.addTree(node)
                parent.open = 1
                docparam.nodeFormats.removeQuiet(TreeMainWin.copyFormat)
                self.updateViews(itemList)
            else:
                print 'Error reading XML string'
        else:
            if item:
                text = item.title()
            widget = self.focusWidgetWithAttr('pasteText')
            if widget:
                widget.pasteText(text)

    def editPasteText(self):
        """Paste text from the clipboard"""
        try:
            text = unicode(QApplication.clipboard().text())
        except UnicodeError:
            return
        item = self.doc.readXmlString(text)
        if item and item.data:
            text = item.title()
        elif item and item.childList:
            text = item.childList[0].title()
        else:
            text = unicode(QApplication.clipboard().text()).\
                   split(u'\n', 1)[0].strip()
        if self.treeView.hasFocus():
            item = self.treeView.currentTreeItem()
            docparam.undoStore.addDataUndo(item, self.doc.selectedNodes)
            item.setTitle(text)
            docparam.modified = 1
            self.treeView.updateTree(item)
            self.updateCmdAvail()
        else:
            widget = self.focusWidgetWithAttr('pasteText')
            if widget:
                widget.pasteText(text)

    def editInBefore(self):
        """Insert new sibling before selection"""
        sibling = self.treeView.currentTreeItem()
        docparam.undoStore.addChildListUndo(sibling.parent, \
                                            self.doc.selectedNodes)
        newItem = sibling.insertSibling()
        if newItem:
            if self.option.boolData('RenameNewNodes'):
                self.updateViews(newItem)
                self.treeView.editRename()
            else:
                self.updateViews(self.doc.selectedNodes)

    def editInAfter(self):
        """Insert new sibling after selection"""
        sibling = self.treeView.currentTreeItem()
        docparam.undoStore.addChildListUndo(sibling.parent, \
                                            self.doc.selectedNodes)
        newItem = sibling.insertSibling(inAfter=1)
        if newItem:
            if self.option.boolData('RenameNewNodes'):
                self.updateViews(newItem)
                self.treeView.editRename()
            else:
                self.updateViews(self.doc.selectedNodes)

    def editAddChild(self):
        """Add a new child to the selected parent"""
        parent = self.treeView.currentTreeItem()
        docparam.undoStore.addChildListUndo(parent, self.doc.selectedNodes)
        newItem = parent.addChild()
        if newItem:
            parent.open = 1
            if self.option.boolData('RenameNewNodes'):
                self.updateViews(newItem)
                self.treeView.editRename()
            else:
                self.updateViews(self.doc.selectedNodes)

    def editDelete(self):
        """Delete the selected items"""
        nextSel = filter(None, [item.parent for item in \
                                self.doc.selectedNodes])
        nextSel.extend(filter(None, [item.prevSibling() for item in \
                                     self.doc.selectedNodes]))
        nextSel.extend(filter(None, [item.nextSibling() for item in \
                                     self.doc.selectedNodes]))
        docparam.undoStore.addParentListUndo(self.doc.selectedNodes, \
                                             self.doc.selectedNodes)
        for item in self.doc.selectedNodes:
            item.delete()
        if not nextSel:
            return
        while nextSel[-1] in self.doc.selectedNodes:
            del nextSel[-1]
        self.updateViews(nextSel[-1])

    def sortSelectionList(self):
        """Replace core selection list with a sorted version"""
        self.doc.selectedNodes = [item for item in \
                                  self.doc.root.descendantList(0) \
                                  if item.viewData and \
                                  item.viewData.isSelected()]

    def editIndent(self):
        """Indent the selected items"""
        self.sortSelectionList()
        parentList = [item.parent for item in self.doc.selectedNodes]
        siblingList = [item.prevSibling() for item in self.doc.selectedNodes]
        docparam.undoStore.addChildListUndo(parentList + siblingList, \
                                            self.doc.selectedNodes)
        for item in self.doc.selectedNodes:
            if item.indent():
                item.parent.open = 1
        self.updateViews(self.doc.selectedNodes)

    def editUnindent(self):
        """Unindent the selected item"""
        self.sortSelectionList()
        parentList = [item.parent for item in self.doc.selectedNodes]
        gpList = [item.parent for item in parentList]
        docparam.undoStore.addChildListUndo(parentList + gpList, \
                                            self.doc.selectedNodes)
        self.doc.selectedNodes.reverse()
        for item in self.doc.selectedNodes:
            item.unindent()
        self.doc.selectedNodes.reverse()
        self.updateViews(self.doc.selectedNodes)

    def editMoveUp(self):
        """Move the selected item up"""
        self.sortSelectionList()
        docparam.undoStore.addParentListUndo(self.doc.selectedNodes, \
                                             self.doc.selectedNodes, 1)
        for item in self.doc.selectedNodes:
            item.move(-1)
        self.updateViews(self.doc.selectedNodes)

    def editMoveDown(self):
        """Move the selected item down"""
        self.sortSelectionList()
        self.doc.selectedNodes.reverse()
        docparam.undoStore.addParentListUndo(self.doc.selectedNodes, \
                                             self.doc.selectedNodes, 1)
        for item in self.doc.selectedNodes:
            item.move(1)
        self.doc.selectedNodes.reverse()
        self.updateViews(self.doc.selectedNodes)

    def viewToggleToolbar(self,index):
        """Toggle display of toolbar with index == itemId"""
        if self.toolbars[index].isVisible():
            self.toolbars[index].hide()
            self.viewToolbarsMenu.setItemChecked(index, False)
        else:
            self.toolbars[index].show()
            self.viewToolbarsMenu.setItemChecked(index, True)

    def viewSelect(self, action):
        """Show right view given by action"""
        if not self.blockViewChg:
            # avoids multiple raise/updates
            self.blockViewChg = 1
            self.rightStack.raiseWidget(self.actToView[action])
            self.blockViewChg = 0
            # self.showHideChildView()

    def viewTabSelect(self, viewNum):
        """Show right view given by tab number"""
        if not self.blockViewChg:
            # avoids multiple raise/updates
            self.blockViewChg = 1
            self.rightStack.raiseWidget(viewNum)
            self.blockViewChg = 0
            # self.showHideChildView()

    def viewChildren(self, action):
        """Set to view item with or without children"""
        self.showItemChildren = self.childActToBool[action]
        # self.showHideChildView()
        for split in [self.dataOutSplit, self.dataEditSplit, \
                      self.titleListSplit]:
            for view in split.children():
                if hasattr(view, 'showChildren') and view.showChildren:
                    if self.showItemChildren:
                        view.show()
                    else:
                        view.hide()
            split.refresh()
        self.updateRightView()

    def showHideChildView(self):
        """Show or hide lower half of right view based on showItemChildren"""
        print 'Should not be called!!!!!!!!'
        childView = [view for view in \
                     self.rightStack.visibleWidget().children() if \
                     hasattr(view, 'showChildren') and view.showChildren][0]
        if self.showItemChildren:
            childView.show()
        else:
            childView.hide()
        self.updateRightView()

    def viewFocusTree(self):
        """Change focus to tree view"""
        self.treeView.setFocus()

    def viewDataPageUp(self):
        """Page up the right-hand view"""
        split = self.rightStack.visibleWidget()
        for view in split.children():
            if hasattr(view, 'showChildren') and view.showChildren:
                view.scrollPage(-1)

    def viewDataPageDown(self):
        """Page down the right-hand view"""
        split = self.rightStack.visibleWidget()
        for view in split.children():
            if hasattr(view, 'showChildren') and view.showChildren:
                view.scrollPage(1)

    def dataTypeChange(self, id):
        """Change type based on submenu selection"""
        names = docparam.nodeFormats.nameList()[:]
        names.sort()
        newFormat = docparam.nodeFormats.findFormat(names[id])
        itemList = self.doc.selectedNodes
        if not itemList:
            itemList = [self.treeView.currentTreeItem()]
        docparam.undoStore.addTypeUndo(itemList, self.doc.selectedNodes)
        for item in itemList:
            item.nodeFormat = newFormat
        docparam.modified = 1
        self.updateViews(self.doc.selectedNodes)

    def dataTypeMenuCheck(self):
        """Set check box in type menu to the current type"""
        itemList = self.doc.selectedNodes
        if not itemList:
            itemList = [self.treeView.currentTreeItem()]
        currentTypes = []
        for item in itemList:
            if item.nodeFormat.name not in currentTypes:
                currentTypes.append(item.nodeFormat.name)
        names = docparam.nodeFormats.nameList()[:]
        names.sort()
        for index, name in zip(range(len(names)), names):
            id = self.typeSubMenu.idAt(index)
            if id != -1:
                self.typeSubMenu.setItemChecked(id, name in currentTypes)

    def dataSet(self, show):
        """Show dialog for setting item data types"""
        if show:
            if not self.setTypeDlg:
                self.setTypeDlg = TypeSetDlg(self, self)
                self.connect(self.setTypeDlg, PYSIGNAL('viewClosed'), \
                             self.dataSetAct.setOn)
            else:
                self.setTypeDlg.loadList()
            self.setTypeDlg.setCurrentSel()
            self.setTypeDlg.updateDlg()
            self.setTypeDlg.show()
        else:
            self.setTypeDlg.hide()

    def dataConfig(self):
        """Show dialog for modifying data types"""
        currentType = self.treeView.currentTreeItem().nodeFormat.name
        dlg = ConfigDlg(self, currentType, self, None, 1)
        if dlg.exec_loop() == QDialog.Accepted:
            docparam.undoStore.addFormatUndo(docparam.nodeFormats, \
                                             docparam.fileInfoItem.nodeFormat, \
                                             dlg.fieldRenameDict, \
                                             dlg.typeRenameDict, \
                                             self.doc.root, \
                                             self.doc.selectedNodes)
            docparam.fileInfoItem.nodeFormat = dlg.fileInfoFormat
            if dlg.fieldRenameDict:
                docparam.nodeFormats.renameFields(dlg.fieldRenameDict, \
                                                  self.doc.root)
            docparam.nodeFormats.update(dlg.formatList, dlg.typeRenameDict)
            docparam.nodeFormats.updateAllLineFields()
            docparam.modified = 1
            self.updateViews(self.treeView.currentTreeItem())
            if self.setTypeDlg and self.setTypeDlg.isVisible():
                self.setTypeDlg.loadList()
            self.loadTypeSubMenu()

    def dataSortChild(self):
        """Sort children of selected item with chosen keys"""
        if not self.doc.selectedNodes:
            return
        children = []
        for item in self.doc.selectedNodes:
            children.extend(item.childList)
        fieldList = self.doc.selectedNodes[0].commonFields(children)
        if fieldList:           # has common fields
            dlg = FieldSelectDlg(fieldList, 'Sort Keys', \
                                 'Select sort keys in order', 1, self, None, 1)
            if dlg.exec_loop() != QDialog.Accepted:
                return
            docparam.sortFields = dlg.getSelList()
        else:                            # no common fields
            if QMessageBox.information(self, 'TreeLine', \
                                  'No common fields found\nSort by title?', \
                                  'Yes', 'No', QString.null, 0, 1) ==  1:
                return
            docparam.sortFields = [('', 1)]             # sort by title
        docparam.undoStore.addChildListUndo(self.doc.selectedNodes, \
                                            self.doc.selectedNodes)
        for item in self.doc.selectedNodes:
            item.sortChildren()
        self.updateViews(self.doc.selectedNodes)

    def dataSortType(self):
        """Sort descendants of a type with keys"""
        if not self.doc.selectedNodes:
            return
        typeList = []
        for item in self.doc.selectedNodes:
            for type in item.descendTypes():
                if type.name not in typeList:
                    typeList.append(type.name)
        if len(typeList) == 1:
            selType = typeList[0]
        else:
            selType, ok = QInputDialog.getItem('Type to Sort', \
                          'Select descendant node type to sort', \
                          QStringList.split(u' ', u' '.join(typeList)), \
                          0, 0, self)
            if not ok:
                return
        selType = str(selType)
        format = docparam.nodeFormats.findFormat(selType)
        fieldList = format.fieldNames()
        dlg = FieldSelectDlg(fieldList, 'Sort Keys', \
                             'Select sort keys in order', 1, self, None, 1)
        if dlg.exec_loop() != QDialog.Accepted:
            return
        docparam.sortFields = dlg.getSelList()
        undoList = []
        for item in self.doc.selectedNodes:
            undoList.extend([parent for parent in item.descendantList(1) \
                             if parent.childList])
        docparam.undoStore.addChildListUndo(undoList, self.doc.selectedNodes)
        for item in self.doc.selectedNodes:
            item.sortType(format)
        self.updateViews(self.doc.selectedNodes)

    def dataSortBranch(self):
        """Sort all children of selected item by title only"""
        dlg = RadioChoiceDlg('Branch Sort', 'Choose Sort Direction', \
                             [('&Ascending', 1), ('&Descending', 0)], \
                             self, None, 1)
        if dlg.exec_loop() != QDialog.Accepted:
            return
        docparam.sortFields = [('', dlg.getResult())]
        undoList = []
        for item in self.doc.selectedNodes:
            undoList.extend([parent for parent in item.descendantList(1) \
                             if parent.childList])
        docparam.undoStore.addChildListUndo(undoList, self.doc.selectedNodes)
        for item in self.doc.selectedNodes:
            item.sortBranch()
        self.updateViews(self.doc.selectedNodes)

    def dataFilter(self):
        """Create a new file with filtered contents"""
        if not self.doc.selectedNodes:
            return
        typeList = []
        for item in self.doc.selectedNodes:
            for type in item.descendTypes():
                if type.name not in typeList:
                    typeList.append(type.name)
        if len(typeList) > 1:
            typeList = QStringList.split(' ', ' '.join(typeList))
            caption = 'Filter Data'
            if sys.platform == 'win32':
                caption += '  (PyQt)'
            type, ok = QInputDialog.getItem(caption, 'Select data type', \
                                            typeList, 0, 0, self)
            if not ok:
                return
            type = str(type)
        else:
            type = typeList[0]
        format = docparam.nodeFormats.findFormat(type)
        dlg = FilterDlg(format.fieldNames(), self, None, 1)
        if dlg.exec_loop() != QDialog.Accepted:
            return
        expr = dlg.exprText()
        if expr and self.savePrompt():
            docparam.modified = 0
            for item in self.doc.selectedNodes:
                item.filterDescendants(format, expr, self.editFormats)
            if docparam.modified:
                dir, name = os.path.split(self.doc.fileName)
                self.doc.fileName = os.path.splitext(self.doc.fileName)[0] \
                                    + '_filter.trl'
                self.setMainCaption()
                self.updateViews(self.doc.selectedNodes)
            else:
                self.statusBar().message('No nodes were filtered',  4000)

    def dataEditField(self):
        """Edit a child field in all selected nodes"""
        nodes = self.doc.selectedNodes
        if not nodes:
            return
        fieldList = nodes[0].commonFields(nodes)
        if fieldList:           # has common fields
            dlg = EditFieldsDlg(fieldList, self, None, 1)
            if dlg.exec_loop() != QDialog.Accepted:
                return
            docparam.undoStore.addDataUndo(nodes, nodes)
            for item in nodes:
                item.editFields(dlg.resultDict)
            self.updateViews(nodes)
        else:
            QMessageBox.warning(self, 'TreeLine', \
                                'No common fields to set')

    def dataNumbering(self):
        """Add numbering to a data field"""
        item = self.doc.selectedNodes[0]
        dlg = NumberingDlg(item.branchFields(), item.maxDescendLevel(), \
                           self, None, 1)
        if dlg.exec_loop() != QDialog.Accepted:
            return
        docparam.undoStore.addBranchUndo(item, self.doc, [item])
        if dlg.currentStyle == NumberingDlg.outlineType:
            item.addNumbering(dlg.getField(), dlg.currentFormat, \
                              dlg.includeRoot(), 0, 0, dlg.startNumber())
        elif dlg.currentStyle == NumberingDlg.sectionType:
            item.addNumbering(dlg.getField(), dlg.currentFormat, \
                              dlg.includeRoot(), 1, 0, dlg.startNumber())
        else:        # singleType
            item.addNumbering(dlg.getField(), dlg.currentFormat, \
                              0, 0, 1, dlg.startNumber())
        self.updateViews(self.doc.selectedNodes)

    def dataAddCat(self):
        """Add child's category items as a new child level"""
        if not self.doc.selectedNodes:
            return
        children = []
        for item in self.doc.selectedNodes:
            children.extend(item.childList)
        fieldList = self.doc.selectedNodes[0].commonFields(children)
        if fieldList:           # has common fields
            dlg = FieldSelectDlg(fieldList, 'Category Fields', \
                                 'Select fields for new level', 0, self, \
                                 None, 1)
            if dlg.exec_loop() != QDialog.Accepted:
                return
            docparam.undoStore.addBranchUndo(self.doc.selectedNodes, self.doc, \
                                             self.doc.selectedNodes)
            for item in self.doc.selectedNodes:
                item.addChildCat(dlg.getSelList())
            self.updateViews(self.doc.selectedNodes)
        else:
            QMessageBox.warning(self, 'TreeLine', \
                                'Cannot expand without common fields')

    def dataFlatCat(self):
        """Collapse data by merging fields"""
        docparam.undoStore.addBranchUndo(self.doc.selectedNodes, self.doc, \
                                         self.doc.selectedNodes)
        for item in self.doc.selectedNodes:
            item.flatChildCat()
        self.updateViews(self.doc.selectedNodes)

    def dataArrangeRef(self):
        """Arrange data using parent references"""
        item = self.doc.selectedNodes[0]
        fieldList = item.childList[0].nodeFormat.fieldNames()
        refField, ok = QInputDialog.getItem('Reference Field', \
                       'Select field with parent references', \
                       QStringList.split(u' ', u' '.join(fieldList)), \
                       0, 0, self)
        if not ok:
            return
        docparam.undoStore.addBranchUndo(item, self.doc, [item])
        item.arrangeByRef(str(refField))
        self.updateViews(self.doc.selectedNodes)

    def dataFlatRef(self):
        """Collapse data after adding references to parents"""
        item = self.doc.selectedNodes[0]
        dlg = FieldEntry('Flatten by Reference', \
                         'Enter new field name for parent references:', '', \
                         [], self, None, 1)
        if dlg.exec_loop() != QDialog.Accepted:
            return
        docparam.undoStore.addBranchUndo(item, self.doc, [item])
        item.flatByRef(dlg.text)
        self.updateViews(self.doc.selectedNodes)

    def dataUpdate(self):
        """Update with new fields from reference file"""
        fileName = str(QFileDialog.getOpenFileName(None, \
                       'TreeLine Files (*.trl *.xml)', self, None, \
                       'Open Reference File'))
        if fileName:
            try:
                f = codecs.open(fileName, 'r', 'utf-8')
                refRoot = self.doc.readXmlString('\n'.join(f.readlines()))
            except IOError:
                QMessageBox.warning(self, 'TreeLine', \
                                    'Error - could not read file')
                return
            if not refRoot:
                QMessageBox.warning(self, 'TreeLine', \
                                    'Error - invalid file')
                return
            docparam.undoStore.addBranchUndo(self.doc.root, self.doc, \
                                             self.doc.selectedNodes)
            total = [0, 0, 0]
            for item in self.doc.selectedNodes:
                result = item.updateByRef(refRoot)
                total[0] += result[0]
                total[1] += result[1]
                total[2] += result[2]
            self.updateViews(self.doc.selectedNodes)
            result = '%d new entries in %d new fields in %s changed types' % \
                     tuple(total)
            self.statusBar().message(result, 6000)

    def toolsExpand(self):
        """Expand all children of selected item"""
        for item in self.doc.selectedNodes:
            item.openBranch(1)
        self.updateViews(self.doc.selectedNodes)

    def toolsCollapse(self):
        """Collapse all children of selected item"""
        for item in self.doc.selectedNodes:
            item.openBranch(0)
        self.updateViews(self.doc.selectedNodes)

    def toolsFind(self, show):
        """Find item matching text string"""
        if show:
            if not self.FindDlg:
                self.FindDlg = FindTextEntry(self, self)
                self.connect(self.FindDlg, PYSIGNAL('viewClosed'), \
                             self.toolsFindAct.setOn)
            self.FindDlg.entry.selectAll()
            self.FindDlg.show()
        else:
            self.FindDlg.hide()

    def toolsSpellCheck(self):
        """Spell check the tree's text data strting in the selected branch"""
        spellPath = self.option.strData('SpellCheckPath', 1)
        try:
            spCheck = SpellCheck(spellPath)
        except SpellCheckError:
            if sys.platform == 'win32':
                ans = QMessageBox.warning(self, 'TreeLine Spell Check Error', \
                             'Could not find either aspell.exe or ispell.exe\n'
                             'Manually locate?', '&Browse', '&Cancel', \
                             QString.null, 0, 1)
                if ans != 0:
                    return
                path = str(QFileDialog.getOpenFileName(QString.null, \
                                            'Program (*.exe)', self, '', \
                                            'Locate aspell.exe or ipsell.exe'))
                if path:
                    path = path[:-4]
                    if ' ' in path:
                        path = '"%s"' % path
                    self.option.changeData('SpellCheckPath', path, 1)
                    self.option.writeChanges()
                    self.toolsSpellCheck()
                return
            else:
                QMessageBox.warning(self, 'TreeLine', \
                                    'Spell Check Error\n'
                                    'Make sure aspell or ispell is installed')
                return
        dlg = SpellCheckDlg(spCheck, self, None, 1)
        startItem = self.treeView.currentTreeItem()
        for item in startItem.descendantList(1):
            for field in item.nodeFormat.fieldNames():
                text = item.data.get(field, '')
                if text:
                    textLines = text.split('\n')
                    for i, line in zip(range(len(textLines)), textLines):
                        tmpIgnoreList = []
                        results = spCheck.checkLine(line)
                        while results:
                            self.treeView.changeSelect(item, 1)
                            dlg.setWord(line, results)
                            if not dlg.newLine and \
                               dlg.exec_loop() != QDialog.Accepted:
                                spCheck.close()
                                return
                            if dlg.ignoreWord:
                                tmpIgnoreList.append(dlg.ignoreWord)
                            if dlg.newLine:
                                docparam.undoStore.addDataUndo(item, [item], 1)
                                line = dlg.newLine
                                textLines[i] = line
                                item.data[field] = '\n'.join(textLines)
                                docparam.modified = 1
                                self.updateViews(self.doc.selectedNodes)
                            results = spCheck.checkLine(line, tmpIgnoreList)
        spCheck.close()
        if startItem.parent:
            ans = QMessageBox.information(self, 'TreeLine Spell Check', \
                                          'Finished checking the branch\n'
                                          'Continue from the root branch?', \
                                          '&Yes', '&No', QString.null, 0, 1)
            if ans == 0:
                self.treeView.changeSelect(self.doc.root)
                self.toolsSpellCheck()
        else:
            QMessageBox.information(self, 'TreeLine Spell Check', \
                                    'Finished checking the branch')
        self.treeView.changeSelect(startItem, 1)

    def toolsRemXslt(self):
        """Delete reference to XSLT export"""
        if self.doc.xlstLink:
            docparam.undoStore.addParamUndo([(self.doc, 'xlstLink')], \
                                            self.doc.selectedNodes)
            self.doc.xlstLink = ''
            docparam.modified = 1
            self.updateCmdAvail()

    def toolsGenOpt(self):
        """Set user preferences for all files"""
        outputViews = [view for view in self.dataOutSplit.children() if \
                       hasattr(view, 'updateView')]
        self.outputFont = outputViews[0].font()
        oldAutoSave = self.option.intData('AutoSaveMinutes', 0, 999)
        dlg = OptionDlg(self.option, self, None, 1)
        dlg.setIcon(QPixmap(treeline_xpm))
        if sys.platform == 'win32':
            dlg.setCaption('General Options  (PyQt)')
        else:
            dlg.setCaption('General Options')
        dlg.startGroupBox('Startup Condition')
        OptionDlgBool(dlg, 'AutoFileOpen', 'Automatically open last file used')
        OptionDlgBool(dlg, 'StartShowChildren', \
                      'Show children in right-hand view')
        OptionDlgBool(dlg, 'PersistTreeState', \
                           'Restore view states of recent files')
        OptionDlgBool(dlg, 'SaveWindowGeom', 
                      'Restore window geometry from last exit')
        dlg.startGroupBox('Features Available')
        OptionDlgBool(dlg, 'ClickRename', 'Click item to rename')
        OptionDlgBool(dlg, 'DragTree', 'Tree drag && drop available')
        OptionDlgBool(dlg, 'InsertOnEnter', 'Insert node with enter')
        OptionDlgBool(dlg, 'RenameNewNodes', 'Rename new nodes when created')
        OptionDlgBool(dlg, 'OpenSearchNodes', 'Automatically open search nodes')
        OptionDlgBool(dlg, 'CompressNewFiles', \
                     'Set new files to compressed by default')
        OptionDlgBool(dlg, 'HtmlNewFields', \
                      'New fields default to HTML content')
        dlg.startNewColumn()
        dlg.startGroupBox('Undo Memory')
        OptionDlgInt(dlg, 'UndoLevels', 'Number of undo levels ', 0, 99)
        dlg.startGroupBox('Auto Save')
        OptionDlgInt(dlg, 'AutoSaveMinutes', \
                     'Minutes between saves \n(Set to 0 to disable)', 0, 999)
        dlg.startGroupBox('Data Editor Formats')
        OptionDlgStr(dlg, 'EditNumFormat', 'Numbers')
        OptionDlgStr(dlg, 'EditDateFormat', 'Dates')
        OptionDlgStr(dlg, 'EditTimeFormat', 'Times')
        dlg.startGroupBox('Appearance')
        OptionDlgInt(dlg, 'IndentOffset', 'Child indent offset (points)', \
                     0, TreeView.maxOffset)
        OptionDlgInt(dlg, 'MaxEditLines', 'Max lines in data edit box', \
                     1, TreeMainWin.maxNumLines)
        OptionDlgPush(dlg, 'Set Data Output Font', self.setOutputFont)
        dlg.startNewColumn()   # add space if necessary
        if dlg.exec_loop() == QDialog.Accepted:
            if self.outputFont != outputViews[0].font():
                for view in outputViews:
                    view.setFont(self.outputFont)
                self.saveFontToOptions(self.outputFont, 'Output')
            if not self.option.boolData('PersistTreeState'):
                self.treeStates.clear()
            self.option.writeChanges()
            if oldAutoSave != self.option.intData('AutoSaveMinutes', 0, 999):
                self.resetAutoSave()
            docparam.undoStore.levels = self.option.intData('UndoLevels', 0, 99)
            docparam.redoStore.levels = self.option.intData('UndoLevels', 0, 99)
            self.treeView.setTreeStepSize(self.option.intData('IndentOffset', \
                                                       0, TreeView.maxOffset))
            self.updateEditFormats()
            self.updateViews(self.treeView.currentTreeItem())

    def setOutputFont(self):
        """Show dialog for setting custom output font"""
        font, ok = QFontDialog.getFont(self.outputFont, self)
        if ok:
            self.outputFont = font

    def toolsFileOpt(self):
        """Set file preferences"""
        self.option.addData('SpaceBetween', \
                            docparam.spaceBetween and 'yes' or 'no', 0)
        self.option.addData('FormHtml', docparam.formHtml and 'yes' or 'no', 0)
        self.option.addData('CompressFile', \
                            self.doc.compressFile and 'yes' or 'no', 0)
        dlg = OptionDlg(self.option, self, None, 1)
        dlg.setIcon(QPixmap(treeline_xpm))
        if sys.platform == 'win32':
            dlg.setCaption('File Options  (PyQt)')
        else:
            dlg.setCaption('File Options')
        dlg.startGroupBox('Output Formating')
        OptionDlgBool(dlg, 'SpaceBetween', 'Add blank lines between nodes', 0)
        OptionDlgBool(dlg, 'FormHtml', 'Allow HTML rich text in formats', 0)
        dlg.startGroupBox('File Storage')
        OptionDlgBool(dlg, 'CompressFile', 'Use file compression', 0)
        if dlg.exec_loop() == QDialog.Accepted:
            space = self.option.boolData('SpaceBetween')
            html = self.option.boolData('FormHtml')
            compress = self.option.boolData('CompressFile')
            if space != docparam.spaceBetween or html != docparam.formHtml or \
                     compress != self.doc.compressFile:
                docparam.undoStore.addParamUndo([(docparam, 'spaceBetween'), \
                                                 (docparam, 'formHtml'), \
                                                 (self.doc, 'compressFile')])
                docparam.spaceBetween = space
                docparam.formHtml = html
                self.doc.compressFile = compress
                docparam.modified = 1
                self.updateViews(self.treeView.currentTreeItem())
                self.updateCmdAvail()

    def toolsBkColor(self):
        """Set view background color"""
        background = QColor(self.option.intData('BackgroundR', 0, 255), \
                            self.option.intData('BackgroundG', 0, 255), \
                            self.option.intData('BackgroundB', 0, 255))
        newColor = QColorDialog.getColor(background, self)
        if newColor.isValid() and newColor != background:
            self.option.changeData('BackgroundR', `newColor.red()`, 1)
            self.option.changeData('BackgroundG', `newColor.green()`, 1)
            self.option.changeData('BackgroundB', `newColor.blue()`, 1)
            self.option.writeChanges()
            self.updateColors()

    def toolsTxtColor(self):
        """Set view text color"""
        foreground = QColor(self.option.intData('ForegroundR', 0, 255), \
                            self.option.intData('ForegroundG', 0, 255), \
                            self.option.intData('ForegroundB', 0, 255))
        newColor = QColorDialog.getColor(foreground, self)
        if newColor.isValid() and newColor != foreground:
            self.option.changeData('ForegroundR', `newColor.red()`, 1)
            self.option.changeData('ForegroundG', `newColor.green()`, 1)
            self.option.changeData('ForegroundB', `newColor.blue()`, 1)
            self.option.writeChanges()
            self.updateColors()

    def findReadMeFile(self, pathList, fileName):
        """Search paths for file, return line list or None""" 
        for path in pathList:
            try:
                f = open(os.path.join(path, fileName), 'r')
                lineList = f.readlines()
                f.close()
                return lineList
            except IOError:
                pass
        return None

    def helpReadMe(self):
        """View the ReadMe file"""
        if not self.helpView:
            modPath = os.path.abspath(sys.path[0])
            pathList = [helpFilePath, os.path.join(modPath, '../doc/'), \
                        modPath]
            lineList = self.findReadMeFile(filter(None, pathList), \
                                           'README.html')
            if not lineList:
                self.statusBar().message('Read Me file not found', 4000)
                return
            self.helpView = HelpView('\n'.join(lineList), \
                                     'TreeLine README File')
            self.helpView.setIcon(QPixmap(treeline_xpm))
        self.helpView.show()

    def helpAbout(self):
        """About this program"""
        QMessageBox.about(self, 'TreeLine', 'TreeLine, Version %s\n by %s' % \
                          (__version__, __author__))

    def savePrompt(self):
        """Ask for save if doc modified, return false on cancel"""
        if self.doc:
            if docparam.modified:
                text = self.fileImported and 'Save changes?' or \
                       'Save changes to "%s"?' % self.doc.fileName
                ans = QMessageBox.information(self, 'TreeLine', text, \
                                              '&Yes', '&No', '&Cancel', 0, 2)
                if ans == 0:
                    self.fileSave()
                elif ans == 2:
                    return 0
                else:
                    self.delAutoSaveFile()
            if self.option.boolData('PersistTreeState'):
                self.treeStates.saveCurrent(self.doc, self.treeView, \
                                            docparam.modified)
        return 1

    def closeEvent(self, event):
        """Ask for save if doc modified"""
        if self.savePrompt():
            if self.option.boolData('SaveWindowGeom'):
                docks = ''
                for toolbar in self.toolbars:
                    params = self.getLocation(toolbar)
                    docks = docks + ',[' + str(int(toolbar.isVisible())) \
                          + ',' + str(params[1]) + ',' + str(params[2]) \
                          + ',' + str(int(params[3])) + ',' + str(params[4]) \
                          + ']'
                self.option.changeData('ToolbarData', docks[1:], 1)
                self.option.changeData('WindowXSize', self.width(), 1)
                self.option.changeData('WindowYSize', self.height(), 1)
                self.option.changeData('WindowXPos', self.x(), 1)
                self.option.changeData('WindowYPos', self.y(), 1)
                treeWidth = self.treeView.width()
                rightWidth = self.rightStack.width()
                treePercent = int(round(treeWidth * 100.0 / \
                                  (treeWidth + rightWidth)))
                self.option.changeData('TreeSplitPercent', treePercent, 1)
                mainHeight, childHeight = self.dataOutSplit.sizes()
                outPercent = int(round(mainHeight * 100.0 / \
                                 (mainHeight + childHeight)))
                self.option.changeData('OutputSplitPercent', outPercent, 1)
                mainHeight, childHeight = self.dataEditSplit.sizes()
                editPercent = int(round(mainHeight * 100.0 / \
                                  (mainHeight + childHeight)))
                self.option.changeData('EditorSplitPercent', editPercent, 1)
                mainHeight, childHeight = self.titleListSplit.sizes()
                titlePercent = int(round(mainHeight * 100.0 / \
                                   (mainHeight + childHeight)))
                self.option.changeData('TitleSplitPercent', titlePercent, 1)
                self.option.writeChanges()
            event.accept()
        else:
            event.ignore()

    def dragEnterEvent(self, event):
        """Accept drags of files to main window"""
        event.accept(QUriDrag.canDecode(event))

    def dropEvent(self, event):
        """Drop a file onto window"""
        fileList = QStringList()
        if QUriDrag.decodeLocalFiles(event, fileList):
            if self.savePrompt():
                self.openFile(str(fileList[0]))

    def loadTypeSubMenu(self):
        """Update type select submenu with current type names"""
        names = docparam.nodeFormats.nameList()[:]
        names.sort()
        self.typeSubMenu.clear()
        id = 0
        usedShortcuts = []
        for name in names:
            shortcutPos = 0
            try:
                while name[shortcutPos] in usedShortcuts:
                    shortcutPos += 1
                usedShortcuts.append(name[shortcutPos])
                text = '%s&%s' % (name[:shortcutPos], name[shortcutPos:])
            except IndexError:
                text = name
            self.typeSubMenu.insertItem(text, id)
            id += 1

    def accelKey(self, cmdName):
        """Return shortcut key from options, or zero"""
        keyList = self.option.strData(cmdName, 1).split()
        if not keyList:
            return 0
        try:
            mainKey = eval('Qt.Key_%s' % keyList[-1])
            modKeys = [eval('Qt.%s' % key) for key in keyList[:-1]]
            return reduce(lambda x, y: x+y, [mainKey] + modKeys)
        except (AttributeError, TypeError):
            return 0

    def accelKeyText(self, cmdName):
        """Return shortcut key text from options, or ''"""
        key = self.option.strData(cmdName, 1)
        if key:
            return '+'.join(key.split()).title()
        return ''

    def restoreToolbars(self, toolbars, params):
        for i in range(len(toolbars)):
            if qVersion()[0] >= '3':
                self.moveDockWindow(toolbars[i], int(params[i][1]),
                                    int(params[i][3]), int(params[i][2]),
                                    int(params[i][4]))
            else:
                self.moveToolBar(toolbars[i], int(params[i][1]),
                                 int(params[i][3]), int(params[i][2]),
                                 int(params[i][4]))
            self.viewToolbarsMenu.setItemChecked(i, params[i][0])
            if params[i][0] == 0:
                self.toolbars[i].hide()

    def setupMenus(self):
        """Add menu and toolbar items"""
        fileMenu = QPopupMenu(self)
        self.menuBar().insertItem('&File', fileMenu)
        self.toolbars = [QToolBar(self), QToolBar(self)]
        self.parentPopup = QPopupMenu(self.treeView)
        self.childPopup = QPopupMenu(self.treeView)
        self.fullActGrp = QActionGroup(self)

        fileNewAct = QAction('New File', QIconSet(QPixmap(filenew_xpm)), \
                             '&New', self.accelKey('FileNew'), self.fullActGrp)
        fileNewAct.setStatusTip('Start a new file')
        fileNewAct.addTo(fileMenu)
        fileNewAct.addTo(self.toolbars[0])
        self.connect(fileNewAct, SIGNAL('activated()'), self.fileNew)

        fileOpenAct = QAction('Open File', QIconSet(QPixmap(fileopen_xpm)), \
                              '&Open...', self.accelKey('FileOpen'), \
                              self.fullActGrp)
        fileOpenAct.setStatusTip('Open a file from disk')
        fileOpenAct.addTo(fileMenu)
        fileOpenAct.addTo(self.toolbars[0])
        self.connect(fileOpenAct, SIGNAL('activated()'), self.fileOpen)

        fileMenu.insertSeparator()
        
        self.fileSaveAct = QAction('Save File', \
                                   QIconSet(QPixmap(filesave_xpm)), \
                                   '&Save', self.accelKey('FileSave'), \
                                   self.fullActGrp)
        self.fileSaveAct.setStatusTip('Save changes to the current file')
        self.fileSaveAct.addTo(fileMenu)
        self.fileSaveAct.addTo(self.toolbars[0])
        self.connect(self.fileSaveAct, SIGNAL('activated()'), self.fileSave)

        fileSaveAsAct = QAction('Save As', 'Save &As...', \
                                self.accelKey('FileSaveAs'), self.fullActGrp)
        fileSaveAsAct.setStatusTip('Save the file with a new name')
        fileSaveAsAct.addTo(fileMenu)
        self.connect(fileSaveAsAct, SIGNAL('activated()'), self.fileSaveAs)

        fileExportAct = QAction('Export', '&Export...', \
                                self.accelKey('FileExport'), self.fullActGrp)
        fileExportAct.setStatusTip('Export the file as html, a table or text')
        fileExportAct.addTo(fileMenu)
        self.connect(fileExportAct, SIGNAL('activated()'), self.fileExport)

        fileMenu.insertSeparator()
        self.toolbars[0].addSeparator()

        filePrintAct = QAction('Print File', \
                               QIconSet(QPixmap(fileprint_xpm)), \
                               '&Print...', self.accelKey('FilePrint'), \
                               self.fullActGrp)
        filePrintAct.setStatusTip('Print starting at the selected node')
        filePrintAct.addTo(fileMenu)
        filePrintAct.addTo(self.toolbars[0])
        self.connect(filePrintAct, SIGNAL('activated()'), self.filePrint)

        filePrintOptAct = QAction('Print Options', 'P&rint Options...', \
                                  self.accelKey('FilePrintOpt'), \
                                  self.fullActGrp)
        filePrintOptAct.setStatusTip('Set margins and page size for printing')
        filePrintOptAct.addTo(fileMenu)
        self.connect(filePrintOptAct, SIGNAL('activated()'), self.filePrintOpt)

        fileMenu.insertSeparator()
        self.recentFiles = RecentFiles(self.option, 4, 30, self)
        self.treeStates = TreeStates(self.option, self.recentFiles.fileList())
        self.connect(self.recentFiles, PYSIGNAL('listChanged'), \
                     self.treeStates.fileListChanged)
        self.connect(self.recentFiles, PYSIGNAL('askOpen'), self.recentOpen)
        self.recentFiles.initMenu(fileMenu, 100)

        fileMenu.insertSeparator()
        self.toolbars[0].addSeparator()

        fileQuitAct = QAction('Quit', '&Quit', self.accelKey('FileQuit'), \
                              self.fullActGrp)
        fileQuitAct.setStatusTip('Exit the application')
        fileQuitAct.addTo(fileMenu)
        self.connect(fileQuitAct, SIGNAL('activated()'), self.close)

        editMenu = QPopupMenu(self)
        self.menuBar().insertItem('&Edit', editMenu)
        self.editUndoAct = QAction('Undo', '&Undo', self.accelKey('EditUndo'), \
                                   self.fullActGrp)
        self.editUndoAct.setStatusTip('Undo the previous action')
        self.editUndoAct.addTo(editMenu)
        self.connect(self.editUndoAct, SIGNAL('activated()'), self.editUndo)

        self.editRedoAct = QAction('Redo', '&Redo', self.accelKey('EditRedo'), \
                                   self.fullActGrp)
        self.editRedoAct.setStatusTip('Redo the previous undo')
        self.editRedoAct.addTo(editMenu)
        self.connect(self.editRedoAct, SIGNAL('activated()'), self.editRedo)

        editMenu.insertSeparator()
        editCutAct = QAction('Cut', QIconSet(QPixmap(editcut_xpm)), \
                             'Cu&t', self.accelKey('EditCut'), self.fullActGrp)
        editCutAct.setStatusTip('Cut the branch or text to the clipboard')
        editCutAct.addTo(editMenu)
        editCutAct.addTo(self.toolbars[0])
        editCutAct.addTo(self.parentPopup)
        editCutAct.addTo(self.childPopup)
        self.connect(editCutAct, SIGNAL('activated()'), self.editCut)

        editCopyAct = QAction('Copy', QIconSet(QPixmap(editcopy_xpm)), \
                              '&Copy', self.accelKey('EditCopy'), \
                              self.fullActGrp)
        editCopyAct.setStatusTip('Copy the branch or text to the clipboard')
        editCopyAct.addTo(editMenu)
        editCopyAct.addTo(self.toolbars[0])
        editCopyAct.addTo(self.parentPopup)
        editCopyAct.addTo(self.childPopup)
        self.connect(editCopyAct, SIGNAL('activated()'), self.editCopy)

        editCopyTextAct = QAction('Copy Title Text', 'Cop&y Title Text', \
                                  self.accelKey('EditCopyText'), \
                                  self.fullActGrp)
        editCopyTextAct.setStatusTip('Copy node title text to the clipboard')
        editCopyTextAct.addTo(editMenu)
        self.connect(editCopyTextAct, SIGNAL('activated()'), self.editCopyText)

        self.editPasteActGrp = QActionGroup(self)
        editPasteAct = QAction('Paste', QIconSet(QPixmap(editpaste_xpm)), \
                               '&Paste', self.accelKey('EditPaste'), \
                               self.editPasteActGrp)
        editPasteAct.setStatusTip('Paste nodes or text from the clipboard')
        editPasteAct.addTo(editMenu)
        editPasteAct.addTo(self.toolbars[0])
        editPasteAct.addTo(self.parentPopup)
        editPasteAct.addTo(self.childPopup)
        self.connect(editPasteAct, SIGNAL('activated()'), self.editPaste)

        editPasteTextAct = QAction('PasteText', 'Pa&ste Node Text', \
                                   self.accelKey('EditPasteText'), \
                                   self.editPasteActGrp)
        editPasteTextAct.setStatusTip('Paste text from the clipboard')
        editPasteTextAct.addTo(editMenu)
        self.connect(editPasteTextAct, SIGNAL('activated()'), \
                     self.editPasteText)

        editRenameAct = QAction('Rename', 'Re&name', \
                                self.accelKey('EditRename'), self.fullActGrp)
        editRenameAct.setStatusTip('Rename the current tree entry')
        editRenameAct.addTo(editMenu)
        editRenameAct.addTo(self.parentPopup)
        editRenameAct.addTo(self.childPopup)
        self.connect(editRenameAct, SIGNAL('activated()'), \
                     self.treeView.editRename)

        editMenu.insertSeparator()
        self.parentPopup.insertSeparator()
        self.childPopup.insertSeparator()

        self.editInsertActGrp = QActionGroup(self)
        editInBeforeAct = QAction('Insert Sibling Before', \
                                  QIconSet(QPixmap(editinsertbefore_xpm)), \
                                  'Insert Sibling &Before', \
                                  self.accelKey('EditInBefore'), \
                                  self.editInsertActGrp)
        editInBeforeAct.setStatusTip('Insert new sibling before selection')
        editInBeforeAct.addTo(editMenu)
        self.connect(editInBeforeAct, SIGNAL('activated()'), self.editInBefore)

        editInAfterAct = QAction('Insert Sibling After', \
                                 QIconSet(QPixmap(editinsertafter_xpm)), \
                                 'Insert Sibling &After', \
                                 self.accelKey('EditInAfter'), \
                                 self.editInsertActGrp)
        editInAfterAct.setStatusTip('Insert new sibling after selection')
        editInAfterAct.addTo(editMenu)
        editInAfterAct.addTo(self.toolbars[1])
        editInAfterAct.addTo(self.parentPopup)
        editInAfterAct.addTo(self.childPopup)
        self.connect(editInAfterAct, SIGNAL('activated()'), self.editInAfter)

        editAddChildAct = QAction('Add Child', \
                                  QIconSet(QPixmap(editadd_xpm)), \
                                  'Add C&hild', \
                                  self.accelKey('EditAddChild'), \
                                  self.fullActGrp)
        editAddChildAct.setStatusTip('Add a new child to the selected parent')
        editAddChildAct.addTo(editMenu)
        editAddChildAct.addTo(self.toolbars[1])
        editAddChildAct.addTo(self.parentPopup)
        editAddChildAct.addTo(self.childPopup)
        self.connect(editAddChildAct, SIGNAL('activated()'), self.editAddChild)

        editMenu.insertSeparator()
        self.toolbars[1].addSeparator()
        self.parentPopup.insertSeparator()
        self.childPopup.insertSeparator()

        self.editDeleteAct = QAction('Delete Nodes', \
                                     QIconSet(QPixmap(editdel_xpm)), \
                                     '&Delete Node', \
                                     self.accelKey('EditDelete'), \
                                     self.fullActGrp)
        self.editDeleteAct.setStatusTip('Delete the selected nodes')
        self.editDeleteAct.addTo(editMenu)
        self.editDeleteAct.addTo(self.toolbars[1])
        self.editDeleteAct.addTo(self.parentPopup)
        self.editDeleteAct.addTo(self.childPopup)
        self.connect(self.editDeleteAct, SIGNAL('activated()'), self.editDelete)

        self.editPrevSibActGrp = QActionGroup(self)
        editIndentAct = QAction('Indent Nodes', \
                                QIconSet(QPixmap(editindent_xpm)), \
                                '&Indent Node', self.accelKey('EditIndent'), \
                                self.editPrevSibActGrp)
        editIndentAct.setStatusTip('Indent the selected nodes')
        editIndentAct.addTo(editMenu)
        editIndentAct.addTo(self.toolbars[1])
        editIndentAct.addTo(self.parentPopup)
        editIndentAct.addTo(self.childPopup)
        self.connect(editIndentAct, SIGNAL('activated()'), self.editIndent)

        self.editUnindentAct = QAction('Unindent Nodes', \
                                       QIconSet(QPixmap(editunindent_xpm)), \
                                       'Unind&ent Node', \
                                       self.accelKey('EditUnindent'), \
                                       self.fullActGrp)
        self.editUnindentAct.setStatusTip('Unindent the selected nodes')
        self.editUnindentAct.addTo(editMenu)
        self.editUnindentAct.addTo(self.toolbars[1])
        self.editUnindentAct.addTo(self.parentPopup)
        self.editUnindentAct.addTo(self.childPopup)
        self.connect(self.editUnindentAct, SIGNAL('activated()'), \
                     self.editUnindent)

        editMenu.insertSeparator()
        self.toolbars[1].addSeparator()
        self.parentPopup.insertSeparator()
        self.childPopup.insertSeparator()

        editMoveUpAct = QAction('MoveUp', QIconSet(QPixmap(editmoveup_xpm)), \
                                '&MoveUp', self.accelKey('EditMoveUp'), \
                                self.editPrevSibActGrp)
        editMoveUpAct.setStatusTip('Move the selected nodes up')
        editMoveUpAct.addTo(editMenu)
        editMoveUpAct.addTo(self.toolbars[1])
        editMoveUpAct.addTo(self.parentPopup)
        editMoveUpAct.addTo(self.childPopup)
        self.connect(editMoveUpAct, SIGNAL('activated()'), self.editMoveUp)

        self.editMoveDownAct = QAction('MoveDown', \
                                       QIconSet(QPixmap(editmovedown_xpm)), \
                                       'M&oveDown', \
                                       self.accelKey('EditMoveDown'), \
                                       self.fullActGrp)
        self.editMoveDownAct.setStatusTip('Move the selected nodes down')
        self.editMoveDownAct.addTo(editMenu)
        self.editMoveDownAct.addTo(self.toolbars[1])
        self.editMoveDownAct.addTo(self.parentPopup)
        self.editMoveDownAct.addTo(self.childPopup)
        self.connect(self.editMoveDownAct, SIGNAL('activated()'), \
                     self.editMoveDown)

        self.toolbars[1].addSeparator()
        self.parentPopup.insertSeparator()
        self.childPopup.insertSeparator()

        viewMenu = QPopupMenu(self)
        self.menuBar().insertItem('&View', viewMenu)

        self.viewToolbarsMenu = QPopupMenu(viewMenu)
        self.viewToolbarsMenu.setCheckable(True)
        self.viewToolbarsMenu.insertItem('&Standard', 0)
        self.viewToolbarsMenu.insertItem('&Nodes', 1)
        viewMenu.insertItem('&Toolbars', self.viewToolbarsMenu)
        self.connect(self.viewToolbarsMenu, SIGNAL('activated(int)'), \
                     self.viewToggleToolbar)

        viewMenu.insertSeparator()

        viewRightViewGrp = QActionGroup(self)
        viewOutAct = QAction('Show Data Output', 'Show &Data Output', \
                             self.accelKey('ViewDataOutput'), \
                             viewRightViewGrp, '', 1)
        viewOutAct.setStatusTip('Show data output in right view')
        viewOutAct.addTo(viewMenu)
        viewEditAct = QAction('Show Data Editor', 'Show Data &Editor', \
                              self.accelKey('ViewDataEdit'), \
                              viewRightViewGrp, '', 1)
        viewEditAct.setStatusTip('Show data editor in right view')
        viewEditAct.addTo(viewMenu)
        viewTitleAct = QAction('Show Title List', 'Show Title &List', \
                               self.accelKey('ViewTitleList'), \
                               viewRightViewGrp, '', 1)
        viewTitleAct.setStatusTip('Show title list in right view')
        viewTitleAct.addTo(viewMenu)
        self.connect(viewRightViewGrp, SIGNAL('selected(QAction*)'), \
                     self.viewSelect)
        self.viewToAct = {self.dataOutSplit: viewOutAct, \
                          self.dataEditSplit: viewEditAct, \
                          self.titleListSplit: viewTitleAct}
        self.actToView = {viewOutAct: self.dataOutSplit, \
                          viewEditAct: self.dataEditSplit, \
                          viewTitleAct: self.titleListSplit}
        viewFocusTreeAct = QAction('', '', self.accelKey('ViewGoTree'), \
                                   self.fullActGrp)
        self.connect(viewFocusTreeAct, SIGNAL('activated()'), \
                     self.viewFocusTree)

        viewMenu.insertSeparator()
        self.toolbars[0].addSeparator()

        viewChildrenGrp = QActionGroup(self)
        viewSelectAct = QAction('Show Selected Node', \
                                QIconSet(QPixmap(viewselect_xpm)), \
                                'Show Selected &Node', \
                                0, viewChildrenGrp, '', 1)
        viewSelectAct.setStatusTip('Show selected node without children ' \
                                   'in the right view')
        viewSelectAct.addTo(viewMenu)
        viewSelectAct.addTo(self.toolbars[0])
        viewChildrenAct = QAction('Show Selected Node with Children', \
                                  QIconSet(QPixmap(viewchild_xpm)), \
                                  'Show Selected Node &with Children', \
                                  0, viewChildrenGrp, '', 1)
        viewChildrenAct.setStatusTip('Show selected node with it\'s ' \
                                     'children in the right view')
        viewChildrenAct.addTo(viewMenu)
        viewChildrenAct.addTo(self.toolbars[0])
        self.connect(viewChildrenGrp, SIGNAL('selected(QAction*)'), \
                     self.viewChildren)
        self.childActToBool = {viewSelectAct: 0, viewChildrenAct: 1}
        viewSelectAct.setOn(not self.showItemChildren)
        viewChildrenAct.setOn(self.showItemChildren)

        viewDataPageUpAct = QAction('Data View Page Up', 'Data View Page &Up', \
                                    self.accelKey('ViewDataPageUp'), \
                                    self.fullActGrp)
        self.connect(viewDataPageUpAct, SIGNAL('activated()'), \
                     self.viewDataPageUp)
        viewDataPageDownAct = QAction('Data View Page Down', \
                                      'Data View &Page Down', \
                                      self.accelKey('ViewDataPageDown'), \
                                      self.fullActGrp)
        self.connect(viewDataPageDownAct, SIGNAL('activated()'), \
                     self.viewDataPageDown)

        self.toolbars[0].addSeparator()
        dataMenu = QPopupMenu(self)
        self.menuBar().insertItem('&Data', dataMenu)
        self.dataSelExistActGrp = QActionGroup(self)

        self.typeSubMenu = QPopupMenu(self)
        self.typeSubMenu.setCheckable(1)
        dataMenu.insertItem('Set &Item Type  (%s)' \
                            % self.accelKeyText('DataSetItemType'), \
                            self.typeSubMenu)
        self.parentPopup.insertItem('Set &Item Type  (%s)' \
                                    % self.accelKeyText('DataSetItemType'), \
                                    self.typeSubMenu)
        self.childPopup.insertItem('Set &Item Type  (%s)' \
                                   % self.accelKeyText('DataSetItemType'), \
                                   self.typeSubMenu)
        typeSubMenuAct = QAction('', '', self.accelKey('DataSetItemType'), \
                                 self.fullActGrp)
        self.connect(typeSubMenuAct, SIGNAL('activated()'), \
                     self.treeView.showTypeMenu)
        self.connect(self.typeSubMenu, SIGNAL('activated(int)'), \
                     self.dataTypeChange)
        self.connect(self.typeSubMenu, SIGNAL('aboutToShow()'), \
                     self.dataTypeMenuCheck)

        self.dataSetAct = QAction('Set Descendant Types', \
                                  QIconSet(QPixmap(dataset_xpm)), \
                                  '&Set Descendant Types...', \
                                  self.accelKey('DataSetDescendType'), \
                                  self, '', 1)
        self.dataSetAct.setStatusTip('Set data type of selections and children')
        self.dataSetAct.addTo(dataMenu)
        self.dataSetAct.addTo(self.toolbars[1])
        self.connect(self.dataSetAct, SIGNAL('toggled(bool)'), self.dataSet)

        dataConfigAct = QAction('Configure Data Types', \
                                QIconSet(QPixmap(dataconf_xpm)), \
                                '&Configure Data Types...', \
                                self.accelKey('DataConfigType'), \
                                self.fullActGrp)
        dataConfigAct.setStatusTip('Modify data fields & output lines')
        dataConfigAct.addTo(dataMenu)
        dataConfigAct.addTo(self.toolbars[1])
        self.connect(dataConfigAct, SIGNAL('activated()'), self.dataConfig)

        dataMenu.insertSeparator()
        self.parentPopup.insertSeparator()
        self.toolbars[1].addSeparator()
        self.dataSelParentActGrp = QActionGroup(self)
        self.dataUnrelParentActGrp = QActionGroup(self)
        self.dataSingleSelActGrp = QActionGroup(self)

        dataSortChildAct = QAction('Sort Node Children', \
                                   QIconSet(QPixmap(toolssort_xpm)), \
                                   'Sort &Node Children...', \
                                   self.accelKey('DataSortChildren'), \
                                   self.dataSelParentActGrp)
        dataSortChildAct.setStatusTip('Sort the children with chosen keys')
        dataSortChildAct.addTo(dataMenu)
        dataSortChildAct.addTo(self.toolbars[1])
        dataSortChildAct.addTo(self.parentPopup)
        dataSortChildAct.addTo(self.childPopup)
        self.connect(dataSortChildAct, SIGNAL('activated()'), \
                     self.dataSortChild)

        dataSortTypeAct = QAction('Sort Type in Branch', \
                                  'Sort &Type in Branch...', \
                                  self.accelKey('DataSortType'), \
                                  self.dataUnrelParentActGrp)
        dataSortTypeAct.setStatusTip('Sort descendants of a type with keys')
        dataSortTypeAct.addTo(dataMenu)
        self.connect(dataSortTypeAct, SIGNAL('activated()'), \
                     self.dataSortType)

        dataSortBranchAct = QAction('Sort Branch by Title', \
                                    'Sort &Branch by Title...', \
                                    self.accelKey('DataSortTitle'), \
                                    self.dataSelParentActGrp)
        dataSortBranchAct.setStatusTip('Sort all descendants by title only')
        dataSortBranchAct.addTo(dataMenu)
        self.connect(dataSortBranchAct, SIGNAL('activated()'), \
                     self.dataSortBranch)

        dataMenu.insertSeparator()

        dataFilterAct = QAction('Filter Data', 'Filter &Data...', \
                                self.accelKey('DataFilter'), \
                                self.dataUnrelParentActGrp)
        dataFilterAct.setStatusTip('Create a new file with filtered contents')
        dataFilterAct.addTo(dataMenu)
        self.connect(dataFilterAct, SIGNAL('activated()'), self.dataFilter)

        dataEditFieldAct = QAction('Change Selected Data', \
                                   'C&hange Selected Data...', \
                                   self.accelKey('DataChange'), \
                                   self.dataSelExistActGrp)
        dataEditFieldAct.setStatusTip('Edit data values for all selected nodes')
        dataEditFieldAct.addTo(dataMenu)
        self.connect(dataEditFieldAct, SIGNAL('activated()'), \
                     self.dataEditField)

        dataNumberingAct = QAction('Numbering', 'N&umbering...', \
                                   self.accelKey('DataNumber'), \
                                   self.dataSingleSelActGrp)
        dataNumberingAct.setStatusTip('Add numbering to a given data field')
        dataNumberingAct.addTo(dataMenu)
        self.connect(dataNumberingAct, SIGNAL('activated()'), \
                     self.dataNumbering)

        dataMenu.insertSeparator()

        dataAddCatAct = QAction('Add Category Level', \
                                '&Add Category Level...', \
                                self.accelKey('DataCategoryAdd'), \
                                self.dataUnrelParentActGrp)
        dataAddCatAct.setStatusTip('Insert category nodes above children')
        dataAddCatAct.addTo(dataMenu)
        self.connect(dataAddCatAct, SIGNAL('activated()'), self.dataAddCat)

        dataFlatCatAct = QAction('Flatten by Category', \
                                 '&Flatten by Category', \
                                 self.accelKey('DataCategoryFlat'), \
                                 self.dataUnrelParentActGrp)
        dataFlatCatAct.setStatusTip('Collapse data by merging fields')
        dataFlatCatAct.addTo(dataMenu)
        self.connect(dataFlatCatAct, SIGNAL('activated()'), self.dataFlatCat)

        dataMenu.insertSeparator()

        dataArrangeRefAct = QAction('Arrange by Reference', \
                                   'Arrange by &Reference...', \
                                    self.accelKey('DataRefArrange'), \
                                    self.dataSingleSelActGrp)
        dataArrangeRefAct.setStatusTip('Arrange data using parent references')
        dataArrangeRefAct.addTo(dataMenu)
        self.connect(dataArrangeRefAct, SIGNAL('activated()'), \
                     self.dataArrangeRef)

        dataFlatRefAct = QAction('Flatten by Reference', \
                                 'F&latten by Reference...', \
                                 self.accelKey('DataRefFlat'), \
                                 self.dataSingleSelActGrp)
        dataFlatRefAct.setStatusTip('Collapse data after adding references')
        dataFlatRefAct.addTo(dataMenu)
        self.connect(dataFlatRefAct, SIGNAL('activated()'), self.dataFlatRef)

        dataUpdateAct = QAction('Update by Reference', \
                                '&Update by Reference...', \
                                self.accelKey('DataRefUpdate'), \
                                self.dataSelParentActGrp)
        dataUpdateAct.setStatusTip('Update with new fields from '\
                                   'reference file')
        dataUpdateAct.addTo(dataMenu)
        self.connect(dataUpdateAct, SIGNAL('activated()'), self.dataUpdate)

        toolsMenu = QPopupMenu(self)
        self.menuBar().insertItem('&Tools', toolsMenu)

        toolsExpandAct = QAction('Expand Full Branch', '&Expand Full Branch', \
                                 self.accelKey('ToolsExpand'), \
                                 self.dataSelParentActGrp)
        toolsExpandAct.setStatusTip('Expand all children of selected node')
        toolsExpandAct.addTo(toolsMenu)
        toolsExpandAct.addTo(self.parentPopup)
        self.connect(toolsExpandAct, SIGNAL('activated()'), self.toolsExpand)

        toolsCollapseAct = QAction('Collapse Full Branch', \
                                   '&Collapse Full Branch', \
                                   self.accelKey('ToolsCollapse'), \
                                   self.dataSelParentActGrp)
        toolsCollapseAct.setStatusTip('Collapse all children of selected node')
        toolsCollapseAct.addTo(toolsMenu)
        toolsCollapseAct.addTo(self.parentPopup)
        self.connect(toolsCollapseAct, SIGNAL('activated()'), \
                     self.toolsCollapse)

        toolsMenu.insertSeparator()

        self.toolsFindAct = QAction('Find', '&Find...', \
                                    self.accelKey('ToolsFind'), \
                                    self, '', 1)
        self.toolsFindAct.setStatusTip('Find node matching text string')
        self.toolsFindAct.addTo(toolsMenu)
        self.connect(self.toolsFindAct, SIGNAL('toggled(bool)'), self.toolsFind)

        toolsSpellCheckAct = QAction('Spell Check', '&Spell Check', \
                                     self.accelKey('ToolsSpellCheck'), \
                                     self.fullActGrp)
        toolsSpellCheckAct.setStatusTip('Spell check the tree\'s text data')
        toolsSpellCheckAct.addTo(toolsMenu)
        self.connect(toolsSpellCheckAct, SIGNAL('activated()'), \
                     self.toolsSpellCheck)

        self.toolsRemXsltAct = QAction('Remove XSLT Ref', '&Remove XSLT Ref', \
                                       self.accelKey('ToolsRemXLST'), \
                                       self.fullActGrp)
        self.toolsRemXsltAct.setStatusTip('Delete reference to XSLT export')
        self.toolsRemXsltAct.addTo(toolsMenu)
        self.connect(self.toolsRemXsltAct, SIGNAL('activated()'), \
                     self.toolsRemXslt)

        toolsMenu.insertSeparator()

        toolsGenOptAct = QAction('General Options', '&General Options...', \
                                 self.accelKey('ToolsGenOptions'), \
                                 self.fullActGrp)
        toolsGenOptAct.setStatusTip('Set user preferences for all files')
        toolsGenOptAct.addTo(toolsMenu)
        self.connect(toolsGenOptAct, SIGNAL('activated()'), self.toolsGenOpt)

        toolsFileOptAct = QAction('File Options', 'File &Options...', \
                                  self.accelKey('ToolsFileOptions'), \
                                  self.fullActGrp)
        toolsFileOptAct.setStatusTip('Set preferences for this file')
        toolsFileOptAct.addTo(toolsMenu)
        self.connect(toolsFileOptAct, SIGNAL('activated()'), self.toolsFileOpt)

        toolsBkColorAct = QAction('Background Color', '&Background Color...', \
                                  self.accelKey('ToolsBackColor'), \
                                  self.fullActGrp)
        toolsBkColorAct.setStatusTip('Set view background color')
        toolsBkColorAct.addTo(toolsMenu)
        self.connect(toolsBkColorAct, SIGNAL('activated()'), self.toolsBkColor)

        toolsTxtColorAct = QAction('Text Color', '&Text Color...', \
                                   self.accelKey('ToolsTextColor'), \
                                   self.fullActGrp)
        toolsTxtColorAct.setStatusTip('Set view text color')
        toolsTxtColorAct.addTo(toolsMenu)
        self.connect(toolsTxtColorAct, SIGNAL('activated()'), \
                     self.toolsTxtColor)

        helpMenu = QPopupMenu(self)
        self.menuBar().insertItem('&Help', helpMenu)

        helpReadMeAct = QAction('View ReadMe', QIconSet(QPixmap(help_xpm)), \
                             '&View ReadMe', self.accelKey('HelpReadMe'), \
                             self.fullActGrp)
        helpReadMeAct.setStatusTip('View the ReadMe file')
        helpReadMeAct.addTo(helpMenu)
        helpReadMeAct.addTo(self.toolbars[0])
        self.connect(helpReadMeAct, SIGNAL('activated()'), self.helpReadMe)

        helpAboutAct = QAction('About TreeLine', '&About TreeLine', \
                               self.accelKey('HelpAbout'), self.fullActGrp)
        helpAboutAct.setStatusTip('About this program')
        helpAboutAct.addTo(helpMenu)
        self.connect(helpAboutAct, SIGNAL('activated()'), self.helpAbout)

        # group list to disable all controls during rename
        # would not be req'd in qt3 where full could be parent of others
        self.actGrpList = [self.fullActGrp, self.editPasteActGrp, \
                           self.editInsertActGrp, self.editPrevSibActGrp, \
                           viewRightViewGrp, viewChildrenGrp, \
                           self.toolsFindAct, self.dataSetAct, \
                           self.dataSelExistActGrp, self.dataSelParentActGrp, \
                           self.dataUnrelParentActGrp, self.dataSingleSelActGrp]
