#!/usr/bin/env python

#****************************************************************************
# treestates.py, class to keep track of recent files' meta-data
#
# Copyright (C) 2004, Jan Hustak
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#*****************************************************************************

from os import stat
from qt import QPoint
from time import time
from option import Option

class TreeStates:
    """Tracks, loads & saves tree states of recently opened files"""
    DefaultState = '0:0:0:0' # timestamp:firstVisible:selected:openNodes
    def __init__(self, option, fileList):
        self.option = option
        self.treeStates = {}
        for i in range(len(fileList)):
            state = self.option.strData(self.optionName(i), 1)
            if not state:
                state = TreeStates.DefaultState
            self.treeStates[fileList[i]] = state
        self.currentFile = ''

    def fileListChanged(self, newList):
        """Sync tree states with the list of recent files"""
        newTreeStates = {}
        for i in range(len(newList)):
            filePath = newList[i]
            treeState = self.getStateByFile(filePath)
            self.option.changeData(self.optionName(i), treeState, 1)
            newTreeStates[filePath] = treeState
        self.option.writeChanges()
        self.treeStates = newTreeStates
        self.currentFile = newList[0]

    def clear(self):
        """Clear tree states of all recent files"""
        for i in range(len(self.treeStates)):
            self.option.changeData (self.optionName(i), \
                                    TreeStates.DefaultState, 1)
        self.option.writeChanges()
        self.treeStates = {}

    def loadCurrent(self, doc, view):
        """Apply the tree state of the currently open file"""
        fileTimestamp = stat(self.currentFile).st_mtime
        stateTokens = self.getStateByFile(self.currentFile).split(':')
        if fileTimestamp > int(stateTokens[0]): # file modified externally
            interimState = str(int(fileTimestamp)) + ':0:0:0' # see DefaultState
            self.treeStates[self.currentFile] = interimState
            self.option.changeData(self.optionName(0), interimState, 1)
            stateTokens = interimState.split(':')
        self.option.writeChanges()
        allNodes = doc.root.descendantList(1)
        self.loadOpenNodes(allNodes, stateTokens[3])
        firstVisibleItem = allNodes[int(stateTokens[1])].viewData
        view.setCurrentItem(firstVisibleItem)
        view.setContentsPos(0, view.itemPos(firstVisibleItem))
        view.setSelected(allNodes[0].viewData, False) # deselect root
        selectedIndex = int(stateTokens[2])
        if selectedIndex < len(allNodes): # there is a selected node
            selectedItem = allNodes[selectedIndex].viewData
            view.setSelected(selectedItem, True)
            view.setCurrentItem(selectedItem)
        view.triggerUpdate()

    def saveCurrent(self, doc, view, modified):
        """Persist the tree state of the currently open file"""
        if not self.currentFile:
            return
        if modified:
            state = str(int(time()))
        else:
            state = self.getStateByFile(self.currentFile).split(':')[0]
        allNodes = doc.root.descendantList(1)
        state = state + ':' \
              + str(allNodes.index(view.itemAt(QPoint(0, 0)).docRef)) + ':'
        state = state + str(allNodes.index(view.currentTreeItem()))
        state = state + ':' + self.saveOpenNodes(allNodes)
        self.treeStates[self.currentFile] = state
        self.option.changeData(self.optionName(0), state, 1)
        self.option.writeChanges()

    def getStateByFile(self, filePath):
        """Retrieve the given file's tree state or the default"""
        try:
            treeState = self.treeStates[filePath]
        except KeyError:
            treeState = TreeStates.DefaultState
        if not treeState:
            treeState = TreeStates.DefaultState
        return treeState

    def optionName(self, index):
        return 'TreeState' + `index + 1`

    def loadOpenNodes(self, allNodes, stateStr):
        """Set nodes' visibility state according to stateStr"""
        openNodeIndices = eval('[' + stateStr + ']')
        for index in openNodeIndices:
            try:
                node = allNodes[index]
            except IndexError:
                return  # nodes don't exist, posssibly due to file import issues
            while node.open != 1 and node != allNodes[0]:
                node.open = 1
                if node.viewData:
                    node.viewData.setOpen(1)
                node = node.parent
    
    def saveOpenNodes(self, allNodes):
        """Serialize visible nodes into a string"""
        stateStr = ''
        if not allNodes[0].open: # root closed, return default
            return '0'
        currentGeneration = [allNodes[0]]
        while len(currentGeneration) > 0:
            nextGeneration = []
            for node in currentGeneration:
                noOpenChildren = True
                for child in node.childList:
                    if child.open:
                        nextGeneration.append (child)
                        noOpenChildren = False
                if noOpenChildren:
                    stateStr = stateStr + ',' + str(allNodes.index(node))
            currentGeneration = nextGeneration
        return stateStr[1:]
