#!/usr/bin/env python

#****************************************************************************
# treeview.py, provides classes for the main tree view
#
# TreeLine, an information storage program
# Copyright (C) 2004, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#*****************************************************************************

import copy, types, string
from treedoc import TreeDoc
import docparam
from tmpcontrol import TmpEdit
from treeitem import TreeItem
from treeformat import NodeFormat
import treemainwin
from qt import *

class TreeViewItem(QListViewItem):
    """Qt tree item, contains ref to treecore TreeItem"""
    def __init__(self, parent, docRef):
        prev = docRef.prevSibling()
        if prev:
            prev = prev.viewData
        QListViewItem.__init__(self, parent, prev)
        self.setText(0, docRef.title())
        self.setExpandable(docRef.numChildren())
        docRef.viewData = self
        self.docRef = docRef
        self.setOpen(docRef.open)

    def setOpen(self, open):
        """Called to open tree item, reads children if reqd"""
        currentItem = self.listView().currentTreeItem()
        if open and not self.childCount():
            for child in self.docRef.childList:
                TreeViewItem(child.parent.viewData, child)
        QListViewItem.setOpen(self, open)
        # scroll if opening a formerly closed item
        if open and not self.docRef.open and self.childCount():
            self.listView().ensureItemVisible(self.docRef.\
                                              childList[-1].viewData)
            self.listView().ensureItemVisible(self)
        self.docRef.open = open
        if not open and currentItem and self.docRef in \
                                        currentItem.ancestorList():
            self.listView().changeSelect([self.docRef])


    def setSelected(self, select):
        """Called when selection changes, updates core select list"""
        QListViewItem.setSelected(self, select)
        self.listView().mainWin.doc.updateSelection(self.docRef, select)


class TreeView(QListView):
    """Left pane view of tree structure"""
    maxOffset = 250
    cmdList = ('TreeSelectPrev', 'TreeSelectNext', 'TreeOpenItem', \
               'TreeCloseItem', 'TreePrevSibling', 'TreeNextSibling', \
               'TreeSelectParent', 'TreeTop', 'TreeBottom', \
               'TreePageUp', 'TreePageDown', 'TreeIncremSearch', \
               'TreeIncremNext', 'TreeIncremPrev')
    def __init__(self, mainWin, parent=None, name=None):
        QListView.__init__(self, parent, name)
        self.mainWin = mainWin
        self.renameEdit = None
        self.stopRename = 0
        self.renameTimer = QTimer(self)
        self.connect( self.renameTimer, SIGNAL('timeout()'), self.editRename)
        self.oldSelItem = None
        self.setTreeStepSize(self.mainWin.option.intData('IndentOffset', 0, \
                                                         TreeView.maxOffset))
        self.setSelectionMode(QListView.Extended)
        self.setRootIsDecorated(1)
        self.setSorting(-1)
        self.header().hide()
        self.addColumn('Main')
        self.viewport().setMouseTracking(0)
        self.viewport().setAcceptDrops(1)
        self.setAcceptDrops(1)
        self.incremSearchMode = 0
        self.incremSearchStr = ''
        self.openSelList = []
        self.cmdDict = {}
        for cmd in TreeView.cmdList:
            accelKey = self.mainWin.accelKey(cmd)
            if accelKey:
                self.cmdDict[accelKey] = cmd

    def updateTree(self, select=None):
        """Replace contents of TreeView from the doc"""
        if not select:
            select = self.mainWin.doc.root
        if type(select) != types.ListType:
            select = [select]
        else:
            select = select[:]  # avoid overwriting select list with clear
        origX, origY = (self.contentsX(), self.contentsY())
        self.clear()
        item = TreeViewItem(self, self.mainWin.doc.root)
        self.blockSignals(1)
        for node in select:
            self.setSelected(node.viewData, 1)
        self.setCurrentItem(select[-1].viewData)
        self.setContentsPos(origX, origY)
        self.ensureItemVisible(select[-1].viewData)
        self.blockSignals(0)
        self.triggerUpdate()

    def changeSelect(self, newSelect, keyboardChg=0):
        """Update view selection and doc select list"""
        if type(newSelect) != types.ListType:
            newSelect = [newSelect]
        openSearch = self.mainWin.option.boolData('OpenSearchNodes')
        openChgs = 0
        if keyboardChg and openSearch:
            for item in self.openSelList:
                item.open = 0
            openChgs = len(self.openSelList)
        self.openSelList = []
        if keyboardChg:
            for item in newSelect:
                self.openSelList.extend(item.openParents(openSearch))
            openChgs += len(self.openSelList)
            if openChgs:
                self.mainWin.updateViews(newSelect)
                return
        self.blockSignals(1)
        for node in self.mainWin.doc.selectedNodes[:]:
            self.setSelected(node.viewData, 0)
        for node in newSelect:
            self.setSelected(node.viewData, 1)
        self.setCurrentItem(newSelect[-1].viewData)
        self.ensureItemVisible(newSelect[-1].viewData)
        self.mainWin.doc.selectedNodes = newSelect[:]
        self.blockSignals(0)
        self.triggerUpdate()
        self.emit(SIGNAL('selectionChanged()'), ())

    def currentTreeItem(self):
        """Returns current treecore item"""
        item = self.currentItem()
        if item:
            return item.docRef
        return None

    def editRename(self):
        """Rename the selected tree entry"""
        item = self.currentItem()
        if not item:
            return
        self.ensureItemVisible(item)
        text = item.text(0)
        self.renameEdit = TmpEdit(text, self.viewport())
        rect = self.itemRect(item)
        offset = self.treeStepSize() * item.depth() + self.itemMargin()
        if self.rootIsDecorated():
            offset += self.treeStepSize()
        rect.setLeft(rect.left() + offset)
        self.renameEdit.setGeometry(rect)
        self.renameEdit.selectAll()
        self.renameEdit.show()
        self.renameEdit.setFocus()
        self.mainWin.disableAllControls()
        self.connect(self.renameEdit, PYSIGNAL('editDone'), self.endEdit)

    def endEdit(self):
        """Hide rename edit view and change tree"""
        if not self.renameEdit:
            return
        item = self.currentTreeItem()
        text = unicode(self.renameEdit.text())
        if item and text and text != item.title():
            docparam.undoStore.addDataUndo(item, [item])
            item.setTitle(text)
            docparam.modified = 1
            self.mainWin.updateViews(item)
        self.mainWin.enableAllControls()
        self.setFocus()
        self.renameEdit.close(1)
        # self.renameEdit.close()
        # self.viewport().removeChild(self.renameEdit)
        # del self.renameEdit
        self.renameEdit = None

    def firstVisibleItem(self):
        """Return item at top of viewport"""
        return self.itemAt(QPoint(0, self.firstChild().height() / 2)).docRef

    def lastVisibleItem(self):
        """Return item at bottom of viewport"""
        item = self.itemAt(QPoint(0, self.visibleHeight() - \
                                  self.firstChild().height() / 2))
        if not item:
            return self.mainWin.doc.root.descendantList()[-1]
        return item.docRef

    def treeSelectPrev(self):
        """Select previous item"""
        item = self.currentTreeItem().prevItem()
        if item:
            self.changeSelect(item)

    def treeSelectNext(self):
        """Select next item"""
        item = self.currentTreeItem().nextItem()
        if item:
            self.changeSelect(item)

    def treeOpenItem(self):
        """Set selection to open"""
        for item in self.mainWin.doc.selectedNodes:
            item.viewData.setOpen(1)
        if not self.mainWin.doc.selectedNodes:
            item = self.currentTreeItem()
            item.viewData.setOpen(1)
            self.changeSelect(item)

    def treeCloseItem(self):
        """Set selection to closed"""
        if self.mainWin.doc.selectedNodes:
            newSelects = []
            for item in self.mainWin.doc.selectedNodes:
                if item.open and item.childList:
                    if item not in newSelects:
                        newSelects.append(item)
                elif item.parent and item.parent not in newSelects:
                    newSelects.append(item.parent)
            for item in newSelects:
                item.viewData.setOpen(0)
            newSelects = [item for item in newSelects \
                               if item.allAncestorsOpen()]
            if newSelects:
                self.changeSelect(newSelects)
        else:
            item = self.currentTreeItem()
            if item.open and item.childList:
                item.viewData.setOpen(0)
                self.changeSelect(item)
            elif item.parent:
                item.parent.viewData.setOpen(0)
                self.changeSelect(item.parent)

    def treePrevSibling(self):
        """Select previous sibling item"""
        item = self.currentTreeItem().prevSibling()
        if item:
            self.changeSelect(item, 1)

    def treeNextSibling(self):
        """Select next sibling item"""
        item = self.currentTreeItem().nextSibling()
        if item:
            self.changeSelect(item, 1)

    def treeSelectParent(self):
        """Select parent item"""
        item = self.currentTreeItem().parent
        if item:
            self.changeSelect(item, 1)

    def treeTop(self):
        """Select root item"""
        item = self.mainWin.doc.root
        self.changeSelect(item)

    def treeBottom(self):
        """Select last open item in tree"""
        item = self.mainWin.doc.root.descendantList()[-1]
        self.changeSelect(item)

    def treePageUp(self):
        """Move up one page"""
        item = self.firstVisibleItem()
        self.scrollBy(0, -self.visibleHeight())
        self.changeSelect(item)

    def treePageDown(self):
        """Move down one page"""
        item = self.lastVisibleItem()
        self.scrollBy(0, self.visibleHeight())
        self.changeSelect(item)

    def treeIncremSearch(self):
        """Begin iterative search"""
        self.incremSearchMode = 1
        self.incremSearchStr = ''
        self.mainWin.statusBar().message('Search for:')

    def doIncremSearch(self):
        """Search for searchStr in all titles"""
        self.mainWin.statusBar().message('Search for: %s' % \
                                         self.incremSearchStr)
        item = self.mainWin.doc.findTitleText(self.currentTreeItem(), \
                                              self.incremSearchStr)
        if item:
            self.changeSelect(item, 1)
            self.mainWin.statusBar().message('Search for: %s' % \
                                             self.incremSearchStr)
        else:
            self.mainWin.statusBar().message('Search for: %s  (not found)' % \
                                             self.incremSearchStr)

    def treeIncremNext(self):
        """Search for next occurance of increm string"""
        if self.incremSearchStr:
            item = self.mainWin.doc.findNextTitle(self.currentTreeItem(), \
                                                  self.incremSearchStr, 1)
            if item:
                self.changeSelect(item, 1)
                self.mainWin.statusBar().message('Next:  %s' % \
                                                 self.incremSearchStr)
            else:
                self.mainWin.statusBar().message('Next:  %s  (not found)' % \
                                                 self.incremSearchStr)

    def treeIncremPrev(self):
        """Search for previous occurance of increm string"""
        if self.incremSearchStr:
            item = self.mainWin.doc.findNextTitle(self.currentTreeItem(), \
                                                  self.incremSearchStr, 0)
            if item:
                self.changeSelect(item, 1)
                self.mainWin.statusBar().message('Previous:  %s' % \
                                                 self.incremSearchStr)
            else:
                self.mainWin.statusBar().message('Previous:  %s  (not found)' \
                                                 % self.incremSearchStr)

    def letterSearch(self, char, forward=1):
        """Move to the next or previous visible item starting with letter"""
        fullList = self.mainWin.doc.root.descendantList()
        currentPos = fullList.index(self.currentTreeItem())
        fullList = fullList[currentPos+1:] + fullList[:currentPos]
        matches = [item for item in fullList \
                   if item.title()[0].upper() == char]
        if matches:
            if forward:
                self.changeSelect(matches[0], 1)
            else:
                self.changeSelect(matches[-1], 1)

    def showTypeMenu(self):
        """Show popup menu for changing the item type"""
        self.ensureItemVisible(self.currentItem())
        rect = self.itemRect(self.currentItem())
        pt = self.mapToGlobal(QPoint(rect.center().x(), rect.bottom()))
        self.mainWin.typeSubMenu.popup(pt)

    def itemAtMouse(self, pt):
        """Return view item at pt"""
        clickedItem = self.itemAt(self.contentsToViewport(pt))
        if clickedItem:
            offset = self.treeStepSize() * clickedItem.depth() + \
                     self.itemMargin()
            if self.rootIsDecorated():
                offset += self.treeStepSize()
            if pt.x() <= offset:
                return None
        return clickedItem

    def contentsMousePressEvent(self, event):
        """Mouse press down event saves selected item for rename"""
        clickedItem = self.itemAtMouse(event.pos())
        if event.button() == Qt.LeftButton and clickedItem and \
               self.currentItem().isSelected():
            self.oldSelItem = self.currentItem()
        else:
            self.oldSelItem = None
        if event.button() == Qt.RightButton and not clickedItem:
            return      # skip unselecting right click
        if self.incremSearchMode:
            self.incremSearchMode = 0
            self.mainWin.statusBar().clear()
        QListView.contentsMousePressEvent(self, event)

    def contentsMouseReleaseEvent(self, event):
        """Mouse release event for rename and popup menus"""
        clickedItem = self.itemAtMouse(event.pos())
        if clickedItem:
            if event.button() == Qt.RightButton and \
                    not self.renameTimer.isActive():
                if clickedItem.docRef.numChildren():
                    popup = self.mainWin.parentPopup
                else:
                    popup = self.mainWin.childPopup
                pt = self.mapToGlobal(self.contentsToViewport(event.pos()))
                popup.popup(pt)
            elif self.mainWin.option.boolData('ClickRename') and \
                 event.state() == Qt.LeftButton and clickedItem == \
                 self.oldSelItem and not self.stopRename:
                self.renameTimer.start(1000, 1)
            QListView.contentsMouseReleaseEvent(self, event) 
        self.stopRename = 0

    def contentsMouseDoubleClickEvent(self, event):
        """Mouse double click event, stops rename"""
        self.renameTimer.stop()
        self.stopRename = 1
        if self.incremSearchMode:
            self.incremSearchMode = 0
            self.mainWin.statusBar().clear()
        QListView.contentsMouseDoubleClickEvent(self, event)

    def focusInEvent(self, event):
        """Stop rename on focus click"""
        if event.reason() == QFocusEvent.Mouse:
            self.stopRename = 1
        QListView.focusInEvent(self, event)

    def focusOutEvent(self, event):
        """Stop incremental search on focus loss"""
        if self.incremSearchMode:
            self.incremSearchMode = 0
            self.mainWin.statusBar().clear()
        QListView.focusOutEvent(self, event)

    def keyPressEvent(self, event):
        """Bind keys to functions"""
        keyText = str(event.text())
        combinedKey = event.key()
        if event.state() & Qt.ShiftButton:
            combinedKey += Qt.SHIFT
        if event.state() & Qt.ControlButton:
            combinedKey += Qt.CTRL
        if event.state() & Qt.AltButton:
            combinedKey += Qt.ALT
        # if event.state() & Qt.MetaButton:
            # combinedKey += Qt.Meta
        cmd = self.cmdDict.get(combinedKey, '')
        if self.incremSearchMode:
            if event.key() in (Qt.Key_Return, Qt.Key_Enter, Qt.Key_Escape):
                self.incremSearchMode = 0
                self.mainWin.statusBar().clear()
            elif event.key() == Qt.Key_Backspace and self.incremSearchStr:
                self.incremSearchStr = self.incremSearchStr[:-1]
                self.doIncremSearch()
            elif cmd in ('TreeIncremNext', 'TreeIncremPrev'):
                self.incremSearchMode = 0
                eval('self.t%s()' % cmd[1:])
            elif keyText and keyText in string.printable:
                self.incremSearchStr += keyText
                self.doIncremSearch()
            event.accept()
            return
        if cmd:
            eval('self.t%s()' % cmd[1:])
            event.accept()
        elif 'a' <= keyText <= 'z':
            self.letterSearch(keyText.upper(), 1)
            event.accept()
        elif 'A' <= keyText <= 'Z':
            self.letterSearch(keyText, 0)
            event.accept()
        elif (combinedKey == Qt.Key_Return or combinedKey == Qt.Key_Enter) and \
           self.mainWin.option.boolData('InsertOnEnter'):
            if self.currentTreeItem().parent:
                self.mainWin.editInAfter()
            else:
                self.mainWin.editAddChild()
            event.accept()
        # else:
            # QListView.keyPressEvent(self, event)

    def contentsMouseMoveEvent(self, event):
        """Move event for drag & drop"""
        if self.renameEdit:
            return
        copyFormat = treemainwin.TreeMainWin.copyFormat
        if self.itemAtMouse(event.pos()) and self.mainWin.doc.selectedNodes:
            if len(self.mainWin.doc.selectedNodes) > 1:
                docparam.nodeFormats.addIfMissing(treemainwin.TreeMainWin.\
                                                  copyFormat)
                item = TreeItem(None, copyFormat)
                for node in self.mainWin.doc.selectedNodes:
                    item.childList.append(copy.copy(node))
                    item.childList[-1].parent = item
                oldList = self.mainWin.doc.selectedNodes[:]
            else:
                item = self.mainWin.doc.selectedNodes[0]
                oldList = [item]
            oldCurrent = self.currentTreeItem()
            dragObj = QTextDrag(u'\n'.\
                      join(item.branchXml([copyFormat])), self)
            docparam.nodeFormats.removeQuiet(treemainwin.TreeMainWin.copyFormat)
            # check for move & can't move to descendant (or to self)
            if dragObj.drag(QDragObject.DragDefault) \
               and dragObj.target() == self.viewport() and \
               self.currentTreeItem() and not \
               filter(None, [node.hasDescendant(self.currentTreeItem()) \
                             for node in oldList]):
                for item in oldList:
                    item.delete()
                self.mainWin.updateViews(self.mainWin.doc.selectedNodes)
            elif not self.currentTreeItem():
                self.setCurrentItem(oldCurrent.viewData)

    def contentsDragEnterEvent(self, event):
        """Starting drag event"""
        event.accept(QTextDrag.canDecode(event) and \
                     self.mainWin.option.boolData('DragTree'))

    def contentsDropEvent(self, event):
        """End drag & drop"""
        item = self.itemAtMouse(event.pos())
        oldList = self.mainWin.doc.selectedNodes[:]
        text = QString('')
        copyFormat = treemainwin.TreeMainWin.copyFormat
        if self.mainWin.option.boolData('DragTree') and \
           QTextDrag.decode(event, text) and item and item.docRef not in \
           self.mainWin.doc.selectedNodes:
            newItem = self.mainWin.doc.readXmlString(unicode(text), \
                                                     event.source() != self)
            if newItem:
                if newItem.data:
                    itemList = [newItem]
                else:
                    itemList = newItem.childList
                parent = item.docRef
                undoParents = [parent] + filter(None, \
                                                [old.parent for old in oldList])
                docparam.undoStore.addChildListUndo(undoParents, [parent])
                for node in itemList:
                    parent.addTree(node)
                parent.open = 1
                docparam.nodeFormats.removeQuiet(treemainwin.TreeMainWin.\
                                                 copyFormat)
                self.mainWin.updateViews(itemList)
            else:
                print 'Error reading XML string'
        elif QUriDrag.canDecode(event):
            self.mainWin.dropEvent(event)
