#!/usr/bin/env python

#****************************************************************************
# undo.py, provides a classes to store and execute undo & redo operations
#
# TreeLine, an information storage program
# Copyright (C) 2004, Douglas W. Bell
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the GNU General Public License, Version 2.  This program is
# distributed in the hope that it will be useful, but WITTHOUT ANY WARRANTY.
#*****************************************************************************

import copy
from treeitem import TreeItem
import docparam


class UndoRedoStore:
    """Stores list of info needed for undo or redo steps"""
    def __init__(self, levels=5):
        self.levels = levels
        self.undoList = []
        self.tmpRedoRef = None  # ref for clearing redo list when undos added

    def addUndoObj(self, obj, clearRedo=1):
        """Add undo/redo object to list, truncate if longer than levels"""
        self.undoList.append(obj)
        if self.levels:
            self.undoList = self.undoList[-self.levels:]
        else:
            self.undoList = []
        if self.tmpRedoRef and clearRedo:
            self.tmpRedoRef.undoList = []
            self.tmpRedoRef = None

    def addDataUndo(self, items, selection=[], skipSame=0, clearRedo=1):
        """Syntactic sugar, adds tree item data changes,
           skips adding item if skipSame is true and a single item matches"""
        if skipSame and self.undoList and \
             isinstance(self.undoList[-1], DataUndo) and \
             len(self.undoList[-1].dataList) == 1 and \
             items == self.undoList[-1].dataList[0][0]:
            return
        self.addUndoObj(DataUndo(items, selection), clearRedo)

    def addChildListUndo(self, items, selection=[], skipSame=0, clearRedo=1):
        """Syntactic sugar, adds tree item child list changes,
           skips adding item if skipSame is true and a single item matches"""
        if skipSame and self.undoList and \
             isinstance(self.undoList[-1], ChildListUndo) and \
             len(self.undoList[-1].dataList) == 1 and \
             items == self.undoList[-1].dataList[0][0]:
            return
        self.addUndoObj(ChildListUndo(items, selection), clearRedo)

    def addParentListUndo(self, items, selection=[], skipSame=0, clearRedo=1):
        """Adds child list changes from the parents of items,
           skips adding item if skipSame is true and a single item matches"""
        if isinstance(items, TreeItem):
            items = [items]
        if skipSame and self.undoList and \
             isinstance(self.undoList[-1], ChildListUndo) and \
             len(self.undoList[-1].dataList) == 1 and len(items) == 1 and \
             items[0].parent == self.undoList[-1].dataList[0][0]:
            return
        self.addUndoObj(ChildListUndo([item.parent for item in items], \
                                      selection), clearRedo)

    def addTypeUndo(self, items, selection=[], clearRedo=1):
        """Syntactic sugar, adds tree item type setting changes"""
        self.addUndoObj(TypeUndo(items, selection), clearRedo)

    def addParamUndo(self, varList, selection=[], clearRedo=1):
        """Syntactic sugar, adds general immutable param changes"""
        self.addUndoObj(ParamUndo(varList, selection), clearRedo)

    def addFormatUndo(self, formatList, fileInfoFormat, fieldRenameDict, \
                      typeRenameDict, rootItem, selection=[], clearRedo=1):
        """Syntactic sugar, adds type formatting changes"""
        self.addUndoObj(FormatUndo(formatList, fileInfoFormat, \
                                   fieldRenameDict, typeRenameDict, \
                                   rootItem, selection), \
                        clearRedo)

    def addBranchUndo(self, items, docRef, selection=[], clearRedo=1):
        """Syntactic sugar, adds tree branch changes"""
        self.addUndoObj(BranchUndo(items, docRef, selection), clearRedo)

    def undo(self, redoRef=None, currentSelect=[]):
        """Restore saved state for next undo item,
           add undo item to redoRef, return saved selection"""
        if self.undoList:
            self.tmpRedoRef = redoRef
            obj = self.undoList.pop()
            obj.undo(redoRef, currentSelect)
            docparam.modified = obj.docModified
            return obj.selection
        return []

    def removeLastUndo(self):
        """Remove most recent undo from list"""
        self.undoList = self.undoList[:-1]


class UndoBase:
    """Base class for undo/redo info"""
    def __init__(self, selection=[]):
        self.dataList = []  # list of changes
        self.selection = selection[:]  # store selected nodes
        self.docModified = docparam.modified


class DataUndo(UndoBase):
    """Info for undo/redo of tree item data changes"""
    def __init__(self, items, selection=[]):
        """Pass an item or list of items with data to save"""
        UndoBase.__init__(self, selection)
        if isinstance(items, TreeItem):
            items = [items]
        for item in items:
            self.dataList.append((item, item.data.copy()))

    def undo(self, redoRef=None, currentSelect=[]):
        """Restore saved state"""
        if redoRef:
            redoRef.addDataUndo([tuple[0] for tuple in self.dataList], \
                                currentSelect, 0, 0)
        for item, data in self.dataList:
            item.data = data


class ChildListUndo(UndoBase):
    """Info for undo/redo of tree item child lists"""
    def __init__(self, items, selection=[]):
        """Pass an item or list of items with child lists to save"""
        UndoBase.__init__(self, selection)
        if isinstance(items, TreeItem):
            items = [items]
        for item in items:
            if item:
                self.dataList.append((item, item.childList[:]))

    def undo(self, redoRef=None, currentSelect=[]):
        """Restore saved state"""
        if redoRef:
            redoRef.addChildListUndo([tuple[0] for tuple in self.dataList], \
                                     currentSelect, 0, 0)
        for item, childList in self.dataList:
            item.childList = childList
            for child in childList:
                child.parent = item


class TypeUndo(UndoBase):
    """Info for undo/redo of tree item type settings"""
    def __init__(self, items, selection=[]):
        """Pass an item or list of items with type settings to save"""
        UndoBase.__init__(self, selection)
        if isinstance(items, TreeItem):
            items = [items]
        for item in items:
            self.dataList.append((item, item.nodeFormat))

    def undo(self, redoRef=None, currentSelect=[]):
        """Restore saved state"""
        if redoRef:
            redoRef.addTypeUndo([tuple[0] for tuple in self.dataList], \
                                currentSelect, 0)
        for item, format in self.dataList:
            item.nodeFormat = format


class ParamUndo(UndoBase):
    """Info for undo/redo of general immutable parameters"""
    def __init__(self, varList, selection=[]):
        """Pass a list containing a tuple for each variable,
           each tuple has the variable's owner and name string"""
        UndoBase.__init__(self, selection)
        for varOwner, varName in varList:
            value = varOwner.__dict__[varName]
            self.dataList.append((varOwner, varName, value))

    def undo(self, redoRef=None, currentSelect=[]):
        """Restore saved state"""
        if redoRef:
            redoRef.addParamUndo([tuple[:2] for tuple in self.dataList], \
                                 currentSelect, 0)
        for varOwner, varName, value in self.dataList:
            varOwner.__dict__[varName] = value


class FormatUndo(UndoBase):
    """Info for undo/redo of type formatting changes"""
    def __init__(self, formatList, fileInfoformat, fieldRenameDict, \
                 typeRenameDict, rootItem, selection=[]):
        """Pass the full list of type formats to save"""
        UndoBase.__init__(self, selection)
        self.formatList = copy.deepcopy(formatList)
        self.fileInfoFormat = copy.deepcopy(fileInfoformat)
        self.rootItem = rootItem
        self.fieldRenameDict = {}
        for typeName, origField in fieldRenameDict.keys():  # swap dict around
            newField = fieldRenameDict[(typeName, origField)]
            self.fieldRenameDict[(typeName, newField)] = origField
        self.typeRenameDict = {}
        for origType, newType in typeRenameDict.items(): #swap type dict around
            self.typeRenameDict[newType] = origType

    def undo(self, redoRef=None, currentSelect=[]):
        """Restore saved state"""
        if redoRef:
            redoRef.addFormatUndo(docparam.nodeFormats, \
                                  docparam.fileInfoItem.nodeFormat, \
                                  self.fieldRenameDict, self.typeRenameDict, \
                                  self.rootItem, currentSelect, 0)
        docparam.fileInfoItem.nodeFormat = self.fileInfoFormat
        if self.fieldRenameDict:
            docparam.nodeFormats.renameFields(self.fieldRenameDict, \
                                              self.rootItem)
        docparam.nodeFormats.update(self.formatList, self.typeRenameDict)
        docparam.nodeFormats.updateAllLineFields()


class BranchUndo(UndoBase):
    """Info for undo/redo of tree branches,
       all data, children, etc., and full format"""
    def __init__(self, items, docRef, selection=[]):
        """Pass an item or list of items that are the roots of the branches"""
        UndoBase.__init__(self, selection)
        self.docRef = docRef
        self.formatList = copy.deepcopy(docparam.nodeFormats)
        if isinstance(items, TreeItem):
            items = [items]
        for item in items:
            self.dataList.append((item, '\n'.join(item.branchXml())))

    def undo(self, redoRef=None, currentSelect=[]):
        """Restore saved state"""
        if redoRef and len(self.dataList) == 1:
            redoRef.addBranchUndo([tuple[0] for tuple in self.dataList], \
                                  self.docRef, currentSelect, 0)
        docparam.nodeFormats.update(self.formatList, {})
        self.selection = []
        for item, xmlText in self.dataList:
            newItem = self.docRef.readXmlString(xmlText, 1)
            self.selection.append(newItem)
            if item.parent:
                item.parent.childList[item.parent.childPos(item)] = newItem
                newItem.parent = item.parent
            else:
                self.docRef.root = newItem
        if redoRef and len(self.dataList) == 1:
            redoRef.undoList[-1].dataList[0] = (newItem, \
                                           redoRef.undoList[-1].dataList[0][1])
